// =================================================================
// 	File: bbox.cpp								12/27/1999
//	By Konstantin Khanyants						kon@unix.teple.edu
//	Description:
//		Bounding box-related routines.
//
//	Original Copyright:	    The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
// 
// =================================================================

#include "stdafx.h"
#include "include/Oort.h"

// AxisAlignedBox implementations
Vector3D
AxisAlignedBox::MinV() const
{
    return minV;
}

Vector3D
AxisAlignedBox::MaxV() const
{
    return maxV;
}

int
AxisAlignedBox::Unbounded() const
{
    return minV == Vector3D(-MAXFLOAT) || maxV == Vector3D(-MAXFLOAT);
}

void
AxisAlignedBox::Include(const Vector3D& newpt)
{
    Minimize(minV, newpt);
    Maximize(maxV, newpt);
}

void
AxisAlignedBox::Include(const AxisAlignedBox& bbox)
{
    Minimize(minV, bbox.minV);
    Maximize(maxV, bbox.maxV);
}

// ComputeMinMaxT computes the minimum and maximum parameters
// of intersection with the ray; it returns 1 if the ray hits
// the bounding box and 0 if it does not.
int
AxisAlignedBox::ComputeMinMaxT(Ray& ray, float *tmin, float *tmax) const
{
    float t1, t2;
    float minx, maxx, miny, maxy, minz, maxz;
    if (fabs(ray.dir.x) < 0.001) {
	if (minV.x < ray.loc.x && maxV.x > ray.loc.x) {
	    minx = -MAXFLOAT;
	    maxx = MAXFLOAT;
	}
	else
	    return 0;
    }
    else {
	t1 = (minV.x - ray.loc.x) / ray.dir.x;
	t2 = (maxV.x - ray.loc.x) / ray.dir.x;
	if (t1 < t2) {
	    minx = t1;
	    maxx = t2;
	}
	else {
	    minx = t2;
	    maxx = t1;
	}
	if (maxx < 0)
	    return 0;
    }

    if (fabs(ray.dir.y) < 0.001) {
	if (minV.y < ray.loc.y && maxV.y > ray.loc.y) {
	    miny = -MAXFLOAT;
	    maxy = MAXFLOAT;
	}
	else
	    return 0;
    }
    else {
	t1 = (minV.y - ray.loc.y) / ray.dir.y;
	t2 = (maxV.y - ray.loc.y) / ray.dir.y;
	if (t1 < t2) {
	    miny = t1;
	    maxy = t2;
	}
	else {
	    miny = t2;
	    maxy = t1;
	}
	if (maxy < 0)
	    return 0;
    }

    if (fabs(ray.dir.z) < 0.001) {
	if (minV.z < ray.loc.z && maxV.z > ray.loc.z) {
	    minz = -MAXFLOAT;
	    maxz = MAXFLOAT;
	}
	else
	    return 0;
    }
    else {
	t1 = (minV.z - ray.loc.z) / ray.dir.z;
	t2 = (maxV.z - ray.loc.z) / ray.dir.z;
	if (t1 < t2) {
	    minz = t1;
	    maxz = t2;
	}
	else {
	    minz = t2;
	    maxz = t1;
	}
	if (maxz < 0)
	    return 0;
    }

    *tmin = minx;
    if (miny > *tmin)
	*tmin = miny;
    if (minz > *tmin)
	*tmin = minz;

    *tmax = maxx;
    if (maxy < *tmax)
	*tmax = maxy;
    if (maxz < *tmax)
	*tmax = maxz;
	
    return 1;
}

int
BoundingBox::NearestInt(Ray& ray, float& t, float maxt)
{
    float tmin, tmax;

    Statistics::Intersections::BBox++;

    if (! AxisAlignedBox::ComputeMinMaxT(ray, &tmin, &tmax))
	return 0;
    if (tmax < tmin)
	return 0;
    if (tmin < 0) {
	// No intersection if both parameters are negative
	if (tmax < 0)
	    return 0;
	// The ray hit the bounding box
	// tmax is parameter of intersection
	t = (tmax < maxt) ? tmax : maxt;
    }
    else {
	// The ray hit the bounding box
	// with parameter tmin.
	t = (tmin < maxt) ? tmin : maxt;
    }
    return 1;
}

SpanList *
BoundingBox::FindAllIntersections(Ray& ray)
{
//REM    cerr << "BoundingBox::FindAllIntersections called\n";
//    exit(1);
    return 0;
}

void
BoundingBox::ApplyTransform(const Matrix &tform)
{
    AxisAlignedBox bbox(Vector3D(MAXFLOAT), Vector3D(-MAXFLOAT));

    for (ObjectList::Iterator sc(contents); sc.Valid(); sc.GotoNext()) {
	sc.Contents()->ApplyTransform(tform);
	bbox = Union(bbox, sc.Contents()->BBox());
    }
    minV = bbox.MinV();
    maxV = bbox.MaxV();
}

Vector3D
BoundingBox::FindNormal(const Vector3D& intersection)
{
//rem    cerr << "BoundingBox::FindNormal called\n";
    return Vector3D(0);
}

Object3D *
BoundingBox::Dup() const
{
    BoundingBox *ret = new BoundingBox;

    for (ObjectList::Iterator sc((ObjectList &)contents); sc.Valid(); sc.GotoNext())
	ret->AddObject(sc.Contents()->Dup());

    return ret;
}

int
BoundingBox::IsInside(Vector3D& v)
{
    return ! (v.x < minV.x ||
	      v.x > maxV.x ||
	      v.y < minV.y ||
	      v.y > maxV.y ||
	      v.z < minV.z ||
	      v.z > maxV.z);
}

float
BoundingBox::PtDistance(Vector3D& v)
{
    cerr << "BoundingBox::PtDistance called\n";
    exit(2);
    return 0;
}

void
BoundingBox::Describe()
{
	AfxMessageBox("Bounding Box");
	for (ObjectList::Iterator bb(contents); bb.Valid(); bb.GotoNext())
		bb.Contents()->Describe();

}

void
BoundingVolume::AddObject(const Object3D& NewObj)
{
    contents.AddToList(NewObj.Dup());
}

void
BoundingVolume::AddObject(Object3D *NewObj)
{
    contents.AddToList(NewObj);
}

/*void
BoundingVolume::TransformSurface(SurfaceList& clist, const Matrix& tform)
{
    ObjectList::Iterator sc(contents);

    while (sc.Valid()) {
	sc.Contents()->TransformSurface(clist, tform);
	sc.GotoNext();
    }
}*/

float
AxisAlignedBox::SurfaceArea() const
{
    Vector3D ext = maxV - minV;
    return 2 * (ext.x * ext.y +
		ext.x * ext.z +
		ext.y * ext.z);
}

float
BoundingBox::SurfaceArea() const
{
    return AxisAlignedBox::SurfaceArea();
}

AxisAlignedBox
BoundingBox::BBox() const
{
    return AxisAlignedBox(MinV(), MaxV());
}

void
BoundingBox::AddObject(const Object3D& NewObj)
{
    BoundingVolume::AddObject(NewObj);
    Include(NewObj.BBox());
}

void
BoundingBox::AddObject(Object3D *NewObj)
{
    BoundingVolume::AddObject(NewObj);
    Include(NewObj->BBox());
}
