// ================================================================
//  File: bump.cpp							12/27/1999
//	By Konstantin Khanyants					kon@unix.temple.edu
//	Description:
//		Implementation file for bump maps.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ================================================================

#include "stdafx.h"
#include "include/Oort.h"

// =========================================================
// BumpMapReference
//	Encapsulates a bump map that also takes a reference
//	matrix.  Reference matrices are applied to input
// 	vectors before using noise & other tools to decide
//	how to perturb the normal.
// =========================================================
BumpMapReference::BumpMapReference(const Matrix& _reference)
{
    reference = _reference;
}

Vector3D
BumpMapReference::PreprocessVector(ShadingInfo& shade)
{
    return reference * Vector3D(shade.p);
}


void
BumpMapReference::ApplyTransform(const Matrix& tform)
{
    reference *= tform;
}

void
BumpMapReference::PreMulTransform(const Matrix& tform)
{
    reference = tform * reference;
}

// =========================================================
// ReverseNormal
//	Simply reverses the input normal.
// =========================================================
void
ReverseNormal::PerturbNormal(ShadingInfo& shade)
{
    if (bump)
	bump->PerturbNormal(shade);
    shade.normal = -shade.normal;
}

// =========================================================
// TwoSided
//	Reverses the normal, if necessary, so it faces
//	toward the incident direction.
// =========================================================
void
TwoSided::PerturbNormal(ShadingInfo& shade)
{
    if (bump)
	bump->PerturbNormal(shade);
    if (DotProd(shade.incident, shade.normal) > 0)
	shade.normal = -shade.normal;
}

// =========================================================
// WaveSource
//	Precomputes a set of vector noise values.
// =========================================================
WaveSource::WaveSource(int N, NoiseMaker *nm)
{
    waves = new Vector3D[n = N];
    for (int i = 0; i < n; i++)
	waves[i] = Normalize(nm->DNoise(Vector3D(float(i), 0, 0)));
}

// =========================================================
// FreqSource
//	Precomputes a set of solid noise values.
// =========================================================
FreqSource::FreqSource(int N, NoiseMaker *nm)
{
    freqs = new float[n = N];
    for (int i = 0; i < n; i++)
	freqs[i] = float((rand() & (RAND_MAX - 1)) / RAND_MAX + 0.01);
}

// =========================================================
// Ripples
//	Makes a surface appear rippled.
// =========================================================
void
Ripples::PerturbNormal(ShadingInfo& shade)
{
    Vector3D ret, point;
    double length, scalar, index;

    ret = shade.normal;
    for (int i = 0; i < ws->N(); i++) {
	point = shade.p - ws->GetWave(i);
	length = DotProd(point, point);
	if (length < Limits::Small)
	    length = 1 - length;
	length = sqrt(length);
	index = length * freq + phase;
	scalar = cycloidal(float(index)) * bumpamt;
	point *= float(scalar / length / ws->N());
	ret += point;
    }
    shade.normal = Normalize(ret);
}

BumpMap *
Ripples::Dup() const
{
    return new Ripples(ws, fs, freq, phase, bumpamt, reference);
}

// =========================================================
// Dented
//	Displaces the intersection point to make the object
//	look dented.  Not sure this can work correctly,
//	since the normal is not recomputed correctly.
// =========================================================
Dented::Dented(float _Km, const Matrix& mat): 
    BumpMapReference(mat)
{
    Km = _Km;
}

void
Dented::PerturbNormal(ShadingInfo& shade)
{
    float size = 1, mag = 0;
    Vector3D p = PreprocessVector(shade);
    for (int i = 0; i < 6; i++) {
	mag += float(fabs(0.5 - GlobalNoise::Noise->Noise(p * size)) / size);
	size *= 2;
    }
    shade.p -= shade.normal * mag * mag * mag * Km;
    shade.normal = shade.obj->FindNormal(p);
}

BumpMap *
Dented::Dup() const
{
    return new Dented(Km, reference);
}

// ---------------------------------------------------------
// ImageBumpMap implementation
// ---------------------------------------------------------

ImageBumpMap::ImageBumpMap(const Matrix& ref,
		   int Once,
		   const char *filename,
		   float rot, 
		   float bumpamt,
		   float xscale, float yscale) : BumpMapReference(ref)
{
    once = (Once != 0);
//	file = new ImageFile(FALSE);	//say no to the Open File dialog
//	file->OpenFile(filename);		//open the file quietly
	rot_angle = rot;
	bump_amt = bumpamt;
    scale = Vector3D(xscale, yscale, 0);

	file = new ImageFile(FALSE);	//say no to the Open File dialog
	if (!file->OpenFile(filename)) {		//try to open the file quietly
		CString pcsMsg = "Can't open texture file ";
		pcsMsg += filename;
		AfxMessageBox(pcsMsg);
	}
}

ImageBumpMap::ImageBumpMap(const Matrix& ref,
		   int Once,
		   ImageFile *File,
		   float rot,
		   float bumpamt,
		   float xscale, float yscale) : 
    BumpMapReference(ref)
{
    once = (Once != 0);
//    mapping = map;
    file = File;
    rot_angle = rot;
    bump_amt = bumpamt;
    scale = Vector3D(xscale, yscale, 0);
}

//added 3/20/2000
ImageBumpMap::ImageBumpMap(const ImageBumpMap& initBump) : 
	BumpMapReference(initBump.reference)
{
	once = initBump.once;
    file = initBump.file->Dup();
    rot_angle = initBump.rot_angle;
    bump_amt = initBump.bump_amt;
    scale = initBump.scale;
}

ImageBumpMap::~ImageBumpMap()
{	
	if (file)
		delete file;
}

Vector3D
ImageBumpMap::PreprocessVector(ShadingInfo& shade)
{
	//now you can access many different mapping types of
	// one surface through shade's currrent object
	Mapping* mapping = shade.obj->GetMapping();

	if (mapping)
		return mapping->Map3DTo2D(shade.obj, BumpMapReference::PreprocessVector(shade));
	else
		return BumpMapReference::PreprocessVector(shade);
}

BumpMap*
ImageBumpMap::Dup() const
{
	return new ImageBumpMap(*this);
}

void
ImageBumpMap::SetImageFile(char* filename)
{
	if (file != NULL) {
		//don't do anything if the name is of the same file
		if (!strcmp(filename, file->GetFileName()))
			return;
		delete file;
	}

	file = new ImageFile(FALSE);	//say no to the Open File dialog
	file->OpenFile(filename);		//open the file quietly

}

void
ImageBumpMap::SetImageFile(ImageFile* newFile)
{
	//if it's the same file return
	if (newFile == file)
		return;

	if (file != NULL)
		delete file;

	file = newFile;
}


void
ImageBumpMap::PerturbNormal(ShadingInfo& shade)
{
	
	Vector3D vec = PreprocessVector(shade); //map the point 

	//rotation of the bitmap (same as ImageMap)
	if (rot_angle != 0.0)
	{
		vec = TranslationMatrix(Vector3D(-0.5, -0.5, 0)) * vec; 
		vec = RotationZMatrix(rot_angle) * vec;	//in radians
		vec = TranslationMatrix(Vector3D(0.5, 0.5, 0)) * vec;
	}
	
	//apply scaling of u and v
	vec *= scale;

	//set map height and width to image dimensions -2 because we need -1 counting 
	// from 0, minus another 1 because we will be accessing GetRGBColorPixel(ulx+1, uly)
	int map_width = file->Width() - 2;
	int map_height = file->Height() - 2;
	
	vec *= Vector3D(float(map_width), float(map_height), 0);	//get actual image coordinates
    
    //if the image is tiled then repeat the image
    if (! once) {
	vec.x = (float)fmod(vec.x, map_width);
	if (vec.x < 0)
	    vec.x += map_width;
	vec.y = (float)fmod(vec.y, map_height);
	if (vec.y < 0)
	    vec.y += map_height;
    }
	
	long ulx = (long) vec.x;
    long uly = (long) vec.y;
    
	shade.obj->GetDerivatives(shade); //fill in shade's dudt and dvdt

	RGBColor colors[3];
	int xp, yp;
	Vector3D v;
	for (int i = 0; i < 3; i++) {
		v = vec;
		if (i == 0) {
			xp = ulx+1; yp = uly;
			v.x += 1;
		}
		if (i == 1) {
			xp = ulx; yp = uly;
		}
		if (i == 2) {
			xp = ulx; yp = uly+1;
			v.y += 1;
		}

		float horzweight = v.x - xp;
		float vertweight = v.y - yp;
		float ul = (1 - horzweight) * (1 - vertweight);
		float ur = horzweight * (1 - vertweight);
		float ll = (1 - horzweight) * vertweight;
		float lr = horzweight * vertweight;
		colors[i] = (ul * file->GetRGBColorPixel(xp, yp) +
	       ur * file->GetRGBColorPixel(xp+1, yp) +
	       ll * file->GetRGBColorPixel(xp, yp+1) +
	       lr * file->GetRGBColorPixel(xp+1, yp+1));
			
	}
	//compute the gradients in u and v directions of the image
	RGBColor du = colors[0] - colors[1];
	RGBColor dv = colors[2] - colors[1];

	//compute the gradients in u and v directions of the image
	//RGBColor du = file->GetRGBColorPixel(ulx+1, uly) - file->GetRGBColorPixel(ulx, uly);
	//RGBColor dv = file->GetRGBColorPixel(ulx, uly+1) - file->GetRGBColorPixel(ulx, uly);
	
	//compute the partial derivatives after applying the gradients
//	shade.dudt = shade.dudt + shade.normal * (bump_amt * GetGray(du));
//	shade.dvdt = shade.dvdt + shade.normal * (bump_amt * GetGray(dv));
	shade.normal = Normalize(shade.normal + (shade.dudt * (bump_amt * du)) + (shade.dvdt * (bump_amt * dv)));

	//get the new normal
//	shade.normal = Normalize(CrossProd(shade.dudt, shade.dvdt));
	
	//reverse the normal, if necessary, so it faces
	//	toward the incident direction.
	if (DotProd(shade.incident, shade.normal) > 0)
	shade.normal = -shade.normal;
}
