/****************************************************************************
 *	
 * 	C3DStudio.cpp 				3/9/2000
 *	By Konstantin Khanyants 	kon@unix.temple.edu
 *	
 *	Description: Class functions that are used in reading a 3DS file.
 *
 ***************************************************************************/

#include "stdafx.h"
#include "include/Oort.h"
#include <string.h>

//Constructors
C3DStudio::C3DStudio()
{
	m_Scale = 1.0;
	m_Version = 3;
	m_BackColor.r = 0;
	m_BackColor.g = 0;
	m_BackColor.b = 0;
							
	m_AmbientLight.r = 0;
	m_AmbientLight.g = 0;
	m_AmbientLight.b = 0;
}

C3DStudio::C3DStudio(CString filename)
{
	m_Scale = 1.0;
	m_Version = 3;
	m_BackColor.r = 0;
	m_BackColor.g = 0;
	m_BackColor.b = 0;
							
	m_AmbientLight.r = 0;
	m_AmbientLight.g = 0;
	m_AmbientLight.b = 0;
	Read3DSFile(filename);
}

//Destructor
C3DStudio::~C3DStudio()
{
	DestroyDatabase();
}

//this function deallocates the entire 3ds file database
//this is used when a new file is chosen or on Application's termination
void
C3DStudio::DestroyDatabase()
{
	//delete the materials in the database
	while (m_MaterialList.ListCount()) {
		material3ds* mat = m_MaterialList.ExtractHead();
		
		//delete the material's name
		if (mat->name != NULL)
			delete [] mat->name;

		//delete the texture map
		if (mat->texture.map != NULL) {
			if (mat->texture.map->name != NULL)
				delete [] mat->texture.map->name;
			
			delete mat->texture.map;
		}

		//delete the bump map
		if (mat->bump.map != NULL) {
			if (mat->bump.map->name != NULL)
				delete [] mat->bump.map->name;
			
			delete mat->bump.map;
		}
	
	}

	//delete the meshes in the database
	while (m_MeshList.ListCount()) {
		mesh3ds* mesh = m_MeshList.ExtractHead();
		
		if (mesh->name != NULL)
			delete [] mesh->name;
		
		//vertices deletion
		if (mesh->vertexarray != NULL)
			delete [] mesh->vertexarray;
		
		//texture vertices deletion
		if (mesh->textarray != NULL)
			delete [] mesh->textarray;
		
		//faces deletion
		if (mesh->facearray != NULL)
			delete [] mesh->facearray;
		
		//smooth indices deletion
		if (mesh->smootharray != NULL)
			delete [] mesh->smootharray;

		//delete the material groups
		if (mesh->nmats > 0) {
			for (ushort3ds i = 0; i < mesh->nmats; i++) {
				delete [] mesh->matarray[i].name;
				delete [] mesh->matarray[i].faceindex;
			}

			//delete the groups array
			delete [] mesh->matarray;
		}
	}

	//delete the lights from the database
	while (m_LightList.ListCount()) {
		light3ds* light = m_LightList.ExtractHead();
		
		if (light->name != NULL)
			delete light->name;
		
		//delete the spotlight property if one exists
		if (light->spot != NULL)
			delete light->spot;
		
		delete light;	
	}

	//delete the cameras from the database
	while (m_CameraList.ListCount()) {
		camera3ds* cam = m_CameraList.ExtractHead();
		
		if (cam->name != NULL)
			delete cam->name;
		
		delete cam;
	}

	//reset local settings
	m_Scale = 1.0;
	m_Version = 3;
	m_BackColor.r = 0;
	m_BackColor.g = 0;
	m_BackColor.b = 0;
	m_AmbientLight.r = 0;
	m_AmbientLight.g = 0;
	m_AmbientLight.b = 0;

}

//Passes back the filename of the 3D Studio file
CString
C3DStudio::GetFilename()
{
	return m_Filename;
}

//Initializes the camera list with the camera info
//obtained from the 3DS database. Camera list must be empty
// prior to calling this function
void
C3DStudio::GetCameraList(CameraList& cameras)
{
	camera3ds	*cam3ds;
	Camera		*currCamera;

	//guarantee at least one camera
	if (m_CameraList.ListCount() == 0) {
		//get space for new camera
		currCamera = new Camera;
	
		currCamera->m_Name = "Default";
		currCamera->m_Pos = Vector3D(0, 0, 50);
		currCamera->m_Target = Vector3D(0, 0, -10);
		currCamera->m_Up = Vector3D(0, 0, 1);

		currCamera->m_Focal_length = 70;
		currCamera->m_Range_far = 1000;  //Depth Of View
		currCamera->m_Range_near = 0.01f;
		currCamera->m_Roll = 0.0;
		currCamera->m_Fov = 45.0;  //in degrees
		
		cameras.AppendToList(currCamera);
	}
	else {
		//now fill in the appropriate camera structures
		for (SimpleList<camera3ds *>::Iterator cam(m_CameraList); cam.Valid(); cam.GotoNext()) {
			//make a database of cameras
			cam3ds = cam.Contents();

			currCamera = new Camera(*cam3ds);
			cameras.AppendToList(currCamera);
		}
	}	//end else if there are cameras in the scene
}

//Obtains the light list from the 3DS database
//lights list must be empty upon entry
void
C3DStudio::GetLightList(LightList& lights)
{
	light3ds  *currLight;
	Light	  *theLight; 

	//guarantee two point lights if all are absent
	if (m_LightList.ListCount() == 0) {

		Vector3D Light1(-3000.0f, 3000.0f, 3000.0f);
		Vector3D Light2(5000.0f, 10000.0f, 5000.0f);
//		Vector3D Light1(-30.0f, 30.0f, 30.0f);
//		Vector3D Light2(50.0f, 10.0f, 50.0f);
		RGBColor color(0.7f, 0.7f, 0.7f);

		theLight = new PointLight(Light1, color, 1, 0, 0);
		theLight->SetRange(10000, 12000, FALSE); //no attenuation

		theLight->SetName("Default 1");
		lights.AppendToList(theLight);

		theLight = new PointLight(Light2, color, 1, 0, 0);
		theLight->SetRange(10000, 12000, FALSE); //no attenuation

		theLight->SetName("Default 2");
		lights.AppendToList(theLight);

	}
	else {

	//now fill in the appropriate light structures
	for (SimpleList<light3ds *>::Iterator lite(m_LightList); lite.Valid(); lite.GotoNext()) {
		//make a database of lights
        currLight = lite.Contents();

		if (currLight->spot == NULL)
			theLight = new PointLight(Vector3D(currLight->pos), RGBColor(currLight->color), 1, 0, 0, !(currLight->dloff));
		else
			theLight = new SpotLight(Vector3D(currLight->pos), Vector3D(currLight->spot->target),
				currLight->spot->hotspot, currLight->spot->falloff, RGBColor(currLight->color), 1, 0, 0, !(currLight->dloff));

		theLight->SetRange(currLight->attenuation.inner, currLight->attenuation.outer, 
			currLight->attenuation.on);

		theLight->SetName(currLight->name);
		
		lights.AppendToList(theLight);

	} //end for all lights in the database
	} //end else -if there are lights in the database
}


//Gets the surface list from the database. Surfaces
// list must be empty before this function is called
void
C3DStudio::GetSurfaceList(SurfaceList& surfaces)
{
	material3ds *currMat;
	HallSurface	*currSurface;
	bitmap3ds	*bitmap;
	int			once;

	//at least one material is guaranteed
	if (m_MaterialList.ListCount() == 0) {
		currSurface = new HallSurface;
		currSurface->SetAmbient(RGBColor(0.1f));
		currSurface->SetDiffuseColor(RGBColor(0.5));
        currSurface->SetSpecular(RGBColor(1.0), 128.0);
		currSurface->SetShinStrength(.5);
		currSurface->SetTransmitted(RGBColor(0), 1.0);

		//OpenGL only
		currSurface->SetPolygonMode(GL_FILL);	     // GL_LINE for wire frame, GL_FILL for all other types
		currSurface->SetShadeModel(GL_SMOOTH);       // GL_FLAT for wire frame and flat shading type, GL_SMOOTH for all other types
		currSurface->SetTwoSided(FALSE);

		//name the material				
		currSurface->SetName("Default");
		surfaces.AppendToList(currSurface);  //put the material in the list

	}
	else {
	//now fill in the appropriate surface structures
	for (SimpleList<material3ds *>::Iterator mat(m_MaterialList); mat.Valid(); mat.GotoNext()) {
		//make a database of sufraces
        currMat = mat.Contents();

		currSurface = new HallSurface;
		currSurface->SetAmbient(RGBColor(currMat->ambient));
		currSurface->SetDiffuseColor(RGBColor(currMat->diffuse));

		float spec_exponent = float(currMat->shininess * 300.0);
		currSurface->SetSpecular(RGBColor(currMat->specular), spec_exponent);
		currSurface->SetShinStrength(currMat->shinstrength);
		currSurface->SetTransmitted(RGBColor(currMat->transparency), 1.0); //and index of refraction
	
		//OpenGL only
		currSurface->SetTwoSided(currMat->twosided);

		// shading must be expressed in several OpenGL flags
		// Wire frame or solid shading <Wire Frame is not supported by RayTracer>
		if (currMat->shading > Wire)
			currSurface->SetPolygonMode(GL_FILL);
		else
			currSurface->SetPolygonMode(GL_LINE);

        // Gouraud, Phong and Metal shading are all expressed as Gouraud shading since this is
        // the only smoothing type OpenGL can render
        if (currMat->shading > Flat)
			currSurface->SetShadeModel(GL_SMOOTH);
		else
			currSurface->SetShadeModel(GL_FLAT);

		CString completeName;
		CString pcsMsg;

		//if the texture map exists
		if (currMat->texture.map != NULL) {
			bitmap = currMat->texture.map;

			once = TRUE;
			if (bitmap->tiling == Tile)
				once = FALSE;
						
			//Open GL part
			// also this is just a guess, since I haven't found any description when to use
			// decal and when modulated texturing (might well be that we even need GL_BLEND
			// here to implement the texture strength value)
			int texModeGL;
			if (spec_exponent == 0)
				texModeGL = GL_DECAL;
			else
				texModeGL = GL_MODULATE;

			if (FindImageFile(bitmap->name, completeName)) {
			
				//also convert rotation angle to radians 			//interpolation = FALSE
				ImageMap *texture = new ImageMap(IdentityMatrix(), TRUE, once, 
						 LPCTSTR(completeName), float(bitmap->rotation * M_PI / 180), bitmap->uscale, bitmap->vscale,
						 GL_NEAREST, GL_NEAREST, GL_REPEAT, GL_REPEAT, texModeGL, bitmap->percent);

				currSurface->SetDiffuseTexture(texture);
				currSurface->SetTextureFlag(TRUE);
			}
			else {
				pcsMsg.Format("Can't find texture file %s. File not loaded.", bitmap->name);
				AfxMessageBox(pcsMsg);
			}
		}

		if (currMat->bump.map != NULL) {
			bitmap = currMat->bump.map;
			
			once = TRUE;
			if (bitmap->tiling == Tile)
				once = FALSE;

			completeName = "";
			if (FindImageFile(bitmap->name, completeName)) {

				//also convert rotation angle to radians 
				ImageBumpMap *bumpMap = new ImageBumpMap(IdentityMatrix(), once, LPCTSTR(completeName), 
					float(bitmap->rotation * M_PI / 180), bitmap->percent, bitmap->uscale, bitmap->vscale); 

				currSurface->SetBumpMap(bumpMap);
				currSurface->SetBumpFlag(TRUE);
			}
			else {
				pcsMsg.Format("Can't find texture file %s. File not loaded.", bitmap->name);
				AfxMessageBox(pcsMsg);
			}
		}

		//name the material
		CString newNameStr(currMat->name);
		int matNum = 0;
		CString numExt = "";

		//see if the material with that name already exists
		while (surfaces.FindSurfaceByName(newNameStr + numExt) != NULL)
			numExt.Format(" #%d", matNum++);

		currSurface->SetName(newNameStr);
		surfaces.AppendToList(currSurface);  //put the material in the list

	} //end for all the materials in the database
	} //end else -if there are materials in the database

}

//This function searches the directory of the .3ds file for
// the specified imageName, it also searches the Maps and Textures 
// subdirectories if they exist
BOOL
C3DStudio::FindImageFile(CString imageName, CString& completeName)
{
	CFileFind search;
	
	//if they gave us a file name, find the file, get the extention
	if (imageName.GetLength() > 0) {

		//find the start of the 3DS filename
		int pos = m_Filename.ReverseFind('\\');

		if (pos >= 0) {
			int nameLen = pos + 1;
			//assign the complete path name of the image file
			CString pathName = m_Filename.Left(nameLen);

			//default directory of the 3ds file
			completeName = pathName + imageName;
			if (search.FindFile(completeName))
				return TRUE;

			//Maps subdirectory search
			completeName = pathName + "Maps\\" + imageName;
			if (search.FindFile(completeName))
				return TRUE;

			//Textures subdirectory search
			completeName = pathName + "Textures\\" + imageName;
			if (search.FindFile(completeName))
				return TRUE;
			
			//if the file is not found anywhere
			return FALSE;
		}
		else 
			return FALSE;
	}	//null string
	else
		return FALSE;
}

