/****************************************************************************
 *	
 * 	Camera.cpp					3/10/2000
 *	By Konstantin Khanyants 	kon@unix.temple.edu
 *	
 *	Description: A camera class based on the camera model in
 *				 3D Studio.
 * 
 *	Camera structures definition as in 3DSFile.h:
 *	typedef struct {
 *		float3ds cam_near; 
 *		float3ds cam_far;
 *	} camranges3ds;
 *
 *	typedef struct {
 *		char3ds *name;
 *		point3ds position;
 *		point3ds target;
 *		float3ds roll;
 *		float3ds focallength
 *		float3ds fov;
 *		camranges3ds ranges;
 *	} camera3ds;
 *
 *
 ***************************************************************************/

#include "stdafx.h"
#include "include/Oort.h"

Camera::Camera()
{
	m_Pos = Vector3D(0.0);
	m_Target = Vector3D(0.0);
	m_Up = Vector3D(0.0);
	m_Roll = 0.0;
	m_Fov = 0.0;
	m_Range_near = 0;
	m_Range_far = 0;
	//set the scene matrix to identity
	m_SceneMatrix = IdentityMatrix();
}

Camera::Camera(camera3ds& cam3ds)
{
	SetCamera3DS(cam3ds);
	m_SceneMatrix = IdentityMatrix();
}

void
Camera::Describe()
{
	CString pcsName = "Camera Name = " + m_Name;
	CString content;
	content.Format("\nFocal Length = %1.2f\nFOV = %1.2f\nNear Range = %1.2f\nFar Range = %1.2f\nPosition = {%1.2f, %1.2f, %1.2f}\nTarget = {%1.2f, %1.2f, %1.2f}",
		m_Focal_length, m_Fov, m_Range_near, m_Range_far, m_Pos.x, m_Pos.y, m_Pos.z, m_Target.x, m_Target.y, m_Target.z);

	AfxMessageBox(pcsName + content);
}

void
Camera::ApplyPerspectiveOpenGL(int width, int height)
{
	gluPerspective(m_Fov, (GLdouble) width / (GLdouble) height, m_Range_near, m_Range_far);
}

void
Camera::ApplyTransformToSceneGL(Vector3D& sceneCenter, BOOL onAuto)
{
	// initialize transformation pipeline
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Vector3D newTarget = m_Target;
	if (onAuto)
		newTarget = sceneCenter;

//	Matrix temp = ViewMatrix(newTarget, m_Pos, m_Up) * m_SceneMatrix;
//	glMultMatrixf((float*)temp.x);
	gluLookAt(m_Pos.x,		m_Pos.y,	 m_Pos.z,
			  newTarget.x,	newTarget.y, newTarget.z,
			  m_Up.x,		m_Up.y,		 m_Up.z);

	glMultMatrixf((float*)m_SceneMatrix.x);
}

void	 
Camera::SetCamera3DS(camera3ds& cam3ds)
{
	m_Name = cam3ds.name;
	
	m_Roll = cam3ds.roll; //in degrees
	m_Focal_length = cam3ds.focallength;
	m_Fov = cam3ds.fov;	//in degrees
	m_Range_near = cam3ds.ranges.cam_near;
	m_Range_far = cam3ds.ranges.cam_far;

	if (m_Range_near < 0.0)
		m_Range_near = -m_Range_near;

	//an OpenGL error occurs if near plane is 0
	if (m_Range_near == 0.0)
		m_Range_near += 0.01f;

	if (m_Range_far < 0.0)
		m_Range_far = -m_Range_far;

	if (m_Range_far < 300)
		m_Range_far = 300;

	if (m_Range_near > m_Range_far) {
		float oldFar = m_Range_far;
		m_Range_far = m_Range_near;
		m_Range_near = oldFar;
	}

	m_Pos = Vector3D(cam3ds.position);
	m_Target = Vector3D(cam3ds.target);
	
	m_Up = Vector3D(0.0, 0.0, 1.0);	//initial upVector is Z in 3DStudio

	//convert m_Roll to radians
	float roll_angle = float(m_Roll * M_PI / 180);

	//rotate the camera according to roll the camera
	Matrix roll = RotationAxisMatrix(Normalize(m_Pos - m_Target), roll_angle);
	m_Up = roll * m_Up;

}

//used to rotate the position of the camera around the target
void
Camera::RotatePos(Matrix& rot_matrix)
{
//	m_Pos = (rot_matrix * Vector3D(m_Pos - m_Target)) + m_Target;
	//rotate the scene instead of camera
	m_SceneMatrix = m_SceneMatrix * rot_matrix;
}


CString
Camera::GetName()
{
	return m_Name;
}

Camera*
CameraList::FindCameraByName(CString camName)
{
	for (Iterator sc(*this); sc.Valid(); sc.GotoNext()) {
	if (sc.Contents()->m_Name == camName)
	    return sc.Contents();
    }
    return NULL;
}


//added 6/12/2000 for accessing objects by pointer and not by
//pointer to pointer
Camera*
CameraList::operator[] (int inx) 
{
//	ModifyCheck();
    if (inx >= n || inx < 0)
        return NULL;
    SimpleListNode *pNode = head;

	if (head) {
		int i = 0;
		while (i != inx) {
    	 	pNode = pNode->next;
			i++;
		}
  	  	return pNode->contents;
	}
	return NULL;
}

//delete all the cameras
void
CameraList::DeleteCameras()
{
	while (n > 0)
		delete (ExtractHead());

}


