// ===================================================================
// 	File: light.cpp					12/26/1999
//	Implementation file for various light sources.
//
//	     The Object-Oriented Ray Tracer (OORT)
//            Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"
#include "include/World.h"
#include <math.h>

// -------------------------------------------------------------------
// Light
//	This class encapsulates the various types of light source
//	available in OORT.  Shadow rays are cast at light sources;
//	they emit light of a given color, which is attenuated by
//	distance.
// -------------------------------------------------------------------

Light::Light(RGBColor& clr, float C1, float C2, float C3, BOOL on)
{
    Color = clr;
    c1 = C1;
    c2 = C2;
    c3 = C3;
    LastOpaque = NULL;
	enabled = on;
}

void 
Light::SetName(CString newName)
{
	name = newName;
}

// -------------------------------------------------------------------
// Light::DistAtten.
//	Computes the attenuation due to distance of the light source,
//	clamped to 1.0.
// -------------------------------------------------------------------
float
Light::DistAtten(float t)
{
    float ret = 1 / (c1 + t * (c2 + c3 * t));
    return (ret < 1) ? ret : 1;
}

// -------------------------------------------------------------------
// Check for a cached object and, if there is one, check if it is
// intersected by the given ray within the given maximum parameter of
// intersection.  If so, return 1; otherwise clear the cached object
// and return 0.
// -------------------------------------------------------------------
int
Light::HitCachedObject(Ray& ray, float maxt)
{
	if (LastOpaque) {

		Statistics::ShadowTotal++;

		if (LastOpaque->HitAtAll(ray, maxt)) {
			Statistics::ShadowHits++;
			return 1;
		}
		else {
			LastOpaque = NULL;
			return 0;
		}
	}
    return 0;
}

RGBColor
Light::ShadowRay(ShadingInfo& shade, 
		 Vector3D& target,
		 Surface *surf)
{
    RGBColor ret(0);

    Vector3D RayDirection = target - shade.p;	// Find direction of ray
    float dist = Magnitude(RayDirection);	// Find distance to light source
    RayDirection /= dist;			// Normalize ray direction
    Ray ShadowRay(shade.p, RayDirection);

    // If hit the cached opaque object, return black.
    if (HitCachedObject(ShadowRay, dist))
	return RGBColor(0);

    // We're going to have to fire a shadow ray.
    Statistics::ShadowRays++;

    RGBColor Atten(1, 1, 1);
    Object3D *obj = shade.world->ShadowAtten(ShadowRay, Atten, dist);
    if (obj) {
	LastOpaque = obj;
	return RGBColor(0);
    }
    return surf->ShadeLight(shade, RayDirection) *
    	   Atten * Color * DistAtten(dist);
}


// -------------------------------------------------------------------
// PointLight
//	This class encapsulates the ubiquitous point light sources in
//	ray tracing.  Only one shadow ray is cast at point light sources,
//	even if distribution is requested by the user.
// -------------------------------------------------------------------

PointLight::PointLight(Vector3D& loc,
		       RGBColor& clr,
		       float C1, float C2, float C3, BOOL on) : 
    Light(clr, C1, C2, C3, on)
{
    Location = loc;
	Near_range = 0;
	Far_range = 0;
	Attn_on = false;
	type = POINT_LIGHT;

	lightGL = 0;	//default is zero because I dont want to disable
					//any important GL stuff
	locationGL = MakeHVector(Location.x, Location.y, Location.z, 1);
}

//ShadowLight///////////3/14/2000
// -------------------------------------------------------------------
// Find the color of the light source as seen from the given
// intersection point, and apply light source attenuation.
// -------------------------------------------------------------------
RGBColor
PointLight::ShadowLight(ShadingInfo& shade, Surface *surf)
{
	Vector3D RayDirection = Location - shade.p;	// Find direction of ray
    float dist = Magnitude(RayDirection);	// Find distance to light source
	float atten = 1.0;

	//if the light is off
	if (!enabled) return RGBColor(0);

	//if attenuation is turned on
	if (Attn_on) {
	//if the point is not within the range of the light then
	// the contribution is zero
	if (dist > Far_range)
		return RGBColor(0);
	//otherwise if within range, but is in the falloff
	//then calculate the linear attenuation
	else if (dist <= Far_range && dist > Near_range) //attenuate
		atten = (Far_range - dist) / (Far_range - Near_range);
	}

    return ShadowRay(shade, Location, surf) * atten;
}

void
PointLight::SetRange(float range_near, float range_far, BOOL on)
{
	Near_range = range_near;
	Far_range = range_far;
	Attn_on = on;
}


void
PointLight::ApplyTransform(const Matrix& tform)
{
    Location = tform * Location;
}

void
PointLight::Describe()
{
	CString nameMsg = "Point Light\nName = " + name;
	CString pcsMsg;
	pcsMsg.Format("\nEnabled = %d\n Color = {%1.2f, %1.2f, %1.2f}\nLocation = {%1.2f, %1.2f, %1.2f}\nNear Range = %1.2f\nFar Range = %1.2f",
		enabled, Color[0], Color[1], Color[2], locationGL[0], locationGL[1], locationGL[2], Near_range, Far_range);
	nameMsg += pcsMsg;
	AfxMessageBox(nameMsg);
}

void
PointLight::DisableLightGL()
{
	if (enabled && lightGL != 0 && glIsEnabled(lightGL))
		glDisable(lightGL);

	lightGL = 0;
}

void
PointLight::SetupLightGL()
{
	GLint MaxLights;

	//disable this light if it is enabled
	DisableLightGL();

	//if the light is off, don't go any further
	if (!enabled)
		return;

	// determine maximum number of supported light sources (8 are guaranteed)
	glGetIntegerv(GL_MAX_LIGHTS, &MaxLights);

	//eight lights are guaranteed by OpenGL, try to get more if possible
	int i = 0;
	while (glIsEnabled(GL_LIGHT0 + i) == GL_TRUE && i < MaxLights) i++;

	//assign the member variable for OpenGL usage if the
	// light it indexes is on
	if (i == MaxLights) {
		//if there is no more room then light is not used
		lightGL = 0;
		return;
	}
	else
		lightGL = GL_LIGHT0 + i;

	glEnable(lightGL);

	glLightfv(lightGL, GL_POSITION, locationGL);

	//distribute the color of the light source among the various components
	glLightfv(lightGL, GL_AMBIENT, MakeHVector(Color.x / 5, Color.y / 5, Color.z / 5, 1));
	glLightfv(lightGL, GL_DIFFUSE, MakeHVector(Color.x, Color.y, Color.z, 1));
	glLightfv(lightGL, GL_SPECULAR, MakeHVector(Color.x, Color.y, Color.z, 1));

	// switch on point light property by setting the spot cut off to 180
	glLightf(lightGL, GL_SPOT_CUTOFF, 180.0);
}

// -------------------------------------------------------------------
// SpotLight
//	This class encapsulates the directional spot light source.
//	only one shadow ray is cast at
//	spot light sources, even if distribution is requested
//	by the user.
// -------------------------------------------------------------------

// --------------------------
// Constructor
// --------------------------
SpotLight::SpotLight(Vector3D& loc, Vector3D& target,
		float& hotspot, float& falloff, RGBColor& clr,	
		float C1, float C2, float C3, BOOL on) :
	PointLight(loc, clr, C1, C2, C3, on)
{
	Target = target;
	//convert to radians
	Hotspot = hotspot * M_PI / 180;
	Falloff = falloff * M_PI / 180;
	type = SPOT_LIGHT;

	lightGL = 0;
	targetGL = MakeHVector(Target.x, Target.y, Target.z, 1);
}

void
SpotLight::Describe()
{
	CString nameMsg = "Spot Light\nName = " + name;
	CString pcsMsg;
	pcsMsg.Format("\nEnabled = %d\n Color = {%1.2f, %1.2f, %1.2f}\nLocation = {%1.2f, %1.2f, %1.2f}\nTarget = {%1.2f, %1.2f, %1.2f}\nNear Range = %1.2f\nFar Range = %1.2f\nHotspot = %1.2f\nFalloff = %1.2f",
		enabled, Color[0], Color[1], Color[2], locationGL[0], locationGL[1], locationGL[2], targetGL[0], targetGL[1], targetGL[2], Near_range, Far_range, Hotspot, Falloff);
	nameMsg += pcsMsg;
	AfxMessageBox(nameMsg);
}


void
SpotLight::ApplyTransform(const Matrix& tform)
{
    Location = tform * Location;
	Target = tform * Target;
}

//ShadowLight///////////3/14/2000
// -------------------------------------------------------------------
// Find the color of the light source as seen from the given
// intersection point, and apply light source attenuation.
// -------------------------------------------------------------------
RGBColor
SpotLight::ShadowLight(ShadingInfo& shade, Surface *surf)
{
	Vector3D RayDirection = Location - shade.p;	// Find direction of ray
    float dist = Magnitude(RayDirection);	// Find distance to light source
	float atten = 1.0; //maximum attnenuation

	//if the light is off
	if (!enabled) return RGBColor(0);

	//if attenuation is turned on
	if (Attn_on) {
	//if the point is not within the range of the light then
	// the contribution is zero
	if (dist > Far_range)
		return RGBColor(0);
	//otherwise if within range, but is in the falloff
	// then calculate the linear attenuation
	else if (dist <= Far_range && dist > Near_range) //attenuate
		atten = (Far_range - dist) / (Far_range - Near_range);
	}

	//find the direction of the spot light
	Vector3D LightDirection = Location - Target;
	//find the angle between the two direction vectors
	RayDirection = Normalize(RayDirection);
	LightDirection = Normalize(LightDirection);
	float alpha = (float) acos(DotProd(RayDirection, LightDirection));

	//if the angle is larger than the cutoff angle then the
	// contribution is zero
	if (alpha > Falloff)
		return RGBColor(0);
	//otherwise if within the attenuation range then attenuate
	else if (alpha <= Falloff && alpha > Hotspot)
		atten *= (Falloff - alpha) / (Falloff - Hotspot);

    return ShadowRay(shade, Location, surf) * atten;
}


void
SpotLight::SetupLightGL()
{
	GLint MaxLights;

	//disable this light if it is enabled
	DisableLightGL();

	//if the light is off, don't go any further
	if (!enabled)
		return;

	// determine maximum number of supported light sources (8 are guaranteed)
	glGetIntegerv(GL_MAX_LIGHTS, &MaxLights);

	//eight lights are guaranteed by OpenGL, try to get more if possible
	int i = 0;
	while (glIsEnabled(GL_LIGHT0 + i) == GL_TRUE && i < MaxLights) i++;

	//assign the member variable for OpenGL usage if the
	// light it indexes is on
	if (i == MaxLights) {
		//if there is no more room then light is not used
		lightGL = 0;
		return;
	}
	else
		lightGL = GL_LIGHT0 + i;

	glEnable(lightGL);

	glLightfv(lightGL, GL_POSITION, locationGL);

	//distribute the color of the light source among the various components
	glLightfv(lightGL, GL_AMBIENT, MakeHVector(Color.x / 5, Color.y / 5, Color.z / 5, 1));
	glLightfv(lightGL, GL_DIFFUSE, MakeHVector(Color.x, Color.y, Color.z, 1));
	glLightfv(lightGL, GL_SPECULAR, MakeHVector(Color.x, Color.y, Color.z, 1));

	//settings specific to a spot light
	glLightfv(lightGL, GL_SPOT_DIRECTION, targetGL);
	//SpotExponent = 0  can the HotSpot angle be used here? 
	glLightf(lightGL, GL_SPOT_EXPONENT, 0.0);
	//take the largest of the two angles (hot spot, falloff)
	glLightf(lightGL, GL_SPOT_CUTOFF, Falloff * 180 / M_PI);
}

// -------------------------------------------------------------------
// DirectLight
//	This class encapsulates the directional light source.
//	They are located at infinity; only one shadow ray is cast at
//	directional light sources, even if distribution is requested
//	by the user.
// -------------------------------------------------------------------

// --------------------------
// Constructor
// --------------------------
DirectLight::DirectLight(const Vector3D& dir, RGBColor& clr, BOOL on) :
    Light(clr, 0, 0, 0, on), Direction(Normalize(dir)) 
{
	type = DIRECTIONAL_LIGHT;	
}

// -------------------------------------
// Find the color of the light source as
// seen from the given intersection 
// point.
// -------------------------------------

void
DirectLight::Describe()
{
	CString nameMsg = "Directional Light\nName = " + name;
	CString pcsMsg;
	pcsMsg.Format("\nEnabled = %d\n Color = {%1.2f, %1.2f, %1.2f}\nDirection = {%1.2f, %1.2f, %1.2f}",
		enabled, Color[0], Color[1], Color[2], Direction[0], Direction[1], Direction[2]);
	nameMsg += pcsMsg;
	AfxMessageBox(nameMsg);
}

RGBColor
DirectLight::ShadowLight(ShadingInfo& shade, Surface *surf)
{
    Ray ShadowRay(shade.p, Direction);

	//if the light is off
	if (!enabled) return RGBColor(0);

    // If hit the cached opaque object, return black.
    if (HitCachedObject(ShadowRay, Limits::Infinity))
	return RGBColor(0);

    // We're going to have to fire a shadow ray.
    Statistics::ShadowRays++;

    RGBColor Atten(1, 1, 1);
    Object3D *obj = shade.world->ShadowAtten(ShadowRay, Atten, Limits::Infinity);
    if (obj) {
	LastOpaque = obj;
	return RGBColor(0);
    }
    return surf->ShadeLight(shade, Direction) * Atten * Color;
}

void
DirectLight::ApplyTransform(const Matrix& tform)
{
    Direction = RotateOnly(tform, Direction);
}


//added 6/12/2000 for accessing objects by pointer and not by
//pointer to pointer
Light*
LightList::operator[] (int inx) 
{
//	ModifyCheck();
    if (inx >= n || inx < 0)
        return 0;
    SimpleListNode *pNode = head;
	if (head) {
		int i = 0;
		while (i != inx) {
    	 	pNode = pNode->next;
			i++;
		}
  	  	return pNode->contents;
	}
	return 0;
}

//delete the lights
void
LightList::DeleteLights()
{
	while (n > 0)
		delete (ExtractHead());
}


Light*
LightList::FindLightByName(CString lightName)
{
	for (Iterator sc(*this); sc.Valid(); sc.GotoNext()) {
	if (sc.Contents()->name == lightName)
	    return sc.Contents();
    }
    return NULL;
}

void
LightList::SetupLightsGL()
{	
	//first disable all the lights
	for (Iterator lc(*this); lc.Valid(); lc.GotoNext())
		lc.Contents()->DisableLightGL();
	
	//setup the lights
	for (Iterator lc2(*this); lc2.Valid(); lc2.GotoNext())
		lc2.Contents()->SetupLightGL();
}
