// ===================================================================
//  File: Mapping.cpp						12/26/1999
//	By Konstantin Khanyants
//	Description:
//		Implementations of inverse mappings. Included are 
//		Cylindrical, Quadrilateral, Spherical, UV vertex, and
//		Projection mappings.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"

// ---------------------------------------------------------
// CylindricalMapping
//	Maps points onto a cylinder centered about the Y axis.
// ---------------------------------------------------------

CylindricalMapping::CylindricalMapping(float _radius, float _height)
{
    radius = _radius;
    height = _height;
}

Mapping *
CylindricalMapping::Dup() const
{
    return new CylindricalMapping(radius, height);
}

Vector3D
CylindricalMapping::Map3DTo2D(Object3D *obj, const Vector3D& v) const
{
    float u = float(acos(v.x / Magnitude(v)) / (2 * M_PI));
    return Vector3D( (v.y < 0) ? 1 - u : u,
		     posmod(v.y, height),
		     0);
/*
    float u = fmod(v.x, radius);
    u = acos(u / radius) / (2 * M_PI);
    return Vector( (v.y < 0) ? 1 - u : u,
		   posmod(v.y, height),
		   0);
*/
}

// ---------------------------------------------------------
// ProjectionMapping
//	Projects points onto the XY coordinate plane.
// ---------------------------------------------------------

Mapping *
ProjectionMapping::Dup() const
{
    return new ProjectionMapping;
}

Vector3D
ProjectionMapping::Map3DTo2D(Object3D *obj, const Vector3D& v) const
{
    return Vector3D(v.x, v.y, 0);
}

// ---------------------------------------------------------
// QuadrilateralMapping
//	Maps onto an arbitrary quadrilateral.
// ---------------------------------------------------------
QuadrilateralMapping::QuadrilateralMapping(const Vector3D *Pts)
{
    for (int i = 0; i < 4; i++)
	pts[i] = Pts[i];
    Vector3D Pa = pts[0] - pts[3] + pts[2] - pts[1];
    Vector3D Pb = pts[3] - pts[0];
    Vector3D Pc = pts[1] - pts[0];
    Vector3D Pd = pts[0];
    Vector3D Pn = -Normalize(CrossProd(pts[2] - pts[1],
				     pts[0] - pts[1]));
    Na = CrossProd(Pa, Pn);
    Nb = CrossProd(Pb, Pn);
    Nc = CrossProd(Pc, Pn);
    Nd = CrossProd(Pd, Pn);
    Du0 = DotProd(Nc, Pd);
    Du1 = DotProd(Na, Pd) + DotProd(Nc, Pb);
    Du2 = DotProd(Na, Pb);
    Dv0 = DotProd(Nb, Pd);
    Dv1 = DotProd(Na, Pd) + DotProd(Nb, Pc);
    Dv2 = DotProd(Na, Pc);
    uaxespar = fabs(Du2) < Limits::Small;
    if (! uaxespar) {
	Qux = Na / (2 * Du2);
	Dux = -Du1 / (2 * Du2);
	Quy = -Nc / Du2;
	Duy = Du0 / Du2;
    }
    vaxespar = fabs(Dv2) < Limits::Small;
    if (! vaxespar) {
	Qvx = Na / (2 * Dv2);
	Dvx = -Dv1 / (2 * Dv2);
	Qvy = -Nb / Dv2;
	Dvy = Dv0 / Dv2;
    }
}

Mapping *
QuadrilateralMapping::Dup() const
{
    return new QuadrilateralMapping(pts);
}

Vector3D
QuadrilateralMapping::Map3DTo2D(Object3D *obj, const Vector3D& x) const
{
    float u, v;
    if (uaxespar)
	u = (DotProd(Nc, x) - Du0) / (Du1 - DotProd(Na, x));
    else {
	float Ka = Dux + DotProd(Qux, x);
	float Kb = Duy + DotProd(Quy, x);
	if (Kb <= Ka*Ka) {
	    float u0 = Ka - (float) sqrt(Ka*Ka - Kb);
	    float u1 = Ka + (float) sqrt(Ka*Ka - Kb);
	    u = u0;
	    if (! InRange((float) 0, (float) 1, u0))
		u = u1;
	}
	else u = 0;
    }
    if (vaxespar)
	v = (DotProd(Nb, x) - Dv0) / (Dv1 - DotProd(Na, x));
    else {
	float Ka = Dvx + DotProd(Qvx, x);
	float Kb = Dvy + DotProd(Qvy, x);
	if (Kb <= Ka*Ka) {
	    float v0 = Ka - (float) sqrt(Ka*Ka - Kb);
	    float v1 = Ka + (float) sqrt(Ka*Ka - Kb);
	    v = v0;
	    if (! InRange((float) 0, (float) 1, v0))
		v = v1;
	}
	else v = 0;
    }
    return Vector3D(u, v, 0);
}

// ---------------------------------------------------------
// SphericalMapping
//	Maps onto a sphere centered about any point.
// ---------------------------------------------------------

SphericalMapping::SphericalMapping(Vector3D _center)
{
    center = _center;
}

Mapping *
SphericalMapping::Dup() const
{
    return new SphericalMapping(center);
}

Vector3D
SphericalMapping::Map3DTo2D(Object3D *obj, const Vector3D& v) const
{
    Vector3D unit = Normalize(v - center);
    float phi = (float) acos(unit.y);
    float theta = float(unit.x / sqrt(1 - unit.y * unit.y));
    if (fabs(theta) > 1)
	return Vector3D(0, float(phi / M_PI), 0);
    theta = float(acos(theta) / (2 * M_PI));
//  if (unit.z > 0)
//	return Vector3D(theta, float(phi / M_PI), 0);
//  else
	return Vector3D(1 - theta, float(phi / M_PI), 0);
}

// ---------------------------------------------------------------
// UV Mapping
//	Interpolates between three uv vertices of the triangle
// ---------------------------------------------------------------
Mapping*
UVMapping::Dup() const
{
	return new UVMapping;
}
	
Vector3D
UVMapping::Map3DTo2D(Object3D *obj, const Vector3D& vec) const
{
	Vector3D v(0); //initialize the vector
	
	if (obj->Flags() == TRIANGLE) {
		Triangle* tri = (Triangle *) obj;
		float alpha = tri->alpha;
		float beta	= tri->beta;
		float u0 = tri->Vertices[0].UV.x;
		float u1 = tri->Vertices[1].UV.x;
		float u2 = tri->Vertices[2].UV.x;
		float v0 = tri->Vertices[0].UV.y;
		float v1 = tri->Vertices[1].UV.y;
		float v2 = tri->Vertices[2].UV.y;
	
		//interpolate
		v.x = (1 - (alpha + beta)) * u0 + alpha * u1 + beta * u2;
		v.y = (1 - (alpha + beta)) * v0 + alpha * v1 + beta * v2; 
		 
	}
		
	return v;	
}