// ===================================================================
//  File: MeshListGL.cpp					5/1/2000
//	By Konstantin Khanyants
//	Description:
//		Class functions that are used when compiling OpenGL display
//		lists from vertex arrays that define the polygonal
//		objects imported from 3D Studio
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"
#include "include/World.h"


//----------------------------------------------------------------------------------------------------------------------

// Gets the oldest error from OpenGL engine and tries to clear the error queue.
// Because under some circumstances reading the error code creates a new error
// and thus hanging up the thread, we limit the loop to 6 reads.
void 
MeshListGL::CheckOpenGLError()
{
	GLenum GLError;
    GLint count;
	FILE *stream;	
	
	GLError = glGetError();
	if (GLError != GL_NO_ERROR) {
		stream = fopen("GLError.out", "w+");
		count = 0;
		//clear the queue
		while ((glGetError() != GL_NO_ERROR) && (count < 6)) 
		{count++; }
		fprintf(stream, "OpenGL Error: %s\n", gluErrorString(GLError));
		fclose(stream);
	}
}


//constructor for the list
MeshListGL::MeshListGL() : SimpleList<MeshGL *>()
{
	m_MainListGL = 0;
	m_Version = 0;
}


//actual drawing routine that executes the main display list
void
MeshListGL::DrawObjectsGL()
{
	if (m_MainListGL > 0)
		glCallList(m_MainListGL);	

}

//----------------------------------------------------------------------------------------------------------------------
//now in the World implelemtation
void
//World::Use3DSDataGL()
MeshListGL::BuildDisplayListGL()
{

// Notification method used by the main form to tell the viewer to upate its display.
// Previously loaded data will be freed and, new data will be loaded.
// Note: The rendering context must be active because display lists will be created herein.
//       All converted data (vertices, normals etc.) is kept until a new file is to be loaded
//       or the application terminates. 

	// clear previously allocated data
	ClearDisplayListGL();

	//paint background color
	glClearColor(m_BackColor.x, m_BackColor.y, m_BackColor.z, 1);
	
	//set the ambient color for the scene
	Vector4f ambient_col = MakeHVector(m_Ambient.x, m_Ambient.y, m_Ambient.z, 1);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_col);

    // FCenter is updated during conversion of the meshs
//    VectorScale(FCenter, 1 / TotalCount);

    for (Iterator cmesh(*this); cmesh.Valid(); cmesh.GotoNext()) {
		MeshGL* currMesh = cmesh.Contents();	

        // allocate a new display list
        currMesh->ListName = glGenLists(1);
        // now compile mesh data into this list
        glNewList(currMesh->ListName, GL_COMPILE);

		//if the version of the mesh database is prior to 3rd
		// then the faces are compiled in a clockwise fashion
		if (m_Version < 3) 
			glFrontFace(GL_CW);
        else 
			glFrontFace(GL_CCW);

		//enter the vertices
        glVertexPointer(3, GL_FLOAT, 0, currMesh->Vertices);
        glEnableClientState(GL_VERTEX_ARRAY);

        // enable the normal array if available
		if (currMesh->Normals != NULL) {
			glNormalPointer(GL_FLOAT, 0, currMesh->Normals);
			glEnableClientState(GL_NORMAL_ARRAY);
		}
		else 
			glDisableClientState(GL_NORMAL_ARRAY);

		// enable the texture coordinates array if available
		if (currMesh->TexCoords != NULL) {
			glTexCoordPointer(2, GL_FLOAT, 0, currMesh->TexCoords);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		else 
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);

		FaceGroup *matGroupsGL = currMesh->Faces.MatGroups;
        // for all of the mat groups in the mesh object
		for (int j = 0; j < currMesh->Faces.nMatGroups; j++) {

			// draw the desired material to set the current state		
			matGroupsGL[j].SurfPtr->RenderMaterialGL();

            // now draw the mesh data
            glDrawElements(GL_TRIANGLES, matGroupsGL[j].IndexCount, GL_UNSIGNED_INT, matGroupsGL[j].Indices);
		} //end for all mat groups

		// finished compiling this display list
		glEndList();
		CheckOpenGLError();
    } //end for all meshes

    //Build main display list
    m_MainListGL = glGenLists(1);
    glNewList(m_MainListGL, GL_COMPILE);
 
	// move scene so that its geometric center is at origin
//    glTranslatef(-FCenter[0], -FCenter[1], -FCenter[2]);

	//for every mesh call the local display lists
	for (Iterator mesh(*this); mesh.Valid(); mesh.GotoNext()) {
		MeshGL *currMesh = mesh.Contents();
		if (currMesh->Visible)
			glCallList(currMesh->ListName);
	}

	glEndList();
    CheckOpenGLError();
}


void
MeshListGL::ClearDisplayListGL()
{
	if (m_MainListGL > 0) {
		glDeleteLists(m_MainListGL, 1);
		m_MainListGL = 0;
	}

	//for every mesh clear the local display lists
	for (Iterator mesh(*this); mesh.Valid(); mesh.GotoNext()) {
		//extract the first element from the list
    	MeshGL *currMesh = mesh.Contents();
		
		if (currMesh->ListName > 0)
			//delete the local display list
			glDeleteLists(currMesh->ListName, 1);
	}
}



//----------------------------------------------------------------------------------------------------------------------
// free locally allocated data and reset structures, variables etc.
// also clear Surfaces, Objects and Lights before doing this
void 
MeshListGL::DeleteMeshList()
{
	ClearDisplayListGL();
	
	//Clear the mesh resources/////////////////
	while (n > 0) {
		//extract the first element from the list
    	MeshGL *currMesh = ExtractHead();
				
		if (currMesh->Vertices != NULL)
			delete [] currMesh->Vertices;

		if (currMesh->Normals != NULL)
			delete [] currMesh->Normals;

		if (currMesh->TexCoords != NULL)
			delete [] currMesh->TexCoords;
	
		if (currMesh->Faces.nMatGroups > 0) {
			// free material assignment data
			for (int j = 0; j < currMesh->Faces.nMatGroups; j++)
				delete [] currMesh->Faces.MatGroups[j].Indices;

			//delete the list of mat groups
			delete [] currMesh->Faces.MatGroups;
		}
		
		//clear the mesh objects list
		delete currMesh;
    }
}
