// ===================================================================
// 	File: plane.cpp								12/27/1999
//	By Kosntantin Khanyants						kon@unix.temple.edu
//	Description:
//		Plane support routines for OORT.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"

Planar::Planar(HallSurface *NewColor, Mapping *NewMapping): 
	Object3D(NewColor, 0, NewMapping)
{
    bad = 1;
}

Planar::Planar(float A, float B, float C, float D, HallSurface *NewColor, Mapping *NewMapping):
    Object3D(NewColor, 0, NewMapping)
{
    Normal = Vector3D(A, B, C);
    d = D;
    float mag = Magnitude(Normal);
    bad = fabs(mag) < Limits::Small;
    if (! bad) {
	Normal /= mag;
	d /= mag;
    }
}

Planar::Planar(const Vector3D& norm, float D, HallSurface *NewColor, Mapping *NewMapping):
    Object3D(NewColor, 0, NewMapping)
{
    Normal = norm;
    d = D;
    float mag = Magnitude(Normal);
    bad = fabs(mag) < Limits::Small;
    if (! bad) {
	Normal /= mag;
	d /= mag;
    }
}

int
Planar::NearestInt(Ray& ray, float& t, float maxt)
{
    if (bad)
	return 0;

    float dot = DotProd(Normal, ray.dir);

    // Watch for near-zero denominator
    if (fabs(dot) < Limits::Small)
		return 0;

    t = (-d - DotProd(Normal, ray.loc)) / dot;

    return t > Limits::Threshold && t < maxt;
}

SpanList *
Planar::FindAllIntersections(Ray& ray)
{
    float dot = DotProd(Normal, ray.dir);

    if (fabs(dot) == 0) {
	// Ray is parallel; if ray is in the plane, it intersects
	// the plane infinitely far away in both directions; otherwise
	// it intersects the ray not at all, because it is parallel.
	if (DotProd(Normal, ray.loc) + d < Limits::Small) {
	    SpanList *ret = new SpanList(Span(-Limits::Infinity, this,
					      Limits::Infinity, this));
	    return ret;
	}
	else return 0;
    }

    float t = (-d - DotProd(Normal, ray.loc)) / dot;

    if (dot < 0)
	return new SpanList(Span(t, this, Limits::Infinity, this));
    else {
	if (t < Limits::Threshold)
	    return 0;
	return new SpanList(Span(-Limits::Infinity, this, t, this));
    }
}

Vector3D
Planar::FindNormal(const Vector3D& v)
{
    return Normal;
}

void
Planar::ApplyTransform(const Matrix& tform)
{
    Vector3D v = Normal * -d;
    Normal = Normalize(RotateOnly(tform, Normal));
    d = -DotProd(tform * v, Normal);
}

float
Planar::PtDistance(Vector3D& v)
{
    return DotProd(v, Normal) + d;
}

/*
// ===================================================================
// Plane implementations
// ===================================================================

Plane::Plane(float A, float B, float C, float D, HallSurface *NewColor, Mapping *NewMapping):
    Planar(A, B, C, D, NewColor, NewMapping)	//added mapping to constructors 6/12/2000
{ }

int
Plane::NearestInt(Ray& ray, float& t, float maxt)
{
    Statistics::Intersections::Plane += 1;
    return Planar::NearestInt(ray, t, maxt);
}


Object3D *
Plane::Dup() const
{
    return new Plane(Normal.x, Normal.y, Normal.z, d, surf, mapping);
}
*/