// ===================================================================
//  File: sphere.cpp							1/15/2000
//	By Konstantin Khanyants						kon@unix.temple.edu
//	Description:
//		Sphere support routines for OORT.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"

// Sphere constructor.	Takes center, radius, color.
Sphere::Sphere(const Vector3D& C, float R, HallSurface *clr, Mapping *NewMapping):
  Object3D(clr, SPHERE, NewMapping), center(C), radius(R)
{
    radiussq = radius*radius;
    flags = SPHERE;	//ADDED
}

Sphere::~Sphere()
{
	if (mapping)
		delete mapping;
}

void
Sphere::Describe()
{
	CString p;
	p.Format("Sphere\n Center = (%1.4f, %1.4f, %1.4f)\n Radius = %f\n flags = %d\n Surface = ",
		center[0], center[1], center[2], radius, flags);
	p += surf->GetName();
	AfxMessageBox(p);
}

// Sphere intersection routine.
int
Sphere::NearestInt(Ray& ray, float& t, float maxt)
{
    float b, c, d;
    float t1, t2;
    Vector3D diff;

    Statistics::Intersections::Sphere++;

    diff = center - ray.loc;
    c = (float) DotProd(diff, diff) - radiussq;
    b = -2 * DotProd(diff, ray.dir);
    d = b*b - 4*c;

    if (d <= 0)
	return 0;

    d = (float) sqrt(d);
    t1 = float((-b+d)/2.0);
    t2 = float((-b-d)/2.0);

    if (t1 < Limits::Threshold || t1 >= maxt)
	t1 = -1;
    if (t2 < Limits::Threshold || t2 >= maxt)
	t2 = -1;
    if (t1 < 0) {
	if (t2 < 0)
	    return 0;
	t = t2;
    }
    else {
	if (t2 < 0)
	    t = t1;
	else
	    t = (t1 < t2) ? t1 : t2;
    }
    if (t2 < t1) {
	float temp = t1;
	t1 = t2;
	t2 = temp;
    }
    return t > Limits::Threshold;
}

SpanList *
Sphere::FindAllIntersections(Ray& ray)
{
    float b, c, d;
    float t1, t2;
    Vector3D diff;

    Statistics::Intersections::Sphere++;

    diff = center - ray.loc;
    c = (float) DotProd(diff, diff) - radiussq;
    b = -2 * DotProd(diff, ray.dir);
    d = b*b - 4*c;

    if (d <= 0)
	return 0;

    d = float(sqrt(d));
    t1 = float((-b+d)/2.0);
    t2 = float((-b-d)/2.0);

    // If both parameters are less than threshold, return no-intersection.
    if (t1 < Limits::Threshold && t2 < Limits::Threshold)
	return 0;

    if (t1 > t2)
	Swap(t1, t2);
    if (t1 < 0) {
	if (IsInside(ray.loc))
	    return new SpanList(Span(t1, this, t2, this));
	else
	    return new SpanList(Span(t2, this, Limits::Infinity, this));
    }
    return new SpanList(Span(t1, this, t2, this));
}

// Fills in the shade object structure with the partial derivatives
// of the object currently being shaded
void
Sphere::GetDerivatives(ShadingInfo& shade)
{
    //translate point of intersection to the origin
	Vector3D newPos = shade.p - center;
	float& X = newPos.x;		//get aliases for vertex components
	float& Y = newPos.y;
	float& Z = newPos.z;
	
	float theta = (float) atan(Y / X);
	float phi = (float) acos(Z / sqrt(X * X + Y * Y + Z * Z));	
		
	//find the derivatives	
	shade.dudt = Vector3D(float(-sin(theta) * sin(phi)), float(cos(theta) * sin(phi)), 0);
	shade.dvdt = Vector3D(float(cos(theta) * cos(phi)), float(sin(theta) * cos(phi)), float(-sin(phi)));
	
}

// Sphere normal-finding routine.
Vector3D
Sphere::FindNormal(const Vector3D& intersection)
{
    return(Normalize(intersection - center));
}

// Apply transform to sphere.  Only center point is affected.
void
Sphere::ApplyTransform(const Matrix& tform)
{
    center = tform * center;
}

// Duplicate sphere.
Object3D *
Sphere::Dup() const
{
    return(new Sphere(center, radius, surf, mapping->Dup()));
}

int
Sphere::IsInside(Vector3D& v)
{
    Vector3D diff = v - center;
    return Magnitude(diff) < radiussq;
}

float
Sphere::PtDistance(Vector3D& v)
{
    Vector3D diff = v - center;
    return DotProd(diff, diff) - radiussq;
}

AxisAlignedBox
Sphere::BBox() const
{
    return AxisAlignedBox(Vector3D(center - Vector3D(radius)),
			  Vector3D(center + Vector3D(radius)));
}