// ==============================================================
// 	File: surface.cpp						12/27/1999
//	By Konstantin Khanyants					kon@unix.temple.edu
//	Description:
//		Implementation of different shading models.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//           			 	Copyright (C) 1993 by Nicholas Wilt.
//
// ==============================================================

#include "stdafx.h"
#include "include/Oort.h"
#include "include/World.h"

HallSurface::~HallSurface()
{
	if (Diffuse_Tex)
		delete Diffuse_Tex;

	if (Bump)
		delete Bump;
}

void
HallSurface::PreMulTransform(const Matrix& tform)
{
    // Bump maps can't get transforms applied to them yet.
//  if (Bump)
//  Bump->PreMulTransform(tform);
    if (Diffuse_Tex)
	Diffuse_Tex->PreMulTransform(tform);

}

void
HallSurface::ApplyTransform(const Matrix& tform)
{
    // Bump maps can't get transforms applied to them yet.
//	if (Bump)
//	Bump->ApplyTransform(tform);	
    if (Diffuse_Tex)
	Diffuse_Tex->ApplyTransform(tform);

}

Surface *
HallSurface::Dup() const
{
    HallSurface *ret = new HallSurface;

	ret->name = name;

	if (Bump)
		ret->SetBumpMap((ImageBumpMap *) Bump->Dup());
	ret->SetBumpFlag(Bump_Used);

	if (Diffuse_Tex)
		ret->SetDiffuseTexture((ImageMap *) Diffuse_Tex->Dup());
	ret->SetDiffuseColor(Diffuse_Col);	
	ret->SetTextureFlag(Tex_Used);	
	
	ret->SetAmbient(Ambient);
	ret->SetSpecular(Specular, specnum);
	ret->SetShinStrength(shin_strength);
	ret->SetTransmitted(Transmitted, index);
	ret->SetReflected(Reflected);
	
	//OpenGL
	ret->PolygonMode = PolygonMode;
    ret->ShadeModel = ShadeModel;     
    ret->TwoSided = TwoSided;

    return ret;
}

int
HallSurface::Transparent()
{
	return Transmitted != RGBColor(0);
}

void
HallSurface::SetBumpMap(ImageBumpMap *_bumpmap)
{
	if (Bump)
		delete Bump;
    Bump = _bumpmap;
}

void
HallSurface::SetAmbient(const RGBColor & _ambient)
{
    Ambient = _ambient;
}

void
HallSurface::SetDiffuseTexture(ImageMap *_diffuse)
{	
	if (Diffuse_Tex)
		delete Diffuse_Tex;
    Diffuse_Tex = _diffuse;
}

void 
HallSurface::SetDiffuseColor(const RGBColor & diff_Col)
{
	Diffuse_Col = diff_Col;
}	

void
HallSurface::SetSpecular(const RGBColor & _specular, float _specnum)
{
    Specular = _specular;
    specnum = _specnum;
}

void
HallSurface::SetTransmitted(const RGBColor& _transmit, float _index)
{
    Transmitted = _transmit;
    index = _index;
}

void
HallSurface::SetReflected(const RGBColor& _reflect)
{
    Reflected = _reflect;
}

void
HallSurface::SetBumpFlag(BOOL state)
{
	if (Bump)
		Bump_Used = state;
}			

void
HallSurface::SetTextureFlag(BOOL state)
{
	if (Diffuse_Tex)
		Tex_Used = state;
}		

void
HallSurface::SetBumpAmount(float val)
{
	Bump_Amount = val;
	if (Bump)
		Bump->SetAmount(val);
}

void
HallSurface::SetTexAmount(float val)
{
	Tex_Amount = val;
	if (Diffuse_Tex)
		Diffuse_Tex->SetAmount(val);
}

float
HallSurface::GetBumpAmount()
{
	if (Bump)
		Bump_Amount = Bump->GetAmount();
	
	return Bump_Amount;
}

float
HallSurface::GetTexAmount()
{
	if (Diffuse_Tex)
		Tex_Amount = Diffuse_Tex->GetAmount();

	return Tex_Amount;
}

CString//char*
HallSurface::GetName()
{
	//char *newStr = NULL;
	//newStr = new char[strlen(name)+1];
	//if (newStr != NULL)
	//	strcpy(newStr, name);
	return name;//newStr;
}

void
HallSurface::SetName(CString newName)//char* newName)
{
	name = newName;
}

//IsSmooth is based on OpenGL's drawing mode
//used in triangle code
BOOL
HallSurface::IsSmooth()
{
	return (ShadeModel == GL_SMOOTH) ? TRUE : FALSE;
}

void
HallSurface::Describe()
{
	CString pcsName = "Surface Name = " + name;
	CString pcsMsg;
	pcsMsg.Format("\nShininess = %1.2f, Shininess Stren. = %1.2f\nIndex of Refraction = %1.4f\nTex Used = %d, Bump Used = %d\nAmbient = {%1.2f, %1.2f, %1.2f}\nDiffuse = {%1.2f, %1.2f, %1.2f}\nSpecular = {%1.2f, %1.2f, %1.2f}\nTransmitted = {%1.2f, %1.2f, %1.2f}\nReflected = {%1.2f, %1.2f, %1.2f}\nTex Amount = %1.2f, Bump Amount = %1.2f, Two Sided = %d\n",
		specnum, shin_strength, index, Tex_Used, Bump_Used, Ambient.x, Ambient.y, Ambient.z, Diffuse_Col.x, Diffuse_Col.y, Diffuse_Col.z, Specular.x, Specular.y, Specular.z, Transmitted.x, Transmitted.y, Transmitted.z, Reflected.x, Reflected.y, Reflected.z, Tex_Amount, Bump_Amount, TwoSided);
	CString pcsGL;

	CString poly_mode = "GL_FILL";
	if (PolygonMode == GL_LINE)
		poly_mode = "GL_LINE";
	CString shade_mode = "GL_SMOOTH";
	if (ShadeModel == GL_FLAT)
		shade_mode = "GL_FLAT";

	pcsGL.Format("Polygon Mode = %s, Shade Model = %s", LPCTSTR(poly_mode), LPCTSTR(shade_mode));
	pcsName += pcsMsg + pcsGL;
	AfxMessageBox(pcsName);

}

// OpenGL rendering function
void
HallSurface::RenderMaterialGL()
{
	float transparency = (float) fabs(GetGray(Transmitted));
	float shininessGL = float(specnum * 128.0 / 300); //convert specularity to OpenGL range
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininessGL);
//  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, Emission);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, MakeHVector(Ambient.x, Ambient.y, Ambient.z, transparency));
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MakeHVector(Diffuse_Col.x, Diffuse_Col.y, Diffuse_Col.z, transparency));
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, MakeHVector(Specular.x, Specular.y, Specular.z, transparency));

    glPolygonMode(GL_FRONT_AND_BACK, PolygonMode);
    glShadeModel(ShadeModel);

	if (TwoSided)
		glDisable(GL_CULL_FACE);
    else 
		glEnable(GL_CULL_FACE);

	//Render the OpenGL texture if the flag is set
    if (Tex_Used)
		//texture is enabled in the RenderTextureGL()
		Diffuse_Tex->RenderTextureGL();
	else
		glDisable(GL_TEXTURE_2D);

	// ambient, diffuse and specular colors values contain the same blend (alpha) factor
	if (transparency > 0) {
		glEnable(GL_BLEND);
        // depth buffer should be disabled while rendering semi transparent surfaces but
        // leads often to somewhat strange results as the mesh aren't sorted properly
        //glDepthMask(GL_FALSE);
        glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);
    }
    else {
        glDisable(GL_BLEND);
        //glDepthMask(GL_TRUE);
    }
}


//This function modulates the color between texture and diffuse color
void
HallSurface::BeginShadeLight(ShadingInfo& shade)
{
	//modulate the diffuse color by the amount of texture (linear interpolation), 
	//the texture is modulated in the ImageMap object
	diffuse = (Tex_Used) ? (1 - Tex_Amount) * Diffuse_Col + Diffuse_Tex->GetColor(shade) : Diffuse_Col;
}

//Scales the specular hilight with strength value
RGBColor
HallSurface::ShadeLight(ShadingInfo& shade, Vector3D& dir)
{
    float diffuseatt = DotProd(shade.normal, dir);
    RGBColor ret = (diffuseatt >= 0) ? (RGBColor) (diffuseatt * diffuse) : RGBColor(0);
    if (Specular != RGBColor(0)) {
		float specularatt = DotProd(shade.reflected, dir);
		if (specularatt >= 0)
			ret += (Specular * shin_strength) * float(pow(specularatt, specnum)); //modulate color by strength 6/13/2000
    }
    return ret;
}

RGBColor
HallSurface::ComputeColor(ShadingInfo& shade, int depth)
{	
	if (Bump_Used)
		Bump->PerturbNormal(shade); 	//BUMP MAPPING
	
    RGBColor AmtLight = RGBColor(0);

    if (Ambient != RGBColor(0))
	AmtLight += Ambient * shade.world->GetAmbientLight();

    if (Diffuse_Col != RGBColor(0) || Specular != RGBColor(0) || Diffuse_Tex) {
		BeginShadeLight(shade);
		AmtLight += shade.world->DiffuseLight(shade, this);
    }

    if (shade.world->HitDepthLimit(depth)) {
		//clamp the color value to 1 if it is greater than 1
		Minimize((Vector3D&)AmtLight, (Vector3D&)RGBColor(1));
		return AmtLight;
	}

    if (Reflected != RGBColor(0)) {
		Ray ReflectedRay(shade.p, shade.ray->ReflectRay(shade.normal));
		Statistics::ReflectedRays++;
		AmtLight += Reflected * shade.world->FindColor(ReflectedRay, depth+1);
    }

    if (Transmitted != RGBColor(0)) {
		Vector3D normal = shade.normal;
		float gamma, dot, sqrterm;

		Statistics::TransmittedRays++;

		dot = DotProd(shade.normal, -shade.incident);
    	if (dot > 0)
    	    gamma = 1 / index;
    	else {
    	    gamma = index;
    	    dot = -dot;
    	    normal = -normal;
    	}
    	sqrterm = float(1.0 - gamma*gamma*(1.0 - dot*dot));

    	// Check for total internal reflection.  
		// Do nothing if it applies.
    	if (sqrterm > 0) {
    	    sqrterm = float(gamma*dot - sqrt(sqrterm));
    	    Ray TransmittedRay = Ray(shade.p,
				     normal*sqrterm + shade.incident*gamma);
    	    AmtLight += Transmitted *
			shade.world->FindColor(TransmittedRay, depth+1);
    	}
    }
	//clamp the color value to 1 if it is greater than 1
	Minimize((Vector3D&)AmtLight, (Vector3D&)RGBColor(1));
	return AmtLight;
}


Surface *
DistributedHall::Dup() const
{
    DistributedHall *ret = new DistributedHall;

	if (Bump)
		ret->SetBumpMap((ImageBumpMap *) Bump->Dup());
	ret->SetBumpFlag(Bump_Used);

	if (Diffuse_Tex)
		ret->SetDiffuseTexture((ImageMap *) Diffuse_Tex->Dup());
	ret->SetDiffuseColor(Diffuse_Col);	
	ret->SetTextureFlag(Tex_Used);	
	
	ret->SetAmbient(Ambient);
	ret->SetSpecular(Specular, specnum);
	ret->SetShinStrength(shin_strength);
	ret->SetTransmitted(Transmitted, index);
	ret->SetReflected(Reflected);
    ret->ReflectParms = ReflectParms;
    ret->TransParms = TransParms;
    return ret;
}

void
DistributedHall::SetReflectParms(const DistribParms& dist)
{
    ReflectParms = dist;
}

void
DistributedHall::SetTransmitParms(const DistribParms& dist)
{
    TransParms = dist;
}

Vector3D
DistributedHall::NewSample(const DistribParms& d, const Vector3D *samples, int n)
{
    float ang = RandomValue(0, float(2*M_PI));
    float rad = RandomValue(0, d.GetAngle());

    Vector3D ret = RotationZMatrix(ang) * (float(sin(rad)) * Vector3D(1, 0, 0));
    if (! n)
		return ret;// + Vector3D(0, 0, cos(rad));
    
	float mindist = MinDist(ret, samples, n);
    ret += Vector3D(0, 0, float(cos(rad)));
    
	for (int i = 0; i < d.GetCandidates(); i++) {
		ang = RandomValue(0, float(2*M_PI));
		rad = RandomValue(0, d.GetAngle());

		Vector3D candidate = RotationZMatrix(ang) * (float(sin(rad)) * Vector3D(1, 0, 0));
		float dist = MinDist(candidate, samples, n);
		if (dist > mindist) {
		    ret = candidate;// + Vector3D(0, 0, cos(rad));
		    mindist = dist;
		}
    }
    return ret;
}

RGBColor
DistributedHall::SampleSpace(ShadingInfo& shade,
			     Vector3D& dir,
			     const DistribParms& dparms,
			     int depth,
			     long *IncMe)
{
    RGBColor Imin(MAXFLOAT), Imax(-MAXFLOAT), Intens(0, 0, 0);
    int n = 0;

    // Compute a transformation matrix that transforms a ray pointing
    // down the +Z axis so it points along dir.
    Vector3D z = dir;
    Vector3D y = Normalize(Vector3D(0, 1, 0) - dir * DotProd(Vector3D(0, 1, 0), dir));
   
	if (Magnitude(y) < Limits::Small) {
		cout << "Small y\n";	// npw-I've never seen this happen.
				// I just want to know when it does.
    }
    Vector3D x = CrossProd(y, z);
    Matrix xform = Transpose(GenRotation(x, y, z));

    Vector3D *pts = new Vector3D[dparms.GetMaxSamples()];

    for (int i = 0; i < dparms.GetMaxSamples(); i++) {
		pts[n] = NewSample(dparms, pts, n);
		n += 1;

		float mag = Magnitude(pts[i]);
		Vector3D shootat = pts[i] + Vector3D(0, 0, 1-mag*mag);
		shootat = xform * shootat;

		*IncMe += 1;
		RGBColor clr = shade.world->FindColor(Ray(shade.p, shootat), depth + 1);
		Minimize(Imin, clr);
		Maximize(Imax, clr);
		Intens += clr;

		if (i > dparms.GetMinSamples()) {
		    if (i >= dparms.GetMaxSamples())
				break;
			if (! MoreRays(Imin, Imax, dparms.GetThreshold()))
				break;
		}
    }
    delete[] pts;
    return Intens / float(n);
}

RGBColor
DistributedHall::ComputeColor(ShadingInfo& shade, int depth)	
{
	if (Bump_Used)						//BUMP MAPPING
		Bump->PerturbNormal(shade);
	
    RGBColor AmtLight = RGBColor(0);

    if (Ambient != RGBColor(0))
		AmtLight += Ambient * shade.world->GetAmbientLight();

    if (Diffuse_Col != RGBColor(0) || Specular != RGBColor(0) || Diffuse_Tex) {
		BeginShadeLight(shade);			//was shade.p);
		AmtLight += shade.world->DiffuseLight(shade, this);
    }

    if (shade.world->HitDepthLimit(depth)) {
		//clamp the color value to 1 if it is greater than 1
		Minimize((Vector3D&)AmtLight, (Vector3D&)RGBColor(1));
		return AmtLight;
	}

    if (Reflected != RGBColor(0)) {
		AmtLight += Reflected *
				SampleSpace(shade,
				shade.ray->ReflectRay(shade.normal),
				ReflectParms, 
				depth, 
				&Statistics::ReflectedRays);
    }


    if (Transmitted != RGBColor(0)) {
		Vector3D normal = shade.normal;
    	float gamma, dot, sqrterm;
    	dot = DotProd(shade.normal, -shade.incident);
    	if (dot > 0)
    	    gamma = 1 / index;
    	else {
    	    gamma = index;
    	    dot = -dot;
    	    normal = -normal;
    	}
    	sqrterm = float(1.0 - gamma*gamma*(1.0 - dot*dot));

    	// Check for total internal reflection.  
		// Do nothing if it applies.
		if (sqrterm > 0) {
		    sqrterm = float(gamma*dot - sqrt(sqrterm));
		    Vector3D transmitteddir = Normalize(shade.normal*sqrterm + 
							shade.incident*gamma);
		    AmtLight += Transmitted *
		    		SampleSpace(shade,
					    transmitteddir,
					    TransParms,
					    depth,
					    &Statistics::TransmittedRays);
		}
	}
	//clamp the color value to 1 if it is greater than 1
	Minimize((Vector3D&)AmtLight, (Vector3D&)RGBColor(1));
    return AmtLight;
}
