// ===================================================================
//  File: texture.cpp						12/26/1999
//	By Konstantin Khanyants					kon@unix.temple.edu
//	Description:
//		Texture-mapping support functions for OORT.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"
#include "include/World.h"


// ---------------------------------------------------------
// Texture functions implemented in-line.
// ---------------------------------------------------------
inline Vector3D
Texture::PreprocessVector(ShadingInfo& shade) const
{
    return reference * Vector3D(shade.p);
}


// ---------------------------------------------------------
// ---------------------------------------------------------
inline void
Texture::ApplyTransform(const Matrix& tform)
{
    reference *= tform;
}

inline void
Texture::PreMulTransform(const Matrix& tform)
{
    reference = tform * reference;
}

// ---------------------------------------------------------
// TextureMapping functions implemented in-line.
// ---------------------------------------------------------
inline
TextureMapping::TextureMapping()
{
//    mapping = 0;	//6/12/2000 no more mapping here
}

inline
//TextureMapping::TextureMapping(const Matrix& ref, Mapping *map): //ditto
TextureMapping::TextureMapping(const Matrix& ref):
    Texture(ref)
{
//    mapping = map;  //6/12/2000 no more mapping here
}

inline
TextureMapping::~TextureMapping()
{
//    delete mapping;  //6/12/2000 since mapping is not in Texture
}

//using mapping get the UV's for the texture process
Vector3D
TextureMapping::PreprocessVector(ShadingInfo& shade) const
{
	//now you can access many different mapping types of
	// one surface through shade's currrent object
	Mapping* mapping = shade.obj->GetMapping();
	if (mapping)
		return mapping->Map3DTo2D(shade.obj, Texture::PreprocessVector(shade));
	else
		return Texture::PreprocessVector(shade);

}



// ---------------------------------------------------------
// PureColor functions implemented in-line.
// ---------------------------------------------------------
inline RGBColor
PureColor::GetColor(ShadingInfo& shade) const
{
    return Color;
}

inline Texture *
PureColor::Dup() const
{
    return new PureColor(Color);
}

// -------------------------------------------------------------------
// ColorMap classes.
//	ColorMapEntry - encapsulates a single entry in the color map.
//	ColorMap - maps a number in the range [0, 1] to an interpolated
//		   vector given a set of ColorMapEntries.
// -------------------------------------------------------------------

ColorMap::~ColorMap()
{
    for (int i = 0; i < num_entries; i++)
        delete entries[i];
    delete[] entries;
}

void
ColorMap::AddEntry(ColorMapEntry *entry)
{
    if (num_entries >= max_entries) {
        ColorMapEntry **newentries = new ColorMapEntry*[max_entries <<= 1];
        for (int i = 0; i < num_entries; i++)
            newentries[i] = entries[i];
        delete[] entries;
        entries = newentries;
    }
    entries[num_entries++] = entry->Dup();
}

RGBColor
ColorMap::ComputeColor(float value)
{
    for (int i = 0; i < num_entries; i++)
	if (entries[i]->IsInside(value))
	    return entries[i]->ComputeColor(value);
    return RGBColor(0);
}

// ---------------------------------------------------------
// Textures
//	These convert points in 3-space into colors.
// ---------------------------------------------------------

// ---------------------------------------------------------
// Checkerboard texture.
// ---------------------------------------------------------

Checkerboard::Checkerboard(float XSize, float YSize,
			   Texture *Text1, Texture *Text2,
			   const Matrix& ref/*, Mapping *map*/):
	TextureMapping(ref/*, map*/), xsize(XSize), ysize(YSize),
	text1(Text1), text2(Text2)
{
}

RGBColor
//REM Checkerboard::GetColor(const Vector3D& Int) const
Checkerboard::GetColor(ShadingInfo& shade) const
{
    Vector3D v = PreprocessVector(shade);
    int U = posmod(v.x, xsize) < 0.5;
    int V = posmod(v.y, ysize) < 0.5;
    if (U ^ V)
    	return text1->GetColor(shade);
    else
    	return text2->GetColor(shade);
}

Texture *
Checkerboard::Dup() const
{
    return new Checkerboard(xsize, ysize, 
    			    text1->Dup(), text2->Dup(), 
			    reference);
}



// ---------------------------------------------------------
// ImageMap texture
// ---------------------------------------------------------

ImageMap::ImageMap(const Matrix& ref, /*Mapping *map,*/
		   int Interp,
		   int Once,
		   const char *filename,
		   float rot, 
		   float xscale, float yscale,
		   GLint minFilter, GLint magFilter,
		   GLenum wrapS, GLenum wrapT, 
		   GLenum mode, float tex_amount) : TextureMapping(ref/*, map*/)
{
    interp = (Interp != 0);
	tex_amt = tex_amount;
    once = (Once != 0);
	rot_angle = rot;	//ADDED for 3D studio
    scale = Vector3D(xscale, yscale, 0);
	
	//OpenGL
	MinFilter = minFilter;
	MagFilter = magFilter;
	WrapS = wrapS;
	WrapT = wrapT;
	Mode = mode;
	TextureObject = 0;

	file = new ImageFile(FALSE);	//say no to the Open File dialog
	if (!file->OpenFile(filename)) {		//try to open the file quietly
		CString pcsMsg = "Can't open texture file ";
		pcsMsg += filename;
		AfxMessageBox(pcsMsg);
	}
	else	//if the file was opened
		//initialize the texture object
		InitTextureObjectGL();
}

ImageMap::ImageMap(const Matrix& ref,
		   int Interp,
		   int Once,
		   ImageFile *File,
		   float rot,
		   float xscale, float yscale,
		   GLint minFilter, GLint magFilter,
		   GLenum wrapS, GLenum wrapT, 
		   GLenum mode, float tex_amount) : 
    TextureMapping(ref/*, map*/)
{
	tex_amt = tex_amount;
    interp = (Interp != 0);
    once = (Once != 0);
    file = File;
    rot_angle = rot;	//ADDED for 3D studio
    scale = Vector3D(xscale, yscale, 0);

	//OpenGL
	MinFilter = minFilter;
	MagFilter = magFilter;
	WrapS = wrapS;
	WrapT = wrapT;
	Mode = mode;
	TextureObject = 0;

	InitTextureObjectGL();
}

ImageMap::ImageMap(const ImageMap& initMap) :
	TextureMapping(initMap.reference)
{
	interp = initMap.interp;
    once = initMap.once;
    file = initMap.file->Dup();

    rot_angle = initMap.rot_angle;	//ADDED for 3D studio
    scale = initMap.scale;

	//OpenGL
	TextureObject = initMap.TextureObject;    // the OpenGL texture object name
	WrapS = initMap.WrapS;
	WrapT = initMap.WrapT;     // texture wrapping (GL_CLAMP, GL_REPEAT)
    MinFilter = initMap.MinFilter;        // minification filter (GL_NEAREST, GL_LINEAR, GL_NEAREST_MIPMAP_NEAREST etc.)
    MagFilter = initMap.MagFilter;        // magnification filter (GL_NEAREST, GL_LINEAR)
    Mode = initMap.Mode;   
}


ImageMap::~ImageMap()
{
	if (file)
		delete file;

	DeleteTextureObjectGL();
}

////////OPEN GL/////////////

// Texture Initializing functions for OpenGL
static inline float Trunc(double x) { return float(int(x));}
//----------------------------------------------------------------------------------------------------------------------

static int RoundDownToPowerOf2(int Value)
{
	double logTwo;
	int result;

	//calculate log base 2
	logTwo = log(Value) / log(2);

	if (Trunc(logTwo) < logTwo)
		result = (int) pow(2, Trunc(logTwo));
    else 
		result = Value;
	
	return result;
}

void
ImageMap::DeleteTextureObjectGL()
{
	if (TextureObject != 0)
		glDeleteTextures(1, &TextureObject);

	//reset the object
	TextureObject = 0;
}


void
ImageMap::InitTextureObjectGL()
{
	unsigned char* pData;
    unsigned char* newPixels;

	if (file == NULL) 
		return;

	if (!file->ImageLoaded())
		return;

	//delete the old object if there was one
	DeleteTextureObjectGL();

	// check width and height being power of 2
	int Width = file->Width();
	int Height = file->Height();
    int TargetWidth = RoundDownToPowerOf2(Width);
    int TargetHeight = RoundDownToPowerOf2(Height);
    
	try {
		//get the pixels and reformat them into a packed OpenGL array
		int nSamples = file->SamplesPerPixel();
		unsigned char** imagePixels = file->GetPixels();
		newPixels = new unsigned char[Width * Height * nSamples];
		
		if (newPixels == NULL)
			throw new CException;

		int bytesPerRow = Width * nSamples;
	
		//copy the data row by row
		for (int i = 0; i < Height; i++)
			memcpy(&newPixels[i * bytesPerRow], imagePixels[i], bytesPerRow);

		//get the maximum tex dimension fo OpenGL
		GLint glMaxTexDim;
		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &glMaxTexDim);
		
		if (TargetWidth >= glMaxTexDim)
			TargetWidth = glMaxTexDim;
		
		if (TargetHeight >= glMaxTexDim)
			TargetHeight = glMaxTexDim;
	
		//assign the pixel format of the data, either RGB or Grayscale
		GLint pixelFormatGL = (nSamples == 3) ? GL_RGB : GL_LUMINANCE;

		//if the image needs to be scaled
		if ((TargetWidth != Width) || (TargetHeight != Height)) {
			pData = new unsigned char[TargetWidth * TargetHeight * nSamples];
			if (pData != NULL)
				gluScaleImage(pixelFormatGL, Width, Height, GL_UNSIGNED_BYTE, newPixels, 
							  TargetWidth, TargetHeight, GL_UNSIGNED_BYTE, pData);
			else
				throw new CException;
		}	
		//otherwise just assign pData to the image
		else
			pData = newPixels;
  
		//get rid of the texture object if one is currently active
		DeleteTextureObjectGL();

		// create a 2D texture object from the given bitmap which must 
		// have width and height as being power of 2
		if (TargetHeight > 2 && TargetWidth > 2) {
			glGenTextures(1, &TextureObject);
			glBindTexture(GL_TEXTURE_2D, TextureObject);

			if ((MinFilter == GL_NEAREST) || (MinFilter == GL_LINEAR))
				glTexImage2D(GL_TEXTURE_2D, 0, pixelFormatGL, TargetWidth, TargetHeight,
								0, GL_RGB, GL_UNSIGNED_BYTE, pData);
			else 
				gluBuild2DMipmaps(GL_TEXTURE_2D, pixelFormatGL, TargetWidth, TargetHeight,
								pixelFormatGL, GL_UNSIGNED_BYTE, pData);
		}
		else
			throw new CException;
	
	
	}
	catch (CException* e) {
		
		TextureObject = 0;
		CString pcsMsg = "Texture object could not be created: ";
		pcsMsg += file->GetFileName();
		AfxMessageBox(pcsMsg);
		e->Delete();
	}

	//get rid of the texture (allocated or not)
	if (newPixels != NULL) {
		if (pData == newPixels)
			delete [] pData;
		else {
			if (pData != NULL)
				delete [] pData;
			delete [] newPixels;	
		}
	}
}

// OpenGL texturing
void
ImageMap::RenderTextureGL()
{
    if (TextureObject != 0) {
		glBindTexture(GL_TEXTURE_2D, TextureObject);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, WrapS);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, WrapT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, MinFilter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, MagFilter);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, Mode);

		glEnable(GL_TEXTURE_2D);
	}
    else 
		glDisable(GL_TEXTURE_2D);
}
//////////////////////end OpenGL////////////////////////////

///Added for changing the texture map in materials dialog box
void
ImageMap::SetImageFile(char* filename)
{
	if (file != NULL) {
		//if the names are the same, don't open it
		if (!strcmp(filename, file->GetFileName()))
			return;
		delete file;
	}

	file = new ImageFile(FALSE);	//say no to the Open File dialog
	file->OpenFile(filename);		//open the file quietly

	//update OpenGL
	InitTextureObjectGL();
}

void
ImageMap::SetImageFile(ImageFile* theFile)
{
	//if we're trying to assign the same file.. don't
	if (theFile == file)
		return;

	if (file != NULL)
		delete file;

	file = theFile;

	InitTextureObjectGL();
}


RGBColor
//REM ImageMap::GetColor(const Vector3D& Intersection) const
ImageMap::GetColor(ShadingInfo& shade) const
{
    Vector3D v = PreprocessVector(shade); //* scale;

	//rotation of the bitmap
	if (rot_angle != 0.0)
	{
		v = TranslationMatrix(Vector3D(-0.5, -0.5, 0)) * v; 
		v = RotationZMatrix(rot_angle) * v;	//in radians
		v = TranslationMatrix(Vector3D(0.5, 0.5, 0)) * v;
	}
	
	v *= scale;
	
    v *= Vector3D(float(file->Width()), float(file->Height()), 0);
    if (! once) {
	v.x = (float) fmod(v.x, file->Width());
	if (v.x < 0)
	    v.x += file->Width();
	v.y = (float) fmod(v.y, file->Height());
	if (v.y < 0)
	    v.y += file->Height();
    }
    long ulx = (long) v.x;
    long uly = (long) v.y;
    if (interp) {
	float horzweight = v.x - ulx;
	float vertweight = v.y - uly;
	float ul = (1 - horzweight) * (1 - vertweight);
	float ur = horzweight * (1 - vertweight);
	float ll = (1 - horzweight) * vertweight;
	float lr = horzweight * vertweight;
	return (ul * file->GetRGBColorPixel(ulx, uly) +
	       ur * file->GetRGBColorPixel(ulx+1, uly) +
	       ll * file->GetRGBColorPixel(ulx, uly+1) +
	       lr * file->GetRGBColorPixel(ulx+1, uly+1)) * tex_amt;
    }
    else
	return file->GetRGBColorPixel(ulx, uly) * tex_amt;
}

Texture*
ImageMap::Dup() const
{
	return new ImageMap(*this);
}

// ---------------------------------------------------------
// SolidTexture class
//	All solid textures are derived from this guy.
// ---------------------------------------------------------

SolidTexture::SolidTexture(const Matrix& ref,
			   ColorMap *CMap): 
    Texture(ref)
{
    cmap = CMap;
}

// -------------------------------------
// Bozo texture
//	Just maps the solid noise value 
//	at the intersection point into 
//	a color map.
// -------------------------------------

Bozo::Bozo(const Matrix& ref, float Turbulence, ColorMap *CMap):
    SolidTexture(ref, CMap)
{
    turbulence = Turbulence;
}

RGBColor
Bozo::GetColor(ShadingInfo& shade) const
{
    Vector3D local = PreprocessVector(shade);
    float noise;

    if (turbulence != 0.0)
	local += GlobalNoise::Noise->DTurbulence(shade.p) * turbulence;
    noise = GlobalNoise::Noise->Noise(local);
    return cmap->ComputeColor(noise);
}

Texture *
Bozo::Dup() const
{
    return new Bozo(reference, turbulence, cmap);
}

// -------------------------------------
// Wood texture.
// -------------------------------------

Wood::Wood(const Matrix& Ref, float Turbulence, ColorMap *CMap):
    SolidTexture(Ref, CMap)
{
    turbulence = Turbulence;
}

RGBColor
Wood::GetColor(ShadingInfo& shade) const
{
    Vector3D v = PreprocessVector(shade);

    Vector3D turb = GlobalNoise::Noise->DTurbulence(v);

    v += Vector3D(cycloidal(v.x + turb.x),
		  cycloidal(v.y + turb.y),
		  0) * turbulence;
    float len = Magnitude(v);
    float noise = triangle_wave(len);
    return cmap->ComputeColor(noise);
}

Texture *
Wood::Dup() const
{
    return new Wood(reference, turbulence, cmap);
}

// -------------------------------------
// Marble texture.
// -------------------------------------

Marble::Marble(const Matrix& Ref, float Turbulence, ColorMap *CMap):
	SolidTexture(Ref, CMap)
{
    turbulence = Turbulence;
}

RGBColor
Marble::GetColor(ShadingInfo& shade) const
{
    Vector3D v = PreprocessVector(shade);
    float noise = triangle_wave(v.x +
				GlobalNoise::Noise->Turbulence(v) *
				turbulence);
    return cmap->ComputeColor(noise);
}

Texture *
Marble::Dup() const
{
    return new Marble(reference, turbulence, cmap);
}

// -------------------------------------
// Gradient texture.
// -------------------------------------

Gradient::Gradient(const Matrix& Ref, float Turbulence, ColorMap *CMap):
    SolidTexture(Ref, CMap)
{
    turbulence = Turbulence;
}

RGBColor
Gradient::GetColor(ShadingInfo& shade) const
{
    Vector3D v = PreprocessVector(shade);
    if (turbulence != 0.0)
	v += GlobalNoise::Noise->DTurbulence(v) * turbulence;
    return cmap->ComputeColor(posmod(posmod(v.x, 1) +
				     posmod(v.y, 1) +
				     posmod(v.z, 1), 1));
}

Texture *
Gradient::Dup() const
{
    return new Gradient(reference, turbulence, cmap);
}

// ---------------------------------------------------------
// WoodUpstill
//	Wood texture derived from Listing 16.15 in Steve 
//	Upstill's book, _The RenderMan Companion_.
//	(Addison-Wesley, 1990).
// ---------------------------------------------------------

WoodUpstill::WoodUpstill(const Matrix& ref,
			 float _ringscale,
			 const RGBColor& _light,
			 const RGBColor& _dark): 
    SolidTexture(ref, 0)
{
    ringscale = _ringscale;
    light = _light;
    dark = _dark;
}

RGBColor
WoodUpstill::GetColor(ShadingInfo& shade) const
{
    Vector3D p = PreprocessVector(shade);
    p += GlobalNoise::Noise->DNoise(p);

    float r = float(hypot(p.y, p.z) * ringscale);
    r += GlobalNoise::Noise->Noise(r);
    r -= float(floor(r));
    r = Hermite(0, 0.8f, r) - Hermite(0.83f, 1.0, r);
    return MixColors(light, dark, r);
}

Texture *
WoodUpstill::Dup() const
{
    return new WoodUpstill(reference, ringscale, light, dark);
}

// ---------------------------------------------------------
// Granite
//	Granite texture derived from Listing 16.15 in Steve 
//	Upstill's book, _The RenderMan Companion_.
//	(Addison-Wesley, 1990).
// ---------------------------------------------------------

Granite::Granite(const RGBColor& _K,
		 const Matrix& _reference):
    SolidTexture(_reference, 0)
{
    K = _K;
}

RGBColor
Granite::GetColor(ShadingInfo& shade) const
{
    Vector3D p = PreprocessVector(shade);
    float sum = 0, freq = 1;
    for (int i = 0; i < 6; i++) {
	sum += float(fabs(0.5 - GlobalNoise::Noise->Noise(4 * freq * p)) / freq);
	freq *= 2;
    }
    return K * sum;
}

Texture *
Granite::Dup() const
{
    return new Granite(K, reference);
}
