// ===================================================================
// 	File: triangle.cpp						1/6/2000
//	By Konstantin Khanyants					kon@unix.temple.edu
//	Description:
//		Triangle support routines for OORT. Added for 3D Studio
//		polygon support.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"

//Triangle constructor
Triangle::Triangle(Vertex *NewVertices, HallSurface *NewColor, Mapping* NewMapping):
					Planar(NewColor, NewMapping)
{
	Vertices = new Vertex[3];
	flags = TRIANGLE;
	for (int i = 0; i < 3; i++)
		Vertices[i] = NewVertices[i];   
	Precompute();
}

Triangle::Triangle(const Triangle& x): Planar(x.surf, x.mapping->Dup())
{
    Vertices = new Vertex[3];
    flags = TRIANGLE;
    for (int i = 0; i < 3; i++)
		Vertices[i] = x.Vertices[i];
    Precompute();
}


//Triangle Destructor
Triangle::~Triangle()
{
	if (mapping)
		delete mapping;
	mapping = NULL;
	delete[] Vertices;
	Vertices = NULL;
}	


// Precompute various parameters about the polygon:
//	The plane equation (including normal)
//	which component of the normal is dominant (needed
//		for intersection routine)
//	Bounding rectangle of projection (used by intersection
//		routine to eliminate intersections)
//	Overall bounding box of polygon (used by intersection
//		routine to intersect with ray bounding boxes)
void 
Triangle::Precompute()
{
    int i;			  // Index variable
    float max;			  // Maximum component of normal
    Vector3D minvert, maxvert;	  // Bounds of polygon

    // Compute the plane equation of the polygon. 5/2/2000
	Vector3D xprod = CrossProd(Vertices[2].Pos - Vertices[1].Pos,
			       Vertices[0].Pos - Vertices[1].Pos);
    
    // Compute the plane equation of the polygon.
    //ADDED
//    Vector3D xprod = CrossProd(Vertices[0].Pos - Vertices[1].Pos,
//			       Vertices[2].Pos - Vertices[1].Pos);
    
    if (fabs(Magnitude(xprod)) > Limits::Small) {
		bad = 0;
		Normal = Normalize(xprod);
    }
    else {
		bad = 1;
		return;
    }
    d = -DotProd(Normal, Vertices[0].Pos);

    // Precompute which.  which is 0 if the normal is dominant
    // in the X direction, 1 if the Y direction, 2 if the Z direction.

    which = 0;			  // Assume we'll ignore X for now
    max = float(fabs(Normal.x));
    if (fabs(Normal.y) > max) {   // Check to see if we ignore Y
		max = float(fabs(Normal.y));	  // Yes, ignore Y instead
		which = 1;
    }
    if (fabs(Normal.z) > max)	  // Check to see if we ignore Z
		which = 2;		  // Yep

    // Compute bounding rectangle of projection.  This is fairly
    // successful at rejecting nonintersections.
    Vertices[0].Pos.ExtractVerts(&umin, &vmin, which);
    umax = umin;
    vmax = vmin;
    for (i = 1; i < 3; i++) {
	float tempx, tempy;

	Vertices[i].Pos.ExtractVerts(&tempx, &tempy, which);
	if (tempx < umin)
	    umin = tempx;
	else if (tempx > umax)
	    umax = tempx;
	if (tempy < vmin)
	    vmin = tempy;
	else if (tempy > vmax)
	    vmax = tempy;
    }
    
}

void Triangle::Describe()
{
	CString desc;
	CString Actual;
	for (int i = 0; i < 3; i++) {
	desc.Format("Vertices[%d] = (%1.2f, %1.2f, %1.2f)\nNormals[%d] = (%1.2f, %1.2f, %1.2f)\nTexVerts[%d] = (%1.2f, %1.2f)\n",
		i, Vertices[i].Pos[0], Vertices[i].Pos[1], Vertices[i].Pos[2], i, Vertices[i].Normal[0], Vertices[i].Normal[1], Vertices[i].Normal[2],
		i, Vertices[i].UV[0], Vertices[i].UV[1]);
		Actual += desc;
	}
	Actual += "Surface = "  + surf->GetName();
	AfxMessageBox(Actual);
}


// Apply transform to a triangle: transform each vertex, then
// re-precompute all the precomputed information.
void
Triangle::ApplyTransform(const Matrix& tform)
{
    for (int i = 0; i < 3; i++) {
		Vertices[i].Pos = tform * Vertices[i].Pos;
   		//Note: vertex normals must be assigned before this
	
		//if rotation matrix, rotate vertex normals
    	Vertices[i].Normal = Normalize(RotateOnly(tform, Vertices[i].Normal));
    }
  
    Precompute();    ///precompute normals there
}

// Duplicate a polygon.
Object3D *
Triangle::Dup() const
{
    return new Triangle(Vertices, surf, mapping->Dup());
}

AxisAlignedBox
Triangle::BBox() const
{
    int i;
    Vector3D minV = Vertices[0].Pos;
    Vector3D maxV = Vertices[0].Pos;

    for (i = 1; i < 3; i++) {
	Minimize(minV, Vertices[i].Pos);
	Maximize(maxV, Vertices[i].Pos);
    }
    // Make sure the bounding box isn't flat.
    Vector3D diff = minV - maxV;
    for (i = 0; i < 3; i++)
	if (fabs(diff[i]) < Limits::Threshold) {
	    minV[i] -= Limits::Threshold;
	    maxV[i] += Limits::Threshold;
	}
    return AxisAlignedBox(minV, maxV);
}

Vector3D
Triangle::FindNormal(const Vector3D& v)
{	
	//interpolation depends on material so OpenGL's shade model
	// variable must be set in the surface
	if (surf->IsSmooth())	//if interpolation of normals is on, then
	{
		Vector3D norm;
    	
		//interpolate the normal between the three vertices
		norm = Normalize((1 - (alpha + beta)) * Vertices[0].Normal +
				alpha * Vertices[1].Normal + beta * Vertices[2].Normal);

		//The POV Ray method
//		 NTemp1 = N1 + u(N2 - N1);
//	     NTemp2 = N1 + u(N3 - N1);
//	     Result = normalize (NTemp1 + v(NTemp2 - NTemp1))
//		Vector3D temp1 = Vertices[0].Normal + alpha * (Vertices[1].Normal - Vertices[0].Normal);
//		Vector3D temp2 = Vertices[0].Normal + alpha * (Vertices[2].Normal - Vertices[0].Normal);
//		norm = Normalize(temp1 + beta * (temp2 - temp1));
		return norm;
	}
	else
		return Planar::FindNormal(v);  //otherwise return the normal to the plane    
}

/*
SpanList *
Triangle::FindAllIntersections(Ray& ray)
{
	//copy the triangle so it can be used in computation
	Triangle *tri = new Triangle(*this); 

	//if normal is to be interpolated(determined by material) then rotate 
	// tri by angle between the two normals, otherwise use the plane normal
	if (surf->IsSmooth()) {//interp) {
		Vector3D newNorm = shade.normal;//tri->FindNormal(shade.p);  //find the new normal
		
		//find the angle between the two normals
		float theta = (float) acos(DotProd(Normal, newNorm));
		
		//get the axis to rotate the triangle by
		Vector3D rotVec = CrossProd(Normal, newNorm);
		
		//rotate the triangle around the axis defined by rotVec by the angle theta.
		tri->ApplyTransform(RotationAxisMatrix(rotVec, theta));
	}
	//else USE PLANE NORMAL

	SpanList *ret = NULL;

	float dot = DotProd(tri->Normal, ray.dir);

    if (fabs(dot) == 0) {
		// Ray is parallel; if ray is in the plane, it intersects
		// the plane infinitely far away in both directions; otherwise
		// it intersects the ray not at all, because it is parallel.
		if (DotProd(tri->Normal, ray.loc) + tri->d < Limits::Small) {
		    ret = new SpanList(Span(-Limits::Infinity, this,
					      Limits::Infinity, this));
		}
    }

    float t = (-tri->d - DotProd(tri->Normal, ray.loc)) / dot;

    if (dot < 0)
		ret = new SpanList(Span(t, this, Limits::Infinity, this));
    else {
		if (t < Limits::Threshold)
			ret = NULL;
		ret = new SpanList(Span(-Limits::Infinity, this, t, this));
    }

	delete tri;
	return ret;
}
*/

Vertex&
Triangle::GetVertex(int n)
{
	if (n == 0)
		return Vertices[0];
	else if (n == 1)
		return Vertices[1];
	else
		return Vertices[2];
}
/*
//Intersection routine for the triangle with calculations for
//alpha and beta terms, call this function before you do shading
//of the object	
int
Triangle::NearestInt(Ray& ray, float& t, float maxt)
{	
	//copy the triangle so it can be used in computation
	Triangle *tri = new Triangle(*this);
	BOOL interFound = FALSE;

	Statistics::Intersections::Triangle++;
	
	if (surf->IsSmooth()) {
		for (int i = 0; i < 3; i++) {
		
			Vector3D newNorm = Vertices[i].Normal;  //find the new normal
		
			//find the angle between the two normals
			float theta = (float) acos(DotProd(Normal, newNorm));
		
			//get the axis to rotate the triangle by
			Vector3D rotVec = CrossProd(Normal, newNorm);
		
			//rotate the triangle around the axis defined by rotVec by the angle theta.
			tri->ApplyTransform(RotationAxisMatrix(rotVec, theta));
			
			if (tri->Planar::NearestInt(ray, t, maxt)) {
				interFound = TRUE;
				break;
			}
			//rotate the triangle back.
			tri->ApplyTransform(RotationAxisMatrix(rotVec, -theta));
		}
	}
	else if (tri->Planar::NearestInt(ray, t, maxt))
			interFound = TRUE;


	//find the intersection point and t 
    if (interFound) {

	float u, v;
	
	// Project the intersection point onto the coordinate plane
	// specified by which.
	ray.Extrap(t).ExtractVerts(&u, &v, tri->which);
	
	// Reject intersection if outside projected bounding rectangle
	if (u < tri->umin || u > tri->umax || v < tri->vmin || v > tri->vmax) {
		delete tri;
	    return 0;
	}

	//code adapted from Graphics Gems III, original by Didier Badouel
	float u0, v0;
	float u1, v1;
	float u2, v2;
	Vector3D diff;
	
	//begin barycentric intersection algorithm
	diff = (ray.Extrap(t) - tri->Vertices[0].Pos);
	diff.ExtractVerts(&u0, &v0, tri->which);
	
	diff = (tri->Vertices[1].Pos - tri->Vertices[0].Pos);
	diff.ExtractVerts(&u1, &v1, tri->which);
	
	diff = (tri->Vertices[2].Pos - tri->Vertices[0].Pos);
	diff.ExtractVerts(&u2, &v2, tri->which);
	 
	//calculate and compare barycentric coordinates
	if (u1 == 0) {		//uncommon case
		beta = u0 / u2;
		if (beta < 0 || beta > 1) {
			delete tri;
			return 0;
		}
		alpha = (v0 - beta * v2) / v1;
	}
	else {			//common case, used for this analysis
		beta = (v0 * u1 - u0 * v1) / (v2 * u1 - u2 * v1);
		if (beta < 0 || beta > 1) {
			delete tri;
			return 0;
		}
		alpha = (u0 - beta * u2) / u1;
	}
	if (alpha < 0 || (alpha + beta) > 1.0) {
		delete tri;
		return 0;
	}
	delete tri;
	return 1;
	
	}
	//if intersection with the plane is not successful
	else {
		delete tri;
		return 0;
	}
}
*/
//Intersection routine for the triangle with calculations for
//alpha and beta terms, call this function before you do shading
//of the object	
int
Triangle::NearestInt(Ray& ray, float& t, float maxt)
{	
	Statistics::Intersections::Triangle++;
	
	//find the intersection point and t 
    if (Planar::NearestInt(ray, t, maxt)) {
	
	float u, v;
	
	// Project the intersection point onto the coordinate plane
	// specified by which.
	ray.Extrap(t).ExtractVerts(&u, &v, which);
	
	// Reject intersection if outside projected bounding rectangle
	if (u < umin || u > umax || v < vmin || v > vmax)
	    return 0;
	
	//code adapted from Graphics Gems III, original by Didier Badouel
	float u0, v0;
	float u1, v1;
	float u2, v2;
	Vector3D diff;
	
	//begin barycentric intersection algorithm
	diff = (ray.Extrap(t) - Vertices[0].Pos);
	diff.ExtractVerts(&u0, &v0, which);
	
	diff = (Vertices[1].Pos - Vertices[0].Pos);
	diff.ExtractVerts(&u1, &v1, which);
	
	diff = (Vertices[2].Pos - Vertices[0].Pos);
	diff.ExtractVerts(&u2, &v2, which);
	 
	//calculate and compare barycentric coordinates
	if (u1 == 0) {		//uncommon case
		beta = u0 / u2;
		if (beta < 0 || beta > 1)
			return 0;
		alpha = (v0 - beta * v2) / v1;
	}
	else {			//common case, used for this analysis
		beta = (v0 * u1 - u0 * v1) / (v2 * u1 - u2 * v1);
		if (beta < 0 || beta > 1)
			return 0;
		alpha = (u0 - beta * u2) / u1;
	}
	if (alpha < 0 || (alpha + beta) > 1.0)
		return 0;
	
	return 1;
	
	}
	//if intersection with the plane is not successful
	else
		return 0;

}


//Fills in the shade object structure with the partial derivatives
// of the object currently being shaded
void
Triangle::GetDerivatives(ShadingInfo& shade)
{
	//copy the triangle so it can be used in computation
	Triangle *tri = new Triangle(*this); 
	
	//if normal is to be interpolated(determined by material) then rotate 
	// tri by angle between the two normals, otherwise use the plane normal
	if (surf->IsSmooth()) {//interp) {
		Vector3D newNorm = shade.normal;//tri->FindNormal(shade.p);  //find the new normal
		
		//find the angle between the two normals
		float theta = (float) acos(DotProd(Normal, newNorm));
		
		//get the axis to rotate the triangle by
		Vector3D rotVec = CrossProd(Normal, newNorm);
		
		//rotate the triangle around the axis defined by rotVec by the angle theta.
		tri->ApplyTransform(RotationAxisMatrix(rotVec, theta));
	}
	//else USE PLANE NORMAL

	float& u0 = tri->Vertices[0].UV.x;	//get aliases for texture vertex components
	float& u1 = tri->Vertices[1].UV.x;
	float& u2 = tri->Vertices[2].UV.x;
	float& v0 = tri->Vertices[0].UV.y;
	float& v1 = tri->Vertices[1].UV.y;
	float& v2 = tri->Vertices[2].UV.y;
	
	//precompute D
	float D = -(u0 * v1) + (u0 * v2) + (u1 * v0) - (u1 * v2) - (u2 * v0) + (u2 * v1);
	
	float V1minV2divD = (v1 - v2) / D;     //precompute commonly used terms 
	float V0minV2divD =	(v0 - v2) / D;
	float V0minV1divD = (v0 - v1) / D;
	
	float U1minU2divD = (u1 - u2) / D;
	float U0minU2divD = (u0 - u2) / D;
	float U0minU1divD = (u0 - u1) / D;
	
	float& x0 = tri->Vertices[0].Pos.x;	//get aliases for vertex positions in 3D
	float& x1 = tri->Vertices[1].Pos.x;
	float& x2 = tri->Vertices[2].Pos.x;
	float& y0 = tri->Vertices[0].Pos.y;
	float& y1 = tri->Vertices[1].Pos.y;
	float& y2 = tri->Vertices[2].Pos.y;
	float& z0 = tri->Vertices[0].Pos.z;
	float& z1 = tri->Vertices[1].Pos.z;
	float& z2 = tri->Vertices[2].Pos.z;
		
	//compute the partial derivatives with respect to texture coordinates
	shade.dudt = Normalize(Vector3D(-V1minV2divD * x0 + V0minV2divD * x1 - V0minV1divD * x2,
									-V1minV2divD * y0 + V0minV2divD * y1 - V0minV1divD * y2,
									-V1minV2divD * z0 + V0minV2divD * z1 - V0minV1divD * z2));
	
	shade.dvdt = Normalize(Vector3D(U1minU2divD * x0 - U0minU2divD * x1 + U0minU1divD * x2,
									U1minU2divD * y0 - U0minU2divD * y1 + U0minU1divD * y2,
									U1minU2divD * z0 - U0minU2divD * z1 + U0minU1divD * z2));
					
	delete tri;
}
