// ===================================================================
//  File: world.cpp			12/24/1999
//  By Konstantin Khanyants	kon@unix.temple.edu
//
//	Description:	
//	World-related routines in ray tracer.
//
// 	Original Copyright:	     The Object-Oriented Ray Tracer (OORT)
//            				 Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#include "stdafx.h"
#include "include/Oort.h"
#include "include/World.h"
#include "../RTWindow.h"
#include "../PreviewPic.h"

// World constructor.
World::World()
{
    m_Ambient = RGBColor(0);
	m_BackColor = RGBColor(1.0f, 0.1f, 0.0f);	//background color
    ViewTransform = IdentityMatrix();		//initialize the transform matrix
	
//  HorzRes = 320;							//horizontal resolution
//  VertRes = 200;							//vertical resolution
//  StartScan = 0;							//scanlines
//  EndScan = 200;								
//  printpers = interactive = 1;			//TAKE OUT PRINTPERS
//  printhierarchy = 0;						//TAKE THIS OUT TOO
//  numtries = 1;							//ONE TRY ONLY
//  hierarchyseed = 0;						//REMAINS THE SAME
    DepthLimit = 3;							//SET TO SIX

    mode = World::None;						//antialiasing mode
    MinSamples = 4;							//for distributed tracing
    MaxSamples = 16;
    NumCandidates = 10;
    ContrastThreshold = RGBColor(0.4f, 0.3f, 0.6f);

	//NEW STUFF 4/5/2000
	m_CameraChanged = FALSE;
	m_MaterialsChanged = FALSE;
	m_bRenderingInProgress = FALSE;
	m_bAutoCenter = FALSE;
	m_LightsChanged = FALSE;
}

World::~World()
{
	Surfaces.DeleteSurfaces();
	Cameras.DeleteCameras();
	Objects.DeleteObjects();
	Lights.DeleteLights();
}

// Set height of screen.  This is the height of the
// virtual screen in 3-space.
void
World::SetScreenHeight(float NewScreenHeight)
{
    ScreenHeight = NewScreenHeight;
}

// Set width of screen.  This is the width of the
// virtual screen in 3-space.
void
World::SetScreenWidth(float NewScreenWidth)
{
    ScreenWidth = NewScreenWidth;
}

// Set screen distance.  This is the distance from
// viewer to screen in 3-space.
void
World::SetScreenDistance(float NewScreenDist)
{
    ScreenDist = NewScreenDist;
}

// Set depth limit on recursion.
void
World::SetDepthLimit(int NewDepthLimit)
{
    DepthLimit = NewDepthLimit;
}

// Set viewer parameters.  ViewerLoc is where the viewer
// looks from; LookAt is the point the viewer looks at;
// and UpVector is the normalized direction of upward
// (traditionally (0.0, 1.0, 0.0).
void
World::SetViewerParameters(Vector3D& NewLookAt,
			   Vector3D& NewViewerLoc,
			   Vector3D& NewUpVector)
{
    ViewerLoc = NewViewerLoc;
    LookAt = NewLookAt;
    UpVector = NewUpVector;
    ViewTransform = ViewMatrix(LookAt, ViewerLoc, UpVector);
}

//added 6/10/2000
// get viewer params from a 3DStudio camera
void
World::SetViewerParamsCamera(Camera* camUser)
{
	ViewerLoc = camUser->m_Pos;

	//if auto centering is turned on
	if (m_bAutoCenter)
		LookAt = m_SceneCenter;
	else
		LookAt = camUser->m_Target;
	
	UpVector = camUser->m_Up;

	//make the view transform matrix
	ViewTransform = ViewMatrix(LookAt, ViewerLoc, UpVector);
//	ViewTransform *= camUser->m_SceneMatrix; APPLY THIS IN THE WORLD

	ScreenDist = camUser->m_Focal_length;
	
	//convert fov from degrees to radians
	float fov = float(camUser->m_Fov * M_PI / 180);
	
	ScreenHeight = 2 * ScreenDist * (float)tan(fov / 2);
	ScreenWidth = HorzRes * ScreenHeight / VertRes;

}


// Set ambient lighting.  Generally a small gray level such
// as (0.1, 0.1, 0.1).
void
World::SetAmbientLight(RGBColor& NewAmbient)
{
    m_Ambient = NewAmbient;
}

//void
//World::SetNumTries(int _numtries)
//{
//    numtries = _numtries;
//}

void
World::SetAntiAliasMode(enum AntiAliasModes _mode)
{
    mode = _mode;
}

void
World::SetSamples(int _MinSamples, int _MaxSamples)
{
    if (_MinSamples == 0) {
	MaxSamples = _MaxSamples;
	return;
    }
    if (_MaxSamples == 0) {
	MinSamples = _MinSamples;
	return;
    }
    if (_MinSamples < _MaxSamples) {
	MinSamples = (_MinSamples < 1) ? 1 : _MinSamples;
	MaxSamples = _MaxSamples;
    }
    else {
	MinSamples = (_MaxSamples < 1) ? 1 : _MaxSamples;
	MaxSamples = _MinSamples;
    }
}

void
World::SetNumCandidates(int _NumCandidates)
{
    NumCandidates = _NumCandidates;
}

void
World::SetContrastThreshold(const RGBColor& _ContrastThreshold)
{
    ContrastThreshold = _ContrastThreshold;
}


// Set background color.  Default is some kind of blue.
void
World::SetBackgroundColor(RGBColor& NewBackground)
{
	m_BackColor = NewBackground;
}

// Set scanline to start tracing.  Default is 0.
//void
//World::SetStart(int Start)
//{
//    StartScan = Start;
//}

// Set scanline to end tracing.  Default is VertRes.
//void
//World::SetEnd(int End)
//{
//    EndScan = End;
//}

// Add a light to the World.
void
World::AddLight(Light *newlight)
{
    Lights.AppendToList(newlight);
}

// Add an instance of an object to the list in the world.
void
World::AddObject(Object3D& Obj)
{
    Objects.AppendToList(Obj.Dup());
}

// Add an object to the list.
void
World::AddObject(Object3D *Obj)
{
    Objects.AppendToList(Obj);
}

void
World::AddCamera(Camera* newCam)
{
	Cameras.AppendToList(newCam);
}

Object3D *
World::Find_Nearest(Ray& ray, float& t)
{
    float mint; 	// Minimum parameter of intersection found thus far
    Object3D *nearest;	// Nearest object found thus far
    float tempt;
    Heap<KayKajiyaNode *> TraversalHeap;

    nearest = 0;
    mint = Limits::Infinity;

    // If intersect with overall bounding volume, place it in heap.
    if (! RTObjects.Empty()) {
 	Object3D *head = *RTObjects.Head();
	if (head->NearestInt(ray, tempt, mint)) {
	    if (head->Flags() == BOUNDING) {
	        KayKajiyaNode *NewNode = new KayKajiyaNode(head, tempt);
	    	TraversalHeap.Insert(NewNode);
	    }
	    else {
	    	if (tempt < mint) {
		    mint = tempt;
		    nearest = head;
	    	}
	    }
    	}
    }

    // Loop to find nearest intersection
    KayKajiyaNode *c;
    while (TraversalHeap.N()) {
		c = TraversalHeap.ExtractMin();

		if (! c) {
//REM	    	cerr << "Null pointer extracted.  Out of memory.\n";
//	    	exit(1);
		}

		// If the nearest bounding volume is farther away than
		// our running minimum intersection, the heap contains no
		// bounding volumes or objects nearer than what we already have.
		if ( (mint < c->t && fabs(mint - c->t) > Limits::Small) &&
	   	   c->t < Limits::Infinity) {
	   		 delete c;
	   	 	 break;
		}

		// Traverse all objects listed in bounding volume
		ObjectList::Iterator sc( ((BoundingVolume *) c->obj)->contents);

		while (sc.Valid()) {
	    	if (sc.Contents()->NearestInt(ray, tempt, mint)) {
	    		// If hit bounding volume, insert into heap.
	    		if (sc.Contents()->Flags() == BOUNDING) {
		   	 		KayKajiyaNode *NewNode = new KayKajiyaNode(sc.Contents(), tempt);
		  	 		TraversalHeap.Insert(NewNode);
	        	}
				// If hit nonbounding object, check if nearest.
				else {
	    		   if (tempt < mint) {
						mint = tempt;
						nearest = sc.Contents();
		    		} // if (tempt < mint)
	        	} // else
	    	} // if (NearestInt)
	    	sc.GotoNext();
    	} // for (each object in bounding volume)
		delete c;
    } // while (heap is not empty)

    // Delete everything remaining in heap
    while (TraversalHeap.N())
	delete TraversalHeap.ExtractMin();

    // Pass back parameter of nearest intersection
    t = mint;

    // Return nearest object intersected, if any
    return nearest;
}

// Function to find the attenuation of a given shadow ray.
// Returns opaque object if one is found (for later caching),
// or 0 if none was found.  The Atten is an RGBColor which
// describes the attenuation (initially (1.0, 1.0, 1.0),
// chipped away by transparent objects on ray).
Object3D *
World::FindAtten(ObjectList &olist,
		 Ray& ray,
		 RGBColor& Atten,
		 float maxt)
{
    ObjectList::Iterator optr(olist);
    Object3D *ret;
    ObjectList trans;		  // List of transparent objects
				  // Used only if no opaque object is found

    // Scan list of objects for opaque or transparent.
    while (optr.Valid()) {
	Object3D *contents = optr.Contents();
	// If object is transparent, save it for later.
	// Intersecting an opaque one will obviate the need to
	// intersect with this one.
	if ((! (contents->flags == BOUNDING)) &&
	    contents->surf &&
	    contents->surf->Transparent()) {
	    trans.AddToList(contents);
	    optr.GotoNext();
	    continue;
	}

	if ((! contents->flags == BOUNDING) && (! contents->surf)) {
	    optr.GotoNext();
	    continue;
	}

	if (contents->HitAtAll(ray, maxt)) {
	    // If intersect with bounding volume, recurse
	    if (contents->flags == BOUNDING) {
		ret = FindAtten( ((BoundingVolume *) contents)->contents,
				     ray, Atten, maxt);
		if (ret)
		    return ret;
	    }
	    // Otherwise, deal with leaf object.
	    else {
		Atten = RGBColor(0);	// Clear Atten
		return contents;
	    }
	}
	// Go on to next object in list
	optr.GotoNext();
    }

    // No opaque object was intersected, go through the list
    // of transparent ones.
    while (trans.ListCount()) {
	Object3D *obj = trans.ExtractHead();
	SpanList *spans = obj->FindAllIntersections(ray);

	if (spans) {
	    // Find first intersection greater than Limits::Threshold
	    while (spans->NumSpans()) {
		Span span = spans->ExtractMinSpan();

		if (span.tmin > Limits::Threshold && span.tmin < maxt) {
		    if (span.omin->surf && span.omin->surf->Transparent())
	        	Atten *= span.omin->surf->GetTransmitted();//ray.Extrap(span.tmin));
		    else {
			Atten = RGBColor(0);
			return span.omin;
		    }
		}
		if (span.tmax > Limits::Threshold && span.tmax < maxt) {
		    if (span.omax->surf && span.omax->surf->Transparent())
			Atten *= span.omax->surf->GetTransmitted();//ray.Extrap(span.tmax));
		    else {
			Atten = RGBColor(0);
			return span.omax;
		    }
		}
	    }
	    delete spans;
	}
    }

    // No opaque object found, return 0
    return 0;
}

Object3D *
World::ShadowAtten(Ray& ray, RGBColor& Atten, float maxt)
{
	return FindAtten(RTObjects, ray, Atten, maxt);
}

// Compute diffuse and specular contributions to an intersection's
// color.  These are computed together because they are both directly
// due to to light sources' light.
RGBColor
World::DiffuseLight(ShadingInfo& shade, Surface *surf)
{
    RGBColor Contrib(0);
    for (LightList::Iterator sc(Lights); sc.Valid(); sc.GotoNext()) 
		Contrib += sc.Contents()->ShadowLight(shade, surf);
    
    return Contrib;
}

// Find color of a given ray.  Take and pass back pointer to
// pointer to nearest object found for subsequent caching.
RGBColor
World::FindColor(Ray& ray, int depth)
{
    Object3D *obj;
    float t;

    if (obj = Find_Nearest(ray, t))
	return obj->ShadeObject(*this, ray, ray.Extrap(t), depth);
    else
	return m_BackColor;  //was Background
}


//TAKE THIS OUT
/*void
World::DrawScreen(long beg, int Col, int Row)
{
    long now;
    clrscr();
    cout << "OORT v1.0 in progress:\n\n";

//    cout << heapsize() << " bytes left\n";
    cout << "Primary rays traced: " << Statistics::PrimaryRays << '\n';

    long pixels = (long) HorzRes * Row + Col;
    cout << "Pixels computed: " << pixels << " / " << NumPixels << '\n';
    time(&now);
    float avg;
    cout << "Seconds elapsed rendering: " << now - beg << '\n';

    if (Statistics::PrimaryRays)
    	avg = (float) (now - beg) / Statistics::PrimaryRays;
    else
    	avg = 0;
    cout << "Average time per primary ray (s): " << avg << " s\n";

    if (pixels) {
	avg = (float) (now - beg) / pixels;
    }
    else
        avg = 0;
    cout << "Average time per pixel (s): " << avg << " s\n";
    float raysperpix = (pixels) ? (float) Statistics::PrimaryRays / pixels : 0;
    cout << "Average rays per pixel: " << raysperpix << '\n';
    cout << "Estimated remaining time: "
	     << avg * (NumPixels - pixels) << " s\n";
}*/

Vector3D
World::NewSample(Vector3D *samples, int n)
{
    Vector3D ret(RandomValue(0, 1), RandomValue(0, 1), 0);
    if (! n)
	return ret;

    float mindist = MinDist(ret, samples, n);
    for (int i = 0; i < NumCandidates; i++) {
	Vector3D candidate(RandomValue(0, 1), RandomValue(0, 1), 0);
	float dist = MinDist(candidate, samples, n);
	if (dist > mindist) {
	    ret = candidate;
	    mindist = dist;
	}
    }
    return ret;
}

RGBColor
World::PixelColor(float xtarg, float ytarg)
{
    RGBColor ret;
	Vector3D target;

    switch (mode) {
	case None:
	    target = Vector3D(xtarg + ScreenWidth / HorzRes / 2,
			    ytarg + ScreenHeight / VertRes / 2,
			    -ScreenDist);
	    Statistics::PrimaryRays += 1;
	    ret = FindColor(Ray(Vector3D(0), target), 0);
	    break;
	case Adaptive:
	    {
		RGBColor Imin(MAXFLOAT), Imax(-MAXFLOAT), Intens(0);
		Vector3D *pts = new Vector3D[MaxSamples];
		int i;

		for (i = 0; i < MaxSamples; i++) {
		    pts[i] = NewSample(pts, i);

		    target = pts[i];
		    target *= Vector3D(ScreenWidth / HorzRes, ScreenHeight / VertRes, 0);
		    target += Vector3D(xtarg, ytarg, -ScreenDist);

		    Statistics::PrimaryRays += 1;
		    RGBColor clr = FindColor(Ray(Vector3D(0), target), 0);
		    Intens += clr;
		    clr = NormalizeColor(clr);
		    Minimize(Imin, clr);
		    Maximize(Imax, clr);

		    if (i >= MinSamples) {
			if (i >= MaxSamples)
			    break;
			if (! MoreRays(Imin, Imax, ContrastThreshold))
			    break;
		    }

		}
		delete[] pts;
		Statistics::RayPixHist[i - 1] += 1;
		ret = Intens / float((i + 1));
	    }
    }
    return ret;
}

//disbands aggregate mesh objects
ObjectList
World::DisbandAggregates(ObjectList& list)
{
    ObjectList ret;

    while (list.ListCount()) {
		Object3D *obj = list.ExtractHead();
    	if (obj->Flags() == AGGREGATE) {
    	    ret.AddToList(DisbandAggregates(((Aggregate *) obj)->List()));
		}
    	else
    	    ret.AddToList(obj);
	}
    return ret;
}


void
World::GetRTObjects(ObjectList& objs)
{
	for (ObjectList::Iterator sc(Objects); sc.Valid(); sc.GotoNext())
		objs.AppendToList(sc.Contents()->Dup());
}

void
World::ClearRTObjects(ObjectList& objs)
{
	//delete all of the objects
	while (objs.ListCount()) {
		Object3D *obj = objs.ExtractHead();

		//if the object is bounding then recurse
		if (obj->Flags() == BOUNDING) {
			ClearRTObjects(((BoundingVolume *) obj)->contents);
			delete obj;
		}
		//otherwise just delete it
		else
			delete obj;
	}
}

// Trace rays, send to output window image.
// pRTWindow is the output window passed in from the View
void
World::RayTrace(CRTWindow* pRTWindow)
{
    float x, y, incx, incy;
    int ix, iy;
	RGBColor pixelColor(0.0);

	//set the thread rendering flag 
	m_bRenderingInProgress = TRUE;

	ImageFile* pImage = pRTWindow->GetImageFile();
	HorzRes = pImage->Width();
	VertRes = pImage->Height();
	
	//set the viewind propertis to the current camera model
	SetViewerParamsCamera(m_pCamera);

 // 	NumPixels = ((long) EndScan - StartScan) * HorzRes;
  
//    long beg, end;	  // Time in seconds

    // Get the time before beginning trace so we can report ET
//    time(&beg);

	GetRTObjects(RTObjects);

	RTObjects.TransformList(m_pCamera->m_SceneMatrix);
    // Transform the object database according to the view transform
    RTObjects.TransformList(ViewTransform);


	Surfaces.TransformSurfaces(Invert(ViewTransform));
    RTObjects = DisbandAggregates(RTObjects);


//rem    if (hierarchyseed)
//rem	srand(hierarchyseed);
//rem    else
//rem	randomize();
    BuildHierarchy(1); /////\\\\\numtries); //NO NUMTRIES only once

    if (RTObjects.Head())
        CullSingletons(*RTObjects.Head(), &RTObjects);

//NO THANKS
//rem    if (printhierarchy) {
//        cout << "Hierarchy built.  It is as follows:\n";

    CountObjects();

    // Transform the lights according to the view transform
    for (LightList::Iterator lightsc(Lights); lightsc.Valid(); lightsc.GotoNext())
        lightsc.Contents()->ApplyTransform(ViewTransform);

    Statistics::HierarchyHeuristic = HierarchyCost(RTObjects);

//   time(&end);
//   Statistics::PrepTime = end - beg;
//   beg = end;

    // Begin trace.
    y = ScreenHeight / 2;
    incy = - ScreenHeight / VertRes;
//  y += incy * StartScan;

    incx = ScreenWidth / HorzRes;

	//TEMPORARY
//	m_BackColor = RGBColor(0.0f, 0.0f, 1.0f);

	try {
    
		for (iy = 0; iy < VertRes; iy++, y += incy) {

			//Get ImageFile Dib here and blit it into the view
	
			x = -ScreenWidth / 2;
			for (ix = 0; ix < HorzRes; ix++, x += incx) {
				pixelColor = PixelColor(x, y);
			
				if (m_bRenderingInProgress)
					//save the pixel in the image
					pImage->PutRGBColorPixel(pixelColor, ix, iy);
				//put a stop to rendering.
				else
					throw new CException;
			}

			//display the newly processed image row in the window
			//update display every five pixel rows or if it's the end
			if ((iy % 5) == 0 || iy == VertRes - 1)
				pRTWindow->UpdateImage();
		}
	}
	//catch the exception caused by termination of the RT Window
	catch (CException* e) {
		e->Delete();
	}	

	//since we duplicated all of the objects we can now delete them
	ClearRTObjects(RTObjects);

	//transform the surfaces back
	Surfaces.TransformSurfaces(ViewTransform);
	
	//transform the lights back
    for (LightList::Iterator lightsc2(Lights); lightsc2.Valid(); lightsc2.GotoNext()) {
        lightsc2.Contents()->ApplyTransform(Invert(ViewTransform));
		lightsc2.Contents()->Reset(); //reset the last opaque object optimization
	}
//    time(&end);
//    Statistics::RenderTime = end - beg;

	//set the thread rendering flag to false 
	m_bRenderingInProgress = FALSE;
	//end the current rendering process
	pImage->EndCurrentRender();
}


// Trace rays, send to output window image.
// pRTWindow is the output window passed in from the View
void
World::RayTracePreview(CPreviewPic* previewPic)
{
    float x, y, incx, incy;
    int ix, iy;
	RGBColor pixelColor(0.0);
	
	//set the thread rendering flag 
	m_bRenderingInProgress = TRUE;

	//two levels of refract/reflect
	DepthLimit = 2;

	//set the resolution
	HorzRes = previewPic->Width();
	VertRes = previewPic->Height();

	//set the viewind propertis to the current camera model
	SetViewerParamsCamera(Cameras[0]);

	GetRTObjects(RTObjects);
	
    // Transform the object database according to the view transform
    RTObjects.TransformList(ViewTransform);

	Surfaces.TransformSurfaces(Invert(ViewTransform));
//	RTObjects = DisbandAggregates(RTObjects);//

//    BuildHierarchy(1); //NO NUMTRIES only once//

//    if (RTObjects.Head())//
//		CullSingletons(*RTObjects.Head(), &RTObjects);//

    CountObjects();

    // Transform the lights according to the view transform
    for (LightList::Iterator lightsc(Lights); lightsc.Valid(); lightsc.GotoNext())
        lightsc.Contents()->ApplyTransform(ViewTransform);

    // Begin trace.
    y = ScreenHeight / 2;
    incy = - ScreenHeight / VertRes;
    incx = ScreenWidth / HorzRes;

	m_BackColor = RGBColor(0.0f, 0.0f, 1.0f);
///	m_Ambient = RGBColor(0.5f, 0.5f, 0.5f);

	unsigned char* pBuffer = new unsigned char[HorzRes * RGBSample];

    for (iy = 0; iy < VertRes; iy++, y += incy) {
	
		x = -ScreenWidth / 2;
		int PixelX = 0;
		for (ix = 0; ix < HorzRes * RGBSample; ix+=3, x += incx) {
			pixelColor = PixelColor(x, y);

			//save the pixel in pBuffer			
			pBuffer[ix + 0]	= unsigned char(pixelColor.z * 255.0);
			pBuffer[ix + 1] = unsigned char(pixelColor.y * 255.0);
			pBuffer[ix + 2] = unsigned char(pixelColor.x * 255.0);
		}
		
		previewPic->AddScanlineToDib(iy, pBuffer);

		//display the newly processed image row in the window
		//update display every five pixel rows and in the final iteration
		if ((iy % 5) == 0 || iy == VertRes - 1)
			previewPic->Invalidate(FALSE);  //do not erase background
    }

	if (pBuffer)
		delete [] pBuffer;

	//since we duplicated all of the objects we can now delete them
	ClearRTObjects(RTObjects);

	//transform the surfaces back
	Surfaces.TransformSurfaces(ViewTransform);
	
	//transform the lights back
    for (LightList::Iterator lightsc2(Lights); lightsc2.Valid(); lightsc2.GotoNext()) {
        lightsc2.Contents()->ApplyTransform(Invert(ViewTransform));
		lightsc2.Contents()->Reset(); //reset the last opaque object optimization
	}
	
	//set the thread rendering flag to false
	m_bRenderingInProgress = FALSE;
	//sent an end of render message to the Material Dialog so that 
	// certain buttons can be enabled
	previewPic->GetParent()->SendMessage(WM_USER_PREVIEW_FINISHED, 0, 0);
}


