// ===================================================================
//  File: bump.h						12/26/1999
//	by Konstantin Khanyants				kon@unix.temple.edu
//	Description:
//		Header file for bump-mapping implementations.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#ifndef __BUMP__

#define __BUMP__

// Forward-declare ShadingInfo class.
class ShadingInfo;

// =========================================================
// BumpMap
//	Takes the ShadingInfo for a point--direction of the 
//	incident ray, intersection point, normal and other
//	information--and modifies the intersection point or 
//	the normal, or both.
// =========================================================

class BumpMap {
public:
    virtual ~BumpMap() { }
    virtual void PerturbNormal(ShadingInfo& shade) = 0;
    virtual BumpMap *Dup() const = 0;
    virtual void ApplyTransform(const Matrix& tform) { }
    virtual void PreMulTransform(const Matrix& tform) { }
};

// =========================================================
// BumpMapReference
//	Bump map with a reference matrix attached.
//	Many bump maps, such as ReverseNormal or TwoSided,
//	do not need a reference matrix at all.
// =========================================================
class BumpMapReference : public BumpMap {
public:
    BumpMapReference(const Matrix& _reference = IdentityMatrix());
    virtual ~BumpMapReference() { }

    virtual void PerturbNormal(ShadingInfo& shade) = 0;
    virtual BumpMap *Dup() const = 0;
    virtual void ApplyTransform(const Matrix& tform);
    virtual void PreMulTransform(const Matrix& tform);
protected:
    Matrix reference;

    virtual Vector3D PreprocessVector(ShadingInfo& shade);
};

// =========================================================
// ReverseNormal
//	Reverses the direction of the normal.
// =========================================================
class ReverseNormal : public BumpMap {
public:
    ReverseNormal(BumpMap *a = 0) { bump = a; }
    void PerturbNormal(ShadingInfo& shade);
    virtual BumpMap *Dup() const { return new ReverseNormal(bump); }
private:
    BumpMap *bump;
};


// =========================================================
// TwoSided bump map
//	Ensures the normal is facing the incident direction.
//	If necessary, the normal is negated.
// =========================================================
class TwoSided : public BumpMap {
public:
    TwoSided(BumpMap *a = 0) { bump = a; }
    void PerturbNormal(ShadingInfo& shade);
    virtual BumpMap *Dup() const { return new TwoSided(bump); }
private:
    BumpMap *bump;
};


// ---------------------------------------------------------
// ImageBumpMap: similar to ImageMap of TextureMapping
// ---------------------------------------------------------
class ImageBumpMap : public BumpMapReference {
public:
	ImageBumpMap(const Matrix& ref, int Once, const char *filename, float rot = 0.0, float bumpamt = .5, float xscale = 1, float yscale = 1);
	ImageBumpMap(const Matrix& ref, int Once, ImageFile *file, float rot = 0.0, float bumpamt = .5, float xscale = 1, float yscale = 1);
	ImageBumpMap(const ImageBumpMap& initBump);

	virtual ~ImageBumpMap();
	virtual void PerturbNormal(ShadingInfo& shade);
	virtual BumpMap* Dup() const;
	virtual ImageFile* GetImageFile() { return file; }

	//added for changing the image file from the Materials dialog
	virtual void SetImageFile(char* filename);
	virtual void SetImageFile(ImageFile* newFile);

	virtual void SetAmount(float val) { bump_amt = val; }
	virtual float GetAmount() { return bump_amt; }

protected:
    int once;			// Boolean--if true, the image is mapped once rather than tiled.
    ImageFile *file;	// Image to map
    Vector3D scale;		// Scale by this vector after mapping.
    float rot_angle;	// rotation from 3D Studio map definition, in radians
    float bump_amt;		// bump amount from 3D Studio definition 0.0 ~ 3.0
	
	virtual Vector3D PreprocessVector(ShadingInfo& shade);

	friend class CMaterialDialog;
};


// =========================================================
// WaveSource
//	Precomputes a set of noise vector values.
// =========================================================
class WaveSource {
public:
    WaveSource(int, NoiseMaker *);

    int N() const { return n; }
    Vector3D GetWave(int inx) const { return waves[inx % n]; }
private:
    int n;
    Vector3D *waves;
};

// =========================================================
// FreqSource
//	Precomputes a set of noise values.
// =========================================================
class FreqSource {
public:
    FreqSource(int, NoiseMaker *);

    int N() const { return n; }
    float GetFreq(int inx) const { return freqs[inx % n]; }
private:
    int n;
    float *freqs;
};


// =========================================================
// Ripples
//	Gives the surface a rippled appearance.
// =========================================================
class Ripples : public BumpMapReference {
public:
    Ripples(WaveSource *WS,
	    FreqSource *FS,
	    float Freq,
	    float Phase,
	    float BumpAmt,
	    const Matrix& _reference): BumpMapReference(_reference) {
	  ws = WS;
	  fs = FS;
	  freq = Freq;
	  phase = Phase;
	  bumpamt = BumpAmt;
      }
    void PerturbNormal(ShadingInfo& shade);
    virtual BumpMap *Dup() const;
private:
    WaveSource *ws;
    FreqSource *fs;
    float freq, phase, bumpamt;
};

// =========================================================
// Dented
//	Gives the surface a dented appearance by
//	modifying the intersection point.
//	An experiment in displacement mapping.
//	From Upstill's _The Renderman Companion_.
// =========================================================
class Dented : public BumpMapReference {
public:
    Dented(float, const Matrix&);
    void PerturbNormal(ShadingInfo& shade);
    virtual BumpMap *Dup() const;
private:
    float Km;
};

#endif
