// ===================================================================
//  File: light.h					12/26/1999
//	By Konstantin Khanyants			kon@unix.temple.edu
//	Description:
//		Header file for Light abstract class and its derivatives.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#ifndef _LIGHT_H_

#define _LIGHT_H_

#define	POINT_LIGHT			1
#define SPOT_LIGHT			2
#define	DIRECTIONAL_LIGHT	3

// -------------------------------------------------------------------
// Light
//	Abstract class encapsulates all types of light source.
//	Color is the color of light emitted by the light source.
//	c1, c2 and c3 define the attention of the light due to
//	distance.  The attenuation is 1 / (c1 + c2*d + c3*d^2) where d
//	is the distance from the light source.	Thus, c1 == 1 means no
//	attenuation, while if c3 is nonzero the attenuation is
//	proportional to the square of the distance (the physically
//	correct model, which does not work well in practice).
// -------------------------------------------------------------------
class Light {
public:
    Light(RGBColor& clr, float C1, float C2, float C3, BOOL on = TRUE);

	void SetName(CString newName);

    virtual float DistAtten(float t);	  // Return attenuation due to distance
	virtual ~Light() { }
    // -----------------------------------------------------------------
    // Return nonzero if the given ray hits LastOpaque.  If it misses,
    // LastOpaque is reset to 0.
    // -----------------------------------------------------------------
    int HitCachedObject(Ray& ray, float maxt);

    // -----------------------------------------------------------------
    // Fires a collection of rays according to the type of light source
    // and number of rays authorized by the user.  Returns the color of
    // the light source as seen from the intersection point.
    // -----------------------------------------------------------------
    virtual RGBColor ShadowLight(ShadingInfo& shade, Surface *surf) = 0;
    virtual void ApplyTransform(const Matrix& tform) = 0;
	virtual BOOL Enabled() { return enabled; }
	virtual void SetRange(float range_near, float range_far, BOOL on) { }
	virtual void Describe() { }
	virtual void Reset() { LastOpaque = NULL; }
	
	virtual void SetupLightGL() { }
	virtual void DisableLightGL() { }

protected:
	CString	 name;
	int		 type;      // Point, Spot, Directional
	BOOL	 enabled;	// light on or off
    RGBColor Color;		// Color of light's light
    float c1, c2, c3;		// Light's attenuation by distance
    Object3D *LastOpaque;	// Cache for last object which
				// shadowed this light

    RGBColor ShadowRay(ShadingInfo& shade,
		       Vector3D& target,
		       Surface *surf);

	friend class LightList;
};

// -------------------------------------------------------------------
// PointLight
//	Encapsulates a point light source.
//	One shadow ray is fired at the location, and if it hits an
//	opaque object between the intersection point and the point,
//	the light is in shadow.
// -------------------------------------------------------------------
class PointLight : public Light {
public:
    PointLight(Vector3D& loc, RGBColor& clr, float C1 = 1, float C2 = 0, float C3 = 0, BOOL on = TRUE);
    virtual ~PointLight() { DisableLightGL(); }
	virtual void SetRange(float range_near, float range_far, BOOL on);
	virtual RGBColor ShadowLight(ShadingInfo& shade, Surface *surf);
    virtual void ApplyTransform(const Matrix& tform);

	//OpenGL's light source implementation
	virtual void SetupLightGL();
	virtual void DisableLightGL();
	virtual void Describe();
protected:
    Vector3D Location;		// Where light is coming from
	float	 Near_range;	//attenuation factors only supported in Ray Tracer
	float    Far_range;
	BOOL	 Attn_on;

	//automatically assigned in SetupLightGL whenever you want
	// to turn a light on in OpenGL
	Vector4f locationGL; //doesn't change from ray tracing
	GLuint	 lightGL;	 	 
};

// -------------------------------------------------------------------
// SpotLight added 3/14/2000
//	Encapsulates a spot light source.
//	One shadow ray is fired at the location, and if it hits an
//	opaque object between the intersection point and the point,
//	the light is in shadow.
// -------------------------------------------------------------------
class SpotLight : public PointLight {
public:
	SpotLight(Vector3D& loc, Vector3D& target, float& hotspot,
		float& falloff, RGBColor& clr, 
		float C1 = 1, float C2 = 0, float C3 = 0, BOOL on = TRUE);
	virtual ~SpotLight() { DisableLightGL(); }
	RGBColor ShadowLight(ShadingInfo& shade, Surface *surf);
    void ApplyTransform(const Matrix& tform);

	//opengl overridable from point light source
	virtual void SetupLightGL();
	virtual void Describe();

protected:
	Vector3D Target;
	float	 Hotspot; //hotspot angle in rad
	float    Falloff; //falloff angle

	Vector4f targetGL;	//this point doesn't change with ray tracing
};


// -------------------------------------------------------------------
// DirectLight
//	Encapsulates a directional light source.  Light comes from
//	a fixed direction (the direction vector points _toward_ the
//	direction the light is coming _from_).	Directional light
//	sources do not attenuate their light by distance, since they
//	are infinitely far away.
// -------------------------------------------------------------------
class DirectLight : public Light {
public:
    DirectLight(const Vector3D& loc, RGBColor& clr, BOOL on = TRUE);
	virtual ~DirectLight() { }

    RGBColor ShadowLight(ShadingInfo& shade, Surface *surf);
    void ApplyTransform(const Matrix& tform);
	virtual void Describe();
	
	virtual void SetupLightGL() { }
	virtual void DisableLightGL() { }
protected:
    Vector3D Direction;		  // Where light is coming from
};


// ---------------------------------------------------------
// Derive list of lights from SimpleList class template.
// ---------------------------------------------------------
class LightList : public SimpleList<Light *> {
public:
	LightList(): SimpleList<Light *>() { }
	virtual ~LightList() { }//DeleteLights(); }
	Light* FindLightByName(CString name);
	virtual void   SetupLightsGL();
	virtual void   DeleteLights();
	virtual Light *operator[] (int inx);
};

#endif