// ===============================================================
//  File: mapping.h						12/26/1999
//	By Konstantin Khanyants
//	Description:
//      Header file for inverse mappings.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===============================================================


#ifndef _MAPPING_

#define _MAPPING_

//#include "Oort.h"
class Object3D;

#define		CYLINDRICAL_MAP		1 	
#define 	PROJECTION_MAP		2
#define 	QUADRILATERAL_MAP	3
#define 	SPHERICAL_MAP		4
#define		UV_MAP				5

// ---------------------------------------------------------------
// Mapping 
//	Implements various methods of taking 3D points and 
//	turning them into 2D points.
// ---------------------------------------------------------------

class Mapping {
public:
    Mapping() { }
    virtual ~Mapping() { }
    virtual Mapping *Dup() const = 0;
    virtual Vector3D Map3DTo2D(Object3D *obj, const Vector3D& v) const = 0;
};

// ---------------------------------------------------------------
// CylindricalMapping
//	Maps onto a cylinder centered on the Y axis.
// ---------------------------------------------------------------
class CylindricalMapping : public Mapping {
public:
    CylindricalMapping(float radius, float height);
    Mapping *Dup() const;
    Vector3D Map3DTo2D(Object3D *obj, const Vector3D& v) const;
private:
    float radius, height;
};


// ---------------------------------------------------------------
// ProjectionMapping
//	Projects the input vector onto the XY coordinate plane.
// ---------------------------------------------------------------
class ProjectionMapping : public Mapping {
public:
    Mapping *Dup() const;
    Vector3D Map3DTo2D(Object3D *obj, const Vector3D& v) const;
};


// ---------------------------------------------------------------
// QuadrilateralMapping
//	Maps points onto a quadrilateral.
// ---------------------------------------------------------------
class QuadrilateralMapping : public Mapping {
public:
    QuadrilateralMapping(const Vector3D *Pts);
    Mapping *Dup() const;
    Vector3D Map3DTo2D(Object3D *obj, const Vector3D& v) const;
private:
    Vector3D pts[4];
    Vector3D Na, Nb, Nc, Nd;
    float Du0, Du1, Du2, Dv0, Dv1, Dv2;
    int uaxespar, vaxespar;
    Vector3D Qux, Quy, Qvx, Qvy;
    float Dux, Duy, Dvx, Dvy;
};


// ---------------------------------------------------------------
// SphericalMapping
//	Maps points onto a sphere centered about an arbitrary point.
// ---------------------------------------------------------------
class SphericalMapping : public Mapping {
public:
    SphericalMapping(Vector3D _center = Vector3D(0));
    Mapping *Dup() const;
    Vector3D Map3DTo2D(Object3D *obj, const Vector3D& v) const;
private:
    Vector3D center;
};


// ---------------------------------------------------------------
// UV Mapping
//	Interpolates between three uv vertices, these are defined 
//  in the triagle object
// ---------------------------------------------------------------
class UVMapping : public Mapping {
public:
    Mapping *Dup() const;
    Vector3D Map3DTo2D(Object3D *obj, const Vector3D& v) const;
};

#endif