// ===================================================================
// 	File: object3d.h			12/25/1999
//	By Konstantin Khanyants
//	Description:
//		Definitions required to implement the Object3D abstract class
//		and the hierarchy rooted by it.
//
//	Original Copyright:    	The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#ifndef _OBJECT3D_

#define _OBJECT3D_

// Forward-declare some classes.
class Object3D;
class AxisAlignedBox;
class Surface;
class ShadingInfo;
class World;
class Ray;
class SpanList;
class Vertex;

// ----------------------------------------------------------------
// SurfaceList
//	List of colors; derives from SimpleList and adds a function
//	to search for a given color and return nonzero if there.
// ----------------------------------------------------------------
class SurfaceList : public SimpleList<HallSurface *> {
public:
	SurfaceList(): SimpleList<HallSurface *>() { }
	virtual ~SurfaceList() { }//DeleteSurfaces(); }
    int SurfaceThere(HallSurface *probe);
	HallSurface* FindSurfaceByName(CString surfName);
	void TransformSurfaces(const Matrix& tform);
	virtual HallSurface *operator[] (int inx);
	virtual void DeleteSurfaces();
};

class ObjectList : public SimpleList<Object3D *> {
public:
    ObjectList(): SimpleList<Object3D *>() { }
    ObjectList(Object3D * & x): SimpleList<Object3D *>(x) { }
    virtual ~ObjectList() { }//DeleteObjects(); }
	void TransformList(const Matrix& tform);
    int SingleObject() { return n == 1; }
	virtual Object3D *operator[] (int inx);
	virtual void DeleteObjects();
};


// --------------------------------------------------------
// AxisAlignedBox class.
//	This is a box in 3-space, defined by min and max
//	corner vectors.  Many useful operations are defined
//	on this
// --------------------------------------------------------

class AxisAlignedBox {
public:
    // Constructors.
    AxisAlignedBox() { }
    AxisAlignedBox(const Vector3D& _min, const Vector3D& _max): minV(_min), maxV(_max) { }

    // Read-only access to min and max vectors.
    Vector3D MinV() const;
    Vector3D MaxV() const;

    // Return whether the box is unbounded.  Unbounded boxes appear
    // when unbounded objects such as quadric surfaces are included.
    int Unbounded() const;

    // Expand the axis-aligned box to include the given object.
    void Include(const Vector3D& newpt);
    void Include(const AxisAlignedBox& bbox);

    // Overlap returns 1 if the two axis-aligned boxes overlap.
    friend int Overlap(const AxisAlignedBox&, const AxisAlignedBox&);

    // Returns the smallest axis-aligned box that includes all points
    // inside the two given boxes.
    friend AxisAlignedBox Union(const AxisAlignedBox& x, const AxisAlignedBox& y);

    // Returns the intersection of two axis-aligned boxes.
    friend AxisAlignedBox Intersect(const AxisAlignedBox& x, const AxisAlignedBox& y);

    friend AxisAlignedBox Transform(const AxisAlignedBox& box, const Matrix& tform);

    virtual float SurfaceArea() const;
    int ComputeMinMaxT(Ray& ray, float *tmin, float *tmax) const;

    friend class Ray;

protected:
    Vector3D minV, maxV;
};

inline int 
Overlap(const AxisAlignedBox& x, const AxisAlignedBox& y)
{
    if (x.maxV.x < y.minV.x ||
	x.minV.x > y.maxV.x ||
	x.maxV.y < y.minV.y ||
	x.minV.y > y.maxV.y ||
	x.maxV.z < y.minV.z ||
	x.minV.z > y.maxV.z) {
	return 0;
    }
    return 1;
}

inline AxisAlignedBox
Intersect(const AxisAlignedBox& x, const AxisAlignedBox& y)
{
    if (x.Unbounded())
	return y;
    else if (y.Unbounded())
	return x;
    AxisAlignedBox ret = x;
    if (Overlap(ret, y)) {
	Maximize(ret.minV, y.minV);
	Minimize(ret.maxV, y.maxV);
	return ret;
    }
    else	    // Null intersection.
	return AxisAlignedBox(Vector3D(0), Vector3D(0));
}

inline AxisAlignedBox
Union(const AxisAlignedBox& x, const AxisAlignedBox& y)
{
    Vector3D _min = x.minV;
    Vector3D _max = x.maxV;
    Minimize(_min, y.minV);
    Maximize(_max, y.maxV);
    return AxisAlignedBox(_min, _max);
}

inline AxisAlignedBox 
Transform(const AxisAlignedBox& box, const Matrix& tform)
{
    AxisAlignedBox ret(Vector3D(MAXFLOAT), Vector3D(-MAXFLOAT));
    ret.Include(tform * Vector3D(box.minV.x, box.minV.y, box.minV.z));
    ret.Include(tform * Vector3D(box.minV.x, box.minV.y, box.maxV.z));
    ret.Include(tform * Vector3D(box.minV.x, box.maxV.y, box.minV.z));
    ret.Include(tform * Vector3D(box.minV.x, box.maxV.y, box.maxV.z));
    ret.Include(tform * Vector3D(box.maxV.x, box.minV.y, box.minV.z));
    ret.Include(tform * Vector3D(box.maxV.x, box.minV.y, box.maxV.z));
    ret.Include(tform * Vector3D(box.maxV.x, box.maxV.y, box.minV.z));
    ret.Include(tform * Vector3D(box.maxV.x, box.maxV.y, box.maxV.z));
    return ret;
}


// Flags that can be attached to Object3D's.
#define BOUNDING	1
#define AGGREGATE	2
#define TRIANGLE	3
#define SPHERE		4

class Object3D {
public:
    // Constructors set flags and color.
    Object3D(int F = 0): mapping(0), surf(0), flags(F) { } //mapping added 6/12/2000
    Object3D(HallSurface *NewColor, int F, Mapping *NewMapping = NULL):
    	mapping(NewMapping), surf(NewColor), flags(F) { }

    // Make destructor explicitly virtual so it can be overloaded.
		virtual ~Object3D() {}

    HallSurface *ReturnColor() const { return surf; }
    int Flags() const { return flags; }
    void OverrideColor(HallSurface *NewColor) { surf = NewColor; }

    // NearestInt returns 1 if the ray hits the object; it passes back
    // the parameter of intersection in t.  maxt specifies the maximum
    // allowable value for the parameter of intersection; if the only 
    // intersections with the object are of greater parameter than maxt,
    // no intersection is returned.
    virtual int NearestInt(Ray& ray, float& t, float maxt) { return 0; };

    // FindAllIntersections returns a pointer to a SpanList that
    // contains all the spans where the ray passes through the
    // object.  This is primarily used for CSG.
    virtual SpanList *FindAllIntersections(Ray& ray) { return NULL; };

    // Returns the normal to the object at the given intersection point.
    virtual Vector3D FindNormal(const Vector3D& p) { return Vector3D(0); };

    // Returns the sum of the object's ambient, diffuse, specular,
    // reflective, and transparent components.
    //
    // This was unneeded (the World class knew how to do this stuff
    // in FindColor) until hypertexture objects came along.  Since
    // these objects get shaded quite a bit differently from other
    // objects, ShadeObject has a default implementation for all
    // objects that gets overloaded by the SoftObject (hypertexture)
    // class.
    virtual RGBColor ShadeObject(World&, Ray&, Vector3D&, int);

    // Returns the number of children the object has.  This is defined
    // to be 0, except for bounding volumes.
    virtual int NumChildren() { return 0; }

    // Apply an arbitrary 3D 4x4 matrix to the object.
    virtual void ApplyTransform(const Matrix& tform) { };

    // Return a duplicate of the object.
    virtual Object3D *Dup() const { return NULL; };

    // Return nonzero if the given point is inside the object, 0 if not.
    virtual int IsInside(Vector3D& v);

    // Return nonzero if the given ray intersects the object with
    // parameter of intersection less than maxt.  This is frequently
    // more efficient to implement than find-nearest-intersection.
    virtual int HitAtAll(Ray& ray, float maxt);

    // Return a measure of how far inside the object the given point is.
    // This distance is 0 on the object's surface, negative if inside the
    // object, or positive if outside the object.  If the number is
    // positive or negative, it should reflect how far inside or outside
    // the object the given point is.

    // This function is used strictly by the SoftObject hypertexture class.
    virtual float PtDistance(Vector3D& v) { return 0; };

    // Returns bounding box of object.	Unbounded by default.
    virtual AxisAlignedBox BBox() const {
    	return AxisAlignedBox(Vector3D(-MAXFLOAT), Vector3D(MAXFLOAT));
    }

    virtual void CountMe() const { };
    
    //Returns the object's surface
    virtual HallSurface* GetSurface() { return surf; }
	virtual Mapping* GetMapping() { return mapping; }

	// This function is used in bump mapping calculations to compute
	// partial derivatives of the surface being shaded. It then fills 
	// in the shade object variables dudt and dvdt 
	virtual void GetDerivatives(ShadingInfo& shade) { }

	//Describes the object
	virtual void Describe() { }

    friend class World;
protected:
    HallSurface *surf;	// Object surface.
    Mapping *mapping;	//this used to be in texture 6/12/2000
    int flags;		// Flags: see just above for #define's
};

// -------------------------------------------------------------------
// BoundingVolume
//	This abstract class is derived from Object3D.  It contains a
//	list of other objects (those contained in the bounding
//	volume).  Member functions to manipulate the list are
//	provided.
// -------------------------------------------------------------------
class BoundingVolume: public Object3D {
public:
    BoundingVolume() : Object3D(BOUNDING) { }
    virtual int NumChildren() { return contents.ListCount(); }
    virtual Vector3D FindNormal(const Vector3D& intersection) = 0;
    virtual void ApplyTransform(const Matrix& tform) = 0;
    virtual SpanList *FindAllIntersections(Ray& ray) = 0;
    virtual Object3D *Dup() const = 0;
    virtual float PtDistance(Vector3D& v) = 0;
    virtual AxisAlignedBox BBox() const = 0;
    virtual void CountMe() const = 0;
	virtual void Describe() { }

    virtual float SurfaceArea() const = 0;

    virtual void AddObject(const Object3D& NewObj); // add new instance of object
    virtual void AddObject(Object3D *NewObj);	    // put object in list
    virtual ObjectList *List() { return &contents; }

    friend class World;

protected:
    ObjectList contents;	// Contents of bounding volume

};


// -------------------------------------------------------------------
// BoundingBox
//	Derived multiply from BoundingVolume and AxisAlignedBox, this
//	class encapsulates the axis-aligned bounding boxes used by
//	OORT.
// -------------------------------------------------------------------
class BoundingBox : public BoundingVolume, public AxisAlignedBox {
public:
    BoundingBox(): BoundingVolume(),
	AxisAlignedBox(Vector3D(MAXFLOAT), Vector3D(-MAXFLOAT)) { }
    BoundingBox(const Vector3D& _Min, const Vector3D& _Max): BoundingVolume(),
	AxisAlignedBox(_Min, _Max) { }
    BoundingBox(AxisAlignedBox& x): BoundingVolume(), AxisAlignedBox(x) { }

    virtual SpanList *FindAllIntersections(Ray& ray);
    virtual int NearestInt(Ray& ray, float& t, float maxt);
    virtual Vector3D FindNormal(const Vector3D& intersection);
    virtual void ApplyTransform(const Matrix& tform);
    virtual Object3D *Dup() const;
    virtual int IsInside(Vector3D& v);
    virtual float PtDistance(Vector3D& v);
    virtual AxisAlignedBox BBox() const;
    virtual void CountMe() const { Statistics::Objects::BBox++; }
    virtual float SurfaceArea() const;
	virtual void Describe();

    virtual void AddObject(const Object3D&);
    virtual void AddObject(Object3D *);

    friend class Ray;
private:
    int ComputeMinMaxT(Ray& ray, float *tmin, float *tmax);
};

// -------------------------------------------------------------------
// Aggregate
//		Basically a collection of objects that get transformed
//		together.
// -------------------------------------------------------------------
class Aggregate : public Object3D {
public:
    Aggregate(): Object3D(AGGREGATE) { surf = 0; }
    //ADDED objName
    Aggregate(const Aggregate& x);
    Aggregate(const ObjectList&, CString name);

    virtual int NearestInt(Ray& ray, float& t, float maxt);
    virtual SpanList *FindAllIntersections(Ray& ray);
    virtual Vector3D FindNormal(const Vector3D& intersection);
    virtual void ApplyTransform(const Matrix& tform);
//    virtual void TransformSurface(SurfaceList& clist, const Matrix& tform);
    virtual Object3D *Dup() const;
    virtual float PtDistance(Vector3D& v);
    virtual AxisAlignedBox BBox() const;
    virtual void CountMe() const;
	virtual void SetName(CString newName);

    ObjectList& List() { return list; }
    void AddObject(Object3D *);
    void AddObject(const Object3D&);
	void AppendObject(Object3D *);
    void AppendObject(const Object3D&);
	Object3D *ObjectAt(int index);

	virtual void Describe();
    void NewSurface(HallSurface *);
    
private:
	CString name;
    ObjectList list;
};
/*
// This type of Object has a canonical form, so you "transform" it
// by transforming rays that you're intersecting with it.
class InverseRayObject : public Object3D {
public:
    InverseRayObject(int f = 0): Object3D(f) {
	mat = inv = IdentityMatrix();
    }
    InverseRayObject(Surface *clr, int f = 0): Object3D(clr, f) {
	mat = inv = IdentityMatrix();
    }

    // Pass along responsibility for intersection calculations.
    virtual int NearestInt(Ray& ray, float& t, float maxt) = 0;
    virtual SpanList *FindAllIntersections(Ray& ray) = 0;

    // Apply a transform to the object.  This function modifies
    // the mat and inv matrices to reflect the transformation.
    virtual void ApplyTransform(const Matrix& tform);

protected:

    // Apply inverse transformation (world space -> canonical space)
    Ray InvertRay(Ray&);
    Vector3D InvertPt(const Vector3D&);
    Vector3D InvRotatePt(const Vector3D&);

    // Apply forward transformation (canonical space -> world space)
    Ray TransformRay(Ray&);
    Vector3D TransformPt(const Vector3D&);
    Vector3D RotatePt(const Vector3D&);

    // Transform spans from canonical space to world space.

    Matrix mat, inv;
};
*/

// -------------------------------------------------------------------
// Sphere
//		Derived from Object3D.
// -------------------------------------------------------------------

class Sphere : public Object3D {
public:
    Sphere(const Vector3D& NewCenter, float NewRadius, 
		HallSurface *NewColor, Mapping* NewMapping = NULL);	//added mapping 6/12/2000

	virtual ~Sphere();
    virtual int NearestInt(Ray& Ray, float& t, float maxt);
    virtual SpanList *FindAllIntersections(Ray& ray);
    virtual Vector3D FindNormal(const Vector3D& intersection);
    virtual void ApplyTransform(const Matrix& tform);
    virtual Object3D *Dup() const;
    virtual int IsInside(Vector3D& v);
    virtual float PtDistance(Vector3D& v);
    virtual AxisAlignedBox BBox() const;
    virtual void CountMe() const { Statistics::Objects::Sphere++; }
	virtual void Describe();
	virtual void GetDerivatives(ShadingInfo& shade);
private:
    Vector3D center;		// location of sphere's center
    float radius;		// radius
    float radiussq;		// radius^2 (precomputed for efficiency)
};


// -------------------------------------------------------------------
// Planar
//	Root of a hierarchy of planar objects (infinite plane, 
//	polygon, ring).
// -------------------------------------------------------------------
class Planar : public Object3D {
public:
    Planar(HallSurface *NewColor, Mapping *NewMapping = NULL);
    Planar(float A, float B, float C, float D, HallSurface *NewColor, Mapping *NewMapping = NULL);
    Planar(const Vector3D& norm, float D, HallSurface *NewColor, Mapping *NewMapping = NULL); 

    virtual int NearestInt(Ray& Ray, float& t, float maxt);
    virtual SpanList *FindAllIntersections(Ray& ray);
    virtual Vector3D FindNormal(const Vector3D& Intersection);
    virtual void ApplyTransform(const Matrix& tform);
    virtual Object3D *Dup() const = 0;
    virtual float PtDistance(Vector3D& v);
    virtual void CountMe() const = 0;
	virtual void Describe() { }
protected:
    Vector3D Normal;			// a, b, c of plane equation
    float d;				// d of plane equation
    int bad;				// nonzero if normal is of
					// zero length.
};
/*
// -------------------------------------------------------------------
// Plane
//	Infinite plane.
// -------------------------------------------------------------------
class Plane : public Planar {
public:
    Plane(float A, float B, float C, float D, HallSurface *NewColor, Mapping *NewMapping = NULL);
    Plane(const Vector3D& norm, float D, HallSurface *NewColor, Mapping *NewMapping = NULL): 
        Planar(norm, D, NewColor, NewMapping) { }

		virtual ~Plane() { if (mapping) delete mapping; }
    virtual int NearestInt(Ray& Ray, float& t, float maxt);
    virtual Object3D *Dup() const;
    virtual void CountMe() const { Statistics::Objects::Plane++; }
};

// -------------------------------------------------------------------
// Polygon
//	Derived from Planar.
//	Much of the information in this class is precomputed to speed
//	intersection calculations.
// -------------------------------------------------------------------
class Polygon : public Planar {
public:
    Polygon(): Planar(0) {}
    Polygon(int NewNumVerts, Vector3D *NewVertices, HallSurface *NewColor, Mapping *NewMapping = NULL);
    Polygon(const Polygon& x);

    virtual ~Polygon();
    virtual int NearestInt(Ray& ray, float& t, float maxt);
    virtual void ApplyTransform(const Matrix& tform);
    virtual Object3D *Dup() const;
    virtual AxisAlignedBox BBox() const;
    virtual void CountMe() const { Statistics::Objects::Polygon++; }
protected:
    int NumVerts;		// Number of vertices
    Vector3D *Vertices;		// Pointer to an array of NumVerts vertices

    // Projection information
    int which;			// Which coordinate to ignore in projection
    float umin, umax;		// Bounding box of projection
    float vmin, vmax;		//

    void Precompute();		// Computes projection and some other stuff
};
*/
// ------------------------------------------------------------------
// Triangle Class
// ------------------------------------------------------------------
class Triangle : public Planar {
public:
	Triangle(): Planar(0) { }
	Triangle(Vertex *NewVertices, HallSurface *NewColor, Mapping *NewMapping = NULL);
	Triangle(const Triangle& tri);
	
	virtual ~Triangle();
	virtual int NearestInt(Ray& ray, float& t, float maxt);
	virtual void ApplyTransform(const Matrix& tform);
	virtual Object3D *Dup() const;
	virtual AxisAlignedBox BBox() const;
	virtual Vector3D FindNormal(const Vector3D& v); //normal interpolation
	virtual void CountMe() const { Statistics::Objects::Triangle++; }
	Vertex&	GetVertex(int n);
	virtual void GetDerivatives(ShadingInfo& shade);
	virtual void Describe();
public:
	Vertex *Vertices;		//Pointer to the array of vertices
	
protected:	

	//Projection information	
	int which;				// Which coordinate to ignore in projection
    float umin, umax;		// Bounding box of projection
    float vmin, vmax;		//

	float alpha, beta;		//used for temporary storage of intersection in the triangle
    void Precompute();		// Computes projection and some other stuff

	friend class UVMapping;
};


// ------------------------------------------------------------------
// Span class
// ------------------------------------------------------------------

class Span {
public:
    Span() { }
    Span(float _tmin, Object3D *_omin, float _tmax, Object3D *_omax)
    {
	tmin = _tmin;  omin = _omin;
	tmax = _tmax;  omax = _omax;
    }
    friend int Compare(const Span& x, const Span& y) {
	if (x.tmin < y.tmin)
	    return -1;
	else if (y.tmin < x.tmin)
	    return 1;
	else return 0;
    }
    friend int Overlap(const Span& x, const Span& y) {
	if (x.tmin > y.tmax || x.tmax < y.tmin)
	    return 0;
	else return 1;
    }

    friend World;
//    friend InverseRayObject;

    // These functions each take a pair of SpanLists, operate
    // on them and return a new SpanList according to the
    // operation (union, intersection, difference) being
    // performed.  The SpanList arguments are destroyed.
    friend SpanList *MergeUnion(SpanList *, SpanList *);
    friend SpanList *MergeIntersection(SpanList *, SpanList *);
    friend SpanList *MergeDifference(SpanList *, SpanList *);
private:
    float tmin, tmax;
    Object3D *omin, *omax;
};

class SpanList {
public:
    SpanList() { }
    SpanList(Span& x) { spans.Insert(x); }
    void AddSpan(const Span& x) { spans.Insert(x); }
    Span ExtractMinSpan() { return spans.ExtractMin(); }
    int NumSpans() const { return spans.N(); }
private:
    Heap<Span> spans;
};

#endif