// ================================================================
//  File: oort.h					12/26/1999
//	By Konstantin Khanyants			kon@unix.temple.edu
//	Description:
//		Main header file for OORT.  Including this should be
//		everything you need.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ================================================================

//OpenGL interface files
#include <gl/gl.h>
#include <gl/glu.h>

#include <malloc.h>
#include <ctype.h>
#include <fstream.h>
#include <iomanip.h>
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
//#include <values.h>

#include <FLOAT.H>

#ifndef MAXFLOAT
#define MAXFLOAT FLT_MAX
#endif

#ifndef M_PI
#define M_PI 3.141592654
#endif

// For Heap (binary heap-based priority queue) class template
#include "Heap.h"

// For SimpleList class template
#include "List.h"

// For SimpleStack class template
#include "Stack.h"

//FOR 3DS file reading
#include "3DSFile.h"

//local OpenGL data types
#include "OpenGLTypes.h"

// For Vector2D class
#include "Vector2D.h"

// For Vector3D class
#include "Vector.h"

#include "ImageFile.h"

//the mesh list structure
#include "MeshListGL.h"

//for the 3d studio class
#include "C3DStudio.h"

// For Matrix class
#include "Matrix.h"

// For Statistics
#include "Stats.h"

//for camera
#include "Camera.h"

// For texture-mapping
#include "Mapping.h"
#include "Colormap.h"
#include "Noise.h"
//surface.h

// For Object3D class and its descendents
#include "Object3d.h"

#include "Texture.h"
#include "Bump.h"

#include "Surface.h"

// For Light class and its descendents
#include "Light.h"

// For adaptive supersampling
#include "Distrib.h"

#ifndef _OORT_H_

#define _OORT_H_

//class NoiseMaker;
//class WaveSource;
//class FreqSource;
//class Object3D;
//class Vector3D;

// ----------------------------------------------------------------
// GlobalNoise.
//	For now, this class encapsulates the classes that I feel
//	may need to be passed into Textures and other places that
//	use them; different solid noise functions may be useful for
//	different Textures, for instance.  But for now, it is
//	simpler to have a global solid noise function, to which the
//	user can specify the random number seed.  All textures look
//	to GlobalNoise::Noise for solid noise, and they similarly
//	look to GlobalNoise::Waves and GlobalNoise::Freqs for waves
//	and frequencies.  I'll have to break these out sometime.
// ----------------------------------------------------------------

class GlobalNoise {
public:
  static NoiseMaker *Noise;
  static WaveSource *Waves;
  static FreqSource *Freqs;
};

// -----------------------------------------------
//	Returns true if the x and y are within
//	Limits::Small of each other.
// -----------------------------------------------
inline int
LessThanEpsilon(const float& x, const float& y)
{
    float diff = x - y;
    Statistics::EpsilonTotal++;
    if (diff >= 0)
      return 0;
    Statistics::EpsilonNeeded++;
    return -diff > Limits::Small;
//  return (diff < 0) && (-diff > Limits::Small);
}

// ------------------------------------------------------------------
// Vertex Class encapsulates information about a vertex
// ------------------------------------------------------------------
class Vertex {
public:
	Vector3D Pos;			//Position
	Vector2D UV;			//texture uv coordinates
	Vector3D Normal;		//normal to the vertex - needed for OpenGL
};

// -------------------------------------------------------------------
// Ray class.  A ray is defined by a location and a direction.
// The direction is always normalized (length == 1).
// -------------------------------------------------------------------
class Ray {
public:
    // constructors
    Ray(const Vector3D& wherefrom, const Vector3D& whichdir) {
	loc = wherefrom;
	dir = Normalize(whichdir);
	Precompute();
    }

    // -----------------------------------
    // Extrapolate ray given parameter
    // -----------------------------------
    Vector3D Extrap(float t) const {
	return loc + dir * t;
    }

    // -----------------------------------
    // Return parameter given point on ray
    // -----------------------------------
    float Interp(Vector3D& x) const {
	for (int i = 0; i < 3; i++)
	    if (fabs(dir[i]) > Limits::Small)
		return (x[i] - loc[i]) / dir[i];
	return 0;
    }

    // -----------------------------------
    // Reflects direction of reflection for the ray, 
    // given the normal to the surface.
    // -----------------------------------
    Vector3D ReflectRay(const Vector3D& N) const {
	return N*2*DotProd(N, -dir) + dir;
    }

    void ApplyTransform(Matrix& tform) {
	loc = tform * loc;
	dir = RotateOnly(tform, dir);
	Precompute();
    }
private:
    Vector3D loc, dir;		// Describes ray
    Vector3D ll, ld, dd;	// Various vectors with corresponding
				// elements multiplied together.
    void Precompute() {
	ll = loc * loc;
	ld = loc * dir;
	dd = dir * dir;
    }

    friend class World;
    friend class AxisAlignedBox;
    friend class Object3D;
    friend class Sphere;
	friend class Triangle;
    friend class BoundingBox;
    friend class Planar;
    friend class Polygon;

    friend class HallSurface;
    friend class DistributedHall;
};

// ---------------------------------------------------------
// KayKajiyaNode
//	Contains the information needed to traverse the 
//	bounding volume hierarchy.  It is named after the 
//	type of bounding volume traversal OORT uses (invented 
//	by Kay and Kajiya).  KayKajiyaNodes exist to get 
//	Insert()'ed and ExtractMin()'ed from the heap used 
//	to traverse the hierarchy.
// ---------------------------------------------------------
class KayKajiyaNode {
public:
    // -------------------------------------------------------
    // Constructors.
    // -------------------------------------------------------
    KayKajiyaNode() { }
    KayKajiyaNode(Object3D *NewObj, float& NewT)
    {
	obj = NewObj;
	t = NewT;
    }

    // -------------------------------------------------------
    // Used by Heap<KayKajiyaNode> to order the nodes.
    // -------------------------------------------------------
    friend int Compare (KayKajiyaNode *x, KayKajiyaNode *y) {
	if (x->t < y->t)
	    return -1;
	else if (x->t > y->t)
	    return 1;
	else return 0;
    }
private:
    float t;		// Intersection parameter
    Object3D *obj;	// Pointer to object intersected

    friend World;
};

// -------------------------------------------------------------------
// Support functions
// -------------------------------------------------------------------

// ---------------------------------------------------------
// RandomValue
//	Returns a random floating-point value between the two 
//	values.  Range is inclusive; the function should 
//	occasionally return exactly a or b.
// ---------------------------------------------------------
inline float
RandomValue(float a, float b)
{
    float range = (float) fabs(a - b);
    return ((float) rand() / RAND_MAX) * range + ((a < b) ? a : b);
}

// ---------------------------------------------------------
// Swaps two values.
// ---------------------------------------------------------
template<class T>
inline void 
Swap(T& x, T& y)
{
    T temp = x;
    x = y;
    y = temp;
}

// ---------------------------------------------------------
// InRange
//	Returns nonzero if min <= candidate <= max.
// ---------------------------------------------------------
template<class T>
inline int 
InRange(T _min, T _max, T candidate)
{
    return candidate >= _min && candidate <= _max;
}

#endif