// ===================================================================
//  File: surface.h							12/26/1999
//	By Konstantin Khanyants					kon@unix.temple.edu
//	Description:
//		Header file for Surface classes.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#ifndef _SURFACE_H_

#define _SURFACE_H_

// Forward-declare classes
class World;

// ---------------------------------------------------------
// ShadingInfo 
//	Encapsulates all the information needed to shade a
//	point.  All members are public, so they can be easily 
//	accessed by the various classes that help do
//	shading in OORT.
// ---------------------------------------------------------
class ShadingInfo {
public:
    Vector3D p;			// Point being shaded
    Vector3D normal;		// Normal at the intersection point
    Vector3D incident;		// Incident direction
    Vector3D reflected;		// Direction of perfect reflection
    
    //added for bump mapping
	Vector3D dudt;			//partial derivative for u on the surface
	Vector3D dvdt;			//partial derivative for v on the surface

    Object3D *obj;		// Object being shaded
    Ray *ray;			// Incident ray
    World *world;		// World the shading is occurring in
};

// ---------------------------------------------------------
// Surface
//	Surface computes the color at an intersection point
//	using the information in a ShadingInfo structure.
// ---------------------------------------------------------

class Surface {
public:
    virtual ~Surface() { }
    virtual void ApplyTransform(const Matrix&) = 0;
    virtual void PreMulTransform(const Matrix&) = 0;
    virtual Surface *Dup() const = 0;

    virtual int Transparent() = 0;
    virtual void BeginShadeLight(ShadingInfo& shade) = 0; //was Vector3D& loc) = 0;
    virtual RGBColor ShadeLight(ShadingInfo& shade, Vector3D& dir) = 0; 

    virtual RGBColor ComputeColor(ShadingInfo& shade, int depth) = 0;
	virtual void Describe() {}
};

class HallSurface : public Surface {
public:
    HallSurface() {
	Bump = NULL;
	Diffuse_Tex = NULL;
	Tex_Used = Bump_Used = FALSE;	
	Tex_Amount = Bump_Amount = 1.0;
	Ambient = Specular = Reflected = Transmitted = Diffuse_Col = RGBColor(0);
	//name = NULL;
    }

	virtual ~HallSurface();
    virtual void SetBumpMap(ImageBumpMap *);
    virtual void SetAmbient(const RGBColor &);
    virtual void SetDiffuseTexture(ImageMap *);
	virtual void SetDiffuseColor(const RGBColor &);
    virtual void SetSpecular(const RGBColor& col, float specnum);
    virtual void SetTransmitted(const RGBColor &, float refract);
    virtual void SetReflected(const RGBColor& col);
	virtual BOOL TextureUsed()  {return Tex_Used;}
	virtual BOOL BumpUsed()		{return Bump_Used;}
	virtual void SetBumpFlag(BOOL);
	virtual void SetTextureFlag(BOOL);
	virtual void SetTexAmount(float val);
	virtual void SetBumpAmount(float val);
	virtual void SetShinStrength(float shin) {shin_strength = shin;}

    virtual void ApplyTransform(const Matrix&);
    virtual void PreMulTransform(const Matrix&);
    virtual Surface *Dup() const;
    virtual int Transparent();
    virtual RGBColor ComputeColor(ShadingInfo& shade, int depth);

    virtual void BeginShadeLight(ShadingInfo& shade); //was Vector3D& loc);
    virtual RGBColor ShadeLight(ShadingInfo& shade, Vector3D& dir); 

	virtual ImageMap* GetDiffuseTexture() 	{return Diffuse_Tex;}
	virtual RGBColor GetDiffuse()			{return Diffuse_Col;}
	virtual ImageBumpMap* GetBumpMapTexture()	{return Bump;}
	virtual RGBColor GetAmbient()	  {return Ambient;}
	virtual RGBColor GetSpecular() 	  {return Specular;}
	virtual RGBColor GetTransmitted() {return Transmitted;}
	virtual RGBColor GetReflected() {return Reflected;}
	virtual float	GetTexAmount();
	virtual float	GetBumpAmount();
	virtual float	GetShininess()	{return specnum;}//shin_strength;}
	virtual float	GetSpecNum()		{return specnum;}
	virtual float	GetIndex()		{return index;}
	virtual CString GetName();
	virtual void	SetName(CString newName);//char* newName);
	
	//OpenGL
	virtual void SetPolygonMode(GLenum polyMode) { PolygonMode = polyMode; }
	virtual void SetShadeModel(GLenum shadeModel) { ShadeModel = shadeModel; }
	virtual void SetTwoSided(BOOL twoSided) { TwoSided = twoSided; }
	
	virtual BOOL IsSmooth();
	virtual void RenderMaterialGL();
	virtual void Describe();

protected:
	CString		name;			//the name according to 3DS material
    float		specnum;		// Specularity factor
	float		shin_strength;   // the presence of the Specular color in the hilight <not variable>
    float		index;			// Index of refraction for object
    RGBColor	diffuse;		// Computed when BeginShadeLight() called
    
	BOOL 		Tex_Used;		//flag: is diffuse texture used or color
	BOOL 		Bump_Used;		//is bump map used
	
	RGBColor	Ambient;		//Ambient term
	RGBColor	Specular;		//Specular term
	RGBColor	Transmitted;	//The transmitted color <gray>
	RGBColor	Reflected;		//Reflected color <gray>	 		
	float		Tex_Amount;		//texture amount
	float		Bump_Amount;	//the amount of bump texture 0.00 ~ 3.0
	ImageMap 	*Diffuse_Tex;		// Diffuse color of the object
	ImageBumpMap *Bump; 
	RGBColor 	Diffuse_Col;	//the color used if no texture is used 
	
	//OpenGL only
	GLenum		PolygonMode;      // GL_LINE for wire frame, GL_FILL for all other types
    GLenum		ShadeModel;       // GL_FLAT for wire frame and flat shading type, GL_SMOOTH for all other types
    BOOL		TwoSided;

	friend class SurfaceList;

};


class DistributedHall : public HallSurface {
public:
    class DistribParms {
    public:

	DistribParms(int _min = 4, int _max = 16, int cand = 10, 
		     float ang = 0.0f, 
		     RGBColor thresh = RGBColor(0.4f, 0.3f, 0.6f)) 
	{
	    MinSamples = _min;  
	    MaxSamples = _max;
	    Candidates = cand;
	    Angle = ang;
	    Threshold = thresh;
	}

	// Routines for read-only access to parameters.
	int GetMinSamples() const { return MinSamples; }
	int GetMaxSamples() const { return MaxSamples; }
	int GetCandidates() const { return Candidates; }
	float GetAngle() const { return Angle; }
	RGBColor GetThreshold() const { return Threshold; }

	// Routines to write to parameters.
	void SetMinSamples(int n) { MinSamples = n; }
	void SetMaxSamples(int n) { MaxSamples = n; }
	void SetCandidates(int n) { Candidates = n; }
	void SetAngle(float ang) { Angle = ang; }
	void SetThreshold(const RGBColor& x) { Threshold = x; }
    private:
	// Number of samples permitted (default = 4, 16).
	int MinSamples, MaxSamples;

	// Number of candidates to consider when computing next sample.
	int Candidates;

	// Angle of disparity permitted distributed rays.  Range [0, pi].
	// Default = 0.
	float Angle;

	// Threshold for contrast--default is (0.4, 0.3, 0.6).
	RGBColor Threshold;
    };

    // Constructor starts us out with no textures or bump maps at all.
    DistributedHall() {
		Bump = NULL;
		Diffuse_Tex = NULL;
		Tex_Used = Bump_Used = false;	
		Ambient = Specular = Reflected = Transmitted = Diffuse_Col = RGBColor(0);
	}

    // Routines to set up 
    virtual void SetTransmitParms(const DistribParms&);
    virtual void SetReflectParms(const DistribParms&);
    virtual RGBColor ComputeColor(ShadingInfo& shade, int depth);
    virtual Surface *Dup() const;
protected:
    virtual RGBColor SampleSpace(ShadingInfo&, Vector3D&, 
				 const DistribParms&, int, long *);
    virtual Vector3D NewSample(const DistribParms&, const Vector3D *, int);
private:
    DistribParms TransParms;	// Ray distribution parameters
				// for transmitted light
    DistribParms ReflectParms;	// Ray distribution parameters
				// for reflected light
};


#endif