// ===================================================================
//  File: texture.h							12/26/1999
//	By Konstantin Khanyants					kon@unix.temple.edu
//	Description:
//		Header file for texture implementations.
//
//	Original Copyright:     The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ===================================================================

#ifndef __TEXTURE__

#define __TEXTURE__

class ShadingInfo;

// ---------------------------------------------------------
// Texture
//	Implements various methods of taking intersection 
//	points on objects and turning them into colors.
// ---------------------------------------------------------

class Texture {
public:
    Texture() { reference = IdentityMatrix(); }
    Texture(const Matrix& ref): reference(ref) { }
    virtual ~Texture() { }

    virtual void ApplyTransform(const Matrix& tform);
    virtual void PreMulTransform(const Matrix& tform);
  	virtual RGBColor GetColor(ShadingInfo& shade) const = 0;
    virtual Texture *Dup() const = 0;

protected:
    Matrix reference;	// Matrix to apply to intersection points

    virtual Vector3D PreprocessVector(ShadingInfo& shade) const;
};

// ---------------------------------------------------------
// TextureMapping
//	Implements a Texture that applies an inverse mapping
//	after applying the reference matrix.
// ---------------------------------------------------------
class TextureMapping : public Texture {
public:
    TextureMapping();
	TextureMapping(const Matrix &ref);

    virtual ~TextureMapping();
    virtual Texture *Dup() const = 0;
protected:

    virtual Vector3D PreprocessVector(ShadingInfo& shade) const;

};

// ---------------------------------------------------------
// PureColor
//	Special texture that just returns an RGB triple 
//	regardless of the intersection point.
// ---------------------------------------------------------

class PureColor : public Texture {
public:
    PureColor() { }
    PureColor(const RGBColor& clr): Color(clr) { }
    PureColor(float r, float g, float b): Color(RGBColor(r, g, b)) { }

    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
    virtual void ApplyTransform(const Matrix& tform) { }
private:
    RGBColor Color;	  // Color associated with this texture
};

// ---------------------------------------------------------
// Checkerboard.
// ---------------------------------------------------------
class Checkerboard : public TextureMapping {
public:
    Checkerboard(float XSize, float YSize,
		 Texture *Text1, Texture *Text2,
		 const Matrix& ref /*,Mapping *map = 0*/);
    RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    float xsize, ysize;
    Texture *text1, *text2;
};


// ---------------------------------------------------------
// ImageMap
//	Maps an image onto an object.  Bilinearly 
//	interpolating the colors is optional.
// ---------------------------------------------------------

class ImageMap : public TextureMapping {
public:
    ImageMap(const Matrix& ref, int Interp, int Once, const char *filename, 
			float rot = 0.0, float xscale = 1, float yscale = 1,
			GLint minFilter = GL_NEAREST, GLint magFilter = GL_NEAREST, 
			GLenum wrapS = GL_REPEAT, GLenum wrapT = GL_REPEAT, GLenum mode = GL_DECAL, float tex_amount = 1.0);
    ImageMap(const Matrix& ref, int Interp, int Once, ImageFile *file,
			float rot = 0.0, float xscale = 1, float yscale = 1, 
			GLint minFilter = GL_NEAREST, GLint magFilter = GL_NEAREST,
			GLenum wrapS = GL_REPEAT, GLenum wrapT = GL_REPEAT, GLenum mode = GL_DECAL, float tex_amount = 1.0);
	ImageMap(const ImageMap& initMap);

    virtual ~ImageMap();
    virtual RGBColor GetColor(ShadingInfo& shade) const;
	virtual ImageFile* GetImageFile() { return file; }

	//to update texture from the Materials dialog
	virtual void SetImageFile(char* filename);
	virtual void SetImageFile(ImageFile* newFile);

    virtual Texture *Dup() const;
	virtual void SetAmount(float val) { tex_amt = val; }
	virtual float GetAmount() { return tex_amt; }

	virtual void InitTextureObjectGL();		//ADDED 4/3/2000 for OpenGL
	//call this only from the routine that is set up with 
	//an OpenGL pipeline
	virtual void RenderTextureGL();

private:  //Internal OpenGL functions
	virtual void DeleteTextureObjectGL();

protected:
    int interp; 	// Boolean--if true, bilinear interpolation
					// is performed.
    int once;		// Boolean--if true, the image is mapped once
					// rather than tiled.
	float tex_amt;  // texture amount
    ImageFile *file;	// Image to map
    Vector3D scale;		// Scale by this vector after mapping.
    float rot_angle;	// rotation from 3D Studio map definition in radians

	//OpenGL texture parameters
	GLuint		TextureObject;    // the OpenGL texture object name
	GLenum		WrapS, WrapT;     // texture wrapping (GL_CLAMP, GL_REPEAT)
    GLenum		MinFilter;        // minification filter (GL_NEAREST, GL_LINEAR, GL_NEAREST_MIPMAP_NEAREST etc.)
    GLenum		MagFilter;        // magnification filter (GL_NEAREST, GL_LINEAR)
    GLenum		Mode;             // texture mode (GL_DECAL, GL_MODULATE etc.)

	friend class CMaterialDialog;
};

// ---------------------------------------------------------
// SolidTexture
//	This class encapsulates the various solid textures
//	such as Bozo, Wood, and Marble.
// ---------------------------------------------------------
class SolidTexture : public Texture {
public:
    SolidTexture() { }
    SolidTexture(const Matrix& ref, ColorMap *CMap);
    virtual RGBColor GetColor(ShadingInfo& shade) const = 0;
    virtual Texture *Dup() const = 0;
protected:
    ColorMap *cmap;
};

// ---------------------------------------------------------
// Bozo
//	A solid texture that maps the solid noise function 
//	directly into a color map.
// ---------------------------------------------------------
class Bozo : public SolidTexture {
public:
    Bozo(const Matrix& Ref, float Turbulence, ColorMap *CMap);
    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    float turbulence;
};

// ---------------------------------------------------------
// Wood
//	Solid texture to simulate wood.
// ---------------------------------------------------------
class Wood : public SolidTexture {
public:
    Wood(const Matrix& Ref, float Turbulence, ColorMap *CMap);
    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    float turbulence;
};

// ---------------------------------------------------------
// Marble
//	Solid texture to simulate marble.
// ---------------------------------------------------------
class Marble : public SolidTexture {
public:
    Marble(const Matrix& Ref, float Turbulence, ColorMap *CMap);
    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    float turbulence;
};

// ---------------------------------------------------------
// Gradient texture.
// ---------------------------------------------------------

class Gradient : public SolidTexture {
public:
    Gradient(const Matrix& Ref, float Turbulence, ColorMap *CMap);
    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    float turbulence;
};

// ---------------------------------------------------------
// WoodUpstill
//	Wood texture derived from Listing 16.15 in Steve 
//	Upstill's book, _The RenderMan Companion_.
//	(Addison-Wesley, 1990).
// ---------------------------------------------------------
class WoodUpstill : public SolidTexture {
public:
    WoodUpstill(const Matrix& ref = IdentityMatrix(),
		float _ringscale = 10, 
		const RGBColor& _light = RGBColor(0.3f, 0.12f, 0.03f), 
		const RGBColor& _dark = RGBColor(0.05f, 0.01f, 0.005f));
    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    float ringscale;
    RGBColor light, dark;
};

// ---------------------------------------------------------
// Granite texture.
//	Also derived from Upstill.
// ---------------------------------------------------------
class Granite : public SolidTexture {
public:
    Granite(const RGBColor& _K,
	    const Matrix& _reference = IdentityMatrix());
    virtual RGBColor GetColor(ShadingInfo& shade) const;
    virtual Texture *Dup() const;
private:
    RGBColor K;
};

// ---------------------------------------------------------
// Some support functions for textures.
// ---------------------------------------------------------

// ---------------------------------------------------------
// posmod()
//	Returns ((x < 0) ? y - x%y : x%y)/y.  
//	Guaranteed to range [0, 1].
// ---------------------------------------------------------
inline float
posmod(float x, float y)
{
    float ret = (float) fmod(x, y);
    if (ret < 0)
	ret += y;
    return ret / y;
}

// ---------------------------------------------------------
// cycloidal()
//	Maps the fractional portion of x to a sinusoid.
// ---------------------------------------------------------
inline float 
cycloidal(float x) {
    float temp = (float) fmod(x, 1);
    if (temp < 0)
	temp += 1;
    return (float) sin(temp * 2 * M_PI);
}

// ---------------------------------------------------------
// triangle_wave() 
//	Maps the fractional portion of x to a triangle wave.
// ---------------------------------------------------------
inline float 
triangle_wave(float x) {
    float offset = (float) fmod(x, 1);
    if (offset < 0)
	offset += 1;
    if (offset > 0.5)
	offset = 1 - offset;
    return offset + offset;
}

// ---------------------------------------------------------
// MixColors
//	Linearly combines two colors according to a (a is
//	in the range [0, 1]).
// ---------------------------------------------------------
inline RGBColor
MixColors(const RGBColor& clr0, const RGBColor& clr1, float a)
{
    return RGBColor(a) * clr0 + RGBColor(1 - a) * clr1;
}

// ---------------------------------------------------------
// Hermite
//	Hermite curve interpolation function.
//	Returns 0 if val < min.  Returns 1 if val > max.
//	Otherwise, returns an interpolated value between 0 and 1.
// ---------------------------------------------------------
inline float 
Hermite(float _min, float _max, float val) 
{
    if (val < _min)
	return 0;
    if (val > _max)
	return 1;
    float t = (val - _min) / (_max - _min);
    return t*t*(-3 + 2*t) + 1;
}

#endif
