// ================================================================
// 	File: World.h					12/24/1999
//	By Konstantin Khanyants
//
//	Description:
//				Header file for World class.
//
//	Original Copyright	    The Object-Oriented Ray Tracer (OORT)
//            				Copyright (C) 1993 by Nicholas Wilt.
//
// ================================================================

#ifndef _WORLD_H_

#define _WORLD_H_

class CRTWindow;
class CPreviewPic;

// ---------------------------------------------------------
// World 
//	Contains all the information required to perform 
//	ray tracing on a set of objects.
// ---------------------------------------------------------
class World {

public:
    // For now, the available anti-aliasing modes are none (one ray
    // per pixel) and adaptive (as described in the book).
    enum AntiAliasModes { None, Adaptive };

//OpenGL members/////////////////////////////////////
public:

	BOOL	GetRenderingInProgress() { return m_bRenderingInProgress; }
	void	SetRenderingInProgress(BOOL on) { m_bRenderingInProgress = on; }

protected:	
	MeshListGL		m_MeshListGL;
	int				m_Version;
	// Default color- ambient and Back color
    RGBColor		m_Ambient;		  // Ambient lighting
	RGBColor		m_BackColor;
	Vector3D		m_SceneCenter;  ///used for auto centering

	BOOL			m_CameraChanged;
	BOOL			m_MaterialsChanged;	//PROTOTYPE RELATED TO MATERIAL DIALOG
	BOOL			m_bAutoCenter; 
	BOOL			m_LightsChanged;
	Camera*			m_pCamera;

	//used in thread communication
	BOOL			m_bRenderingInProgress;

//Ray Tracing members////////////////////////////////
protected:
    // Viewer parameters
    Vector3D ViewerLoc;		  // Location of viewer
    Vector3D LookAt;		  // Point that viewer is looking at
    Vector3D UpVector;		  // Up vector

    // Global list of objects
    Matrix ViewTransform;	  // View transform
    ObjectList Objects; 	  // Object tree <BASE>
	ObjectList RTObjects;	  // Object list based on the <BASE> but for ray tracer,
							  // this list is released at the end of the RT session
    LightList  Lights;		  // List of lights
	SurfaceList Surfaces;	  //added 6/9/2000 
	CameraList Cameras;		  //added 6/10/2000

    // Depth limit.  Limits recursion due to spawned rays.
    // Default is 6 was 8.
    int DepthLimit;     	  // Depth limit

    // Image parameters
    int HorzRes, VertRes;	  // Horizontal and vertical resolution

    float ScreenHeight; 	  // Height of screen
    float ScreenWidth;		  // Width of screen
    float ScreenDist;		  // Screen distance

    int StartScan, EndScan;	  // Starting and ending scanlines
            			  	  // to put out
    long NumPixels;
	
    // Sampling parameters
    enum AntiAliasModes mode;	// Naive or Adaptive (default = Naive)
    int MinSamples, MaxSamples;	// Minimum and maximum samples to take
								//	(default = 4 and 16)
    int NumCandidates;		  // Number of candidates to consider
							  //	(default = 10)
    RGBColor ContrastThreshold;	// Contrast threshold (higher contrasts
								//	cause more rays).  Default =
								// 	(0.4, 0.3, 0.6) per Mitchell.

    // State variables
//    int printpers;		// True if a period is printed for
				// every scan line computed
//    int interactive;		// True if user can hit a key and
				// get an update during the 'trace
//    int hierarchyseed;		// Seed to give srand() before
				// before hierarchy is generated
//    int numtries;		// Number of hierarchies to try

    // Goes through list of objects and breaks all aggregates into
    // their component objects.  Returns the resulting list.
    // Destroys the parameter list.  Thus, to disband all the
    // aggregates in the list of bounded objects in the World,
    // this function is called as follows:
    //		Objects = DisbandAggregate(Objects);
    ObjectList DisbandAggregates(ObjectList&);

	//ADDED 4/4/2000 for Temporary Ray Tracer Objects
	void GetRTObjects(ObjectList& objs);
	void ClearRTObjects(ObjectList& objs);

    // Goes through objects and culls the ones that are unbounded,
    // like planes and quadrics.  Those get tested always.
//    void RemoveUnboundeds();

    // Builds a hierarchy in the given list out of the given.
    static void BuildOneHierarchy(ObjectList& list, Object3D **objs, int n);

    // Assumes ObjectList contains the list of all objects in the World;
    // replaces ObjectList with a Goldsmith-Salmon-manufactured BVH.
    // Parameter is the number of hierarchies to consider before deciding
    // on one.
    void BuildHierarchy(int tries);

    // Culls those bounding boxes that only contain one object
    void CullSingletons(Object3D *bv, ObjectList *ParentList);

    Object3D *FindAtten(ObjectList &optr, Ray& ray,
			RGBColor& Atten, float maxt);
    Object3D *Find_Nearest(Ray& ray, float& t);

    Vector3D NewSample(Vector3D *, int);

    void CountObjectsRec(ObjectList& list);
    static void HierarchyCostRec(BoundingVolume *, float&, float);

public:
    // -------------------------------------------------------
    // Constructor sets some default values.
    // -------------------------------------------------------
    World();

	// -------------------------------------------------------
    // Destructor deallocates the image array.
    // -------------------------------------------------------
    ~World();//{ }//delete theImageFile; }
    
	//To get the surfaces from Material Dialog
	SurfaceList* GetSurfaces() { return &Surfaces; }
	CameraList*  GetCameras() { return &Cameras; }
	ObjectList*  GetObjects() { return &Objects; }
	LightList*	 GetLights() { return &Lights; }

    // -------------------------------------------------------
    // Functions to set parameters of the World.
    // -------------------------------------------------------

    // Sets width and height of the viewing frustrum.
    void SetScreenWidth(float NewScreenWidth);
    void SetScreenHeight(float NewScreenHeight);
    void SetScreenDistance(float NewScreenDist);

    // Sets background color.  Pixels where no object is
    // intersected will be assigned this color.
    void SetBackgroundColor(RGBColor& NewBackground);

    // Set ambient light present in World.
    void SetAmbientLight(RGBColor& NewAmbient);

    // Set starting/ending scanlines to ray trace.
    // The ending scanline is not inclusive, i.e. to ray-trace
    // scanlines 0-479 for a 640x480 image you set Start to
    // 0 and End to 480.
//    void SetStart(int Start);
//    void SetEnd(int End);

    // Set depth limit of recursion.
    void SetDepthLimit(int NewDepthLimit);

    // Set viewer parameters
    //	LookAt will be in the middle of the rendered image.
    //	ViewerLoc will be the location of the viewer.
    //	UpVector will point toward the top of the rendered
    //	    image (this vector should be normalized).
    void SetViewerParameters(Vector3D& NewLookAt,
			     Vector3D& NewViewerLoc,
			     Vector3D& UpVector);

	void SetViewerParamsCamera(Camera* camUser);	//for 3DS camera
    
	// Set number of hierarchies to generate before deciding
    // on one.  Half a dozen is more than enough.
//    void SetNumTries(int _numtries);

    // Anti-aliasing parameters.
    void SetAntiAliasMode(enum AntiAliasModes _mode);
    void SetSamples(int _min, int _max);
    void SetNumCandidates(int);
    void SetContrastThreshold(const RGBColor& thresh);

    // Returns the ambient light in the World.
    RGBColor GetAmbientLight() { return m_Ambient; }

    // Returns nonzero if the depth given has exceeded the
    // World's depth limit of recursion.
    int HitDepthLimit(int depth) { return depth >= DepthLimit; }

    // Add objects and light sources to World.
    void AddObject(Object3D& Obj);
    void AddObject(Object3D *Obj);
    void AddLight(Light *newlight);
	void AddCamera(Camera* newCam);

    // Count the objects in the World.
    void CountObjects();

    // Returns the cost of the World's hierarchy.
    static float HierarchyCost(const ObjectList&);

    // Finds the nearest intersection along the ray, shades
    // that point.  Call initially with depth=0.
    RGBColor FindColor(Ray& ray, int depth);

    // Computes the color of a pixel.  See world.cpp for how
    // to use.
    RGBColor PixelColor(float xtarg, float ytarg);

    // Computes the color due to light sources using the
    // intersection point and other information in 
    RGBColor DiffuseLight(ShadingInfo& shade, Surface *surf);
    Object3D *ShadowAtten(Ray& ray, RGBColor& Atten, float maxt);

    // Using the parameters in the World so far, generate a
    // bounding volume hierarchy, ray-trace the image and
    // copy it to the window.
	// pass in the pointer to the current ray tracing window
    void RayTrace(CRTWindow* pRTWindow);

	// using the parameters in the World.. mainly a sphere,
	// and an associated texture.. ray trace an interactive preview
	// of the material
	void RayTracePreview(CPreviewPic* previewPic);  

};

// Routine that adds an object to a bounding box hierarchy using 
// the Goldsmith-Salmon heuristic.
void AddNewObject(ObjectList *root, Object3D *addme);

#endif