// ===================================================================
//  File: TgaFuncs.cpp					5/1/2000
//	By Konstantin Khanyants
//	Description:
//		Functions of the ImageFile class that are used
//		to open and convert Targa files into internal image format 
//		and also convert the format into a TGA file.
//
// ===================================================================

#include "stdafx.h"
#include "RayTracer/include/ImageFile.h"
#include <memory.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//////
// Macros to deal with unsigned chars as efficiently as compiler allows
typedef unsigned char U_CHAR;
#define UCH(x)	((int) (x))

// Private version of data source object

typedef struct _TGA_source_struct * TGA_source_ptr;

typedef struct _TGA_source_struct {
  
  	unsigned char *buffer; 	// a pointer to one row of our image
	unsigned char *rleBuf;	// the rle buffer used in ReadRleRow

  	CFile *pFile;			//pointer to a file object
  
	// State info for reading RLE-coded pixels; both counts must be init to 0
	int		inRawPacket;	// flags processing state for RLE data
	int		inRLEPacket;	// flags processing state for RLE data

	unsigned int packetSize;	// records current RLE packet size in bytes

} TGA_source_struct;


//static function declarations
static void ReadRleRow(TGA_source_ptr tga_ptr, unsigned int n, int bpp);
static void StripAlpha(unsigned char *s, int n );

static const int Get_8bit_gray_row = 0;
static const int Get_8bit_row	   = 1;
static const int Get_16bit_row     = 2;
static const int Get_24bit_row	   = 3;

// Targa also defines a 32-bit pixel format with order B,G,R,A.
// We presently ignore the attribute byte, so the code for reading
// these pixels is identical to the 24-bit routine above.
// This works because the actual pixel length is only known to read_pixel.
#define Get_32bit_row  Get_24bit_row

#define RLEBUFSIZE	512			// size of largest possible RLE packet

// For expanding 5-bit pixel values to 8-bit with best rounding

static const unsigned char c5to8bits[32] = {
    0,   8,  16,  25,  33,  41,  49,  58,
   66,  74,  82,  90,  99, 107, 115, 123,
  132, 140, 148, 156, 165, 173, 181, 189,
  197, 206, 214, 222, 230, 239, 247, 255
};


//ReadTga()/////////////////////////////////////////////////////////////
// Initializes our image with the tga file that it reads
BOOL ImageFile::LoadTga()
{
	unsigned char targaheader[18];
  	int idlen, cmaptype, subtype, flags, interlace_type, components; 
  	int byteCount, col, t, s_index;
  	unsigned int width, height, maplen, pixel_size, mapWidth, bCount;
  	int is_bottom_up;
	TGA_source_struct  source;
	BOOL Rle_Data, AlphaPresent;
	int Row_Type;
	RGBTRIPLE* colormap;
	unsigned char *rowPtr;
	

	//initialize the buffers
	source.rleBuf = NULL;
	source.buffer = NULL;
	colormap = NULL;

	//initialize RLE processing flags of the tga struct
	source.inRawPacket = 0;
	source.inRLEPacket = 0;	
	source.packetSize = 0;


	//try .. catch block
	try {
	
	CFile file(filename, CFile::modeReadWrite);
	source.pFile = &file;
	
#define GET_2B(offset)	((unsigned int) UCH(targaheader[offset]) + \
			 (((unsigned int) UCH(targaheader[offset+1])) << 8))

	//read the 18 byte header
	file.Read((LPVOID) targaheader, 18);

	// Pretend "15-bit" pixels are 16-bit --- we ignore attribute bit anyway
    if (targaheader[16] == 15)
    	targaheader[16] = 16;
    	
  	idlen = UCH(targaheader[0]);
  	cmaptype = UCH(targaheader[1]);
  	subtype = UCH(targaheader[2]);
  	maplen = GET_2B(5);
  	mapWidth = UCH(targaheader[7]);
  	width = GET_2B(12);
  	height = GET_2B(14);
  	pixel_size = UCH(targaheader[16]) >> 3;
  	flags = UCH(targaheader[17]);			// Image Descriptor byte

  	is_bottom_up = ((flags & 0x20) == 0);	// bit 5 set => top-down
  	interlace_type = flags >> 6;			// bits 6/7 are interlace code

  	if (cmaptype > 1 ||					// cmaptype must be 0 or 1
        pixel_size < 1 || pixel_size > 4 ||
        (UCH(targaheader[16]) & 7) != 0 || // bits/pixel must be multiple of 8
        interlace_type != 0)				 // currently don't allow interlaced image
		throw new CFileException;
  	
  	if (subtype > 8) {
    	// It's an RLE-coded file
    	Rle_Data = TRUE;
    	//allocate space for the largest possible rle row
		source.rleBuf = new unsigned char[RLEBUFSIZE];
		
		if (source.rleBuf == NULL)
			throw new CFileException;
		subtype -= 8;
  	
  	} else
    // Non-RLE file
    	Rle_Data = FALSE;

  	// Now should have subtype 1, 2, or 3
  	components = 3;		// until proven different
  	AlphaPresent = FALSE;
  	SamplePerPixel = RGBSample;

 	switch (subtype) {
  	case 1:			// Colormapped image
    	if (pixel_size == 1 && cmaptype == 1)
      		Row_Type = Get_8bit_row;
    	else
      		throw new CFileException;
    	break;
  	case 2:			// RGB image
    	switch (pixel_size) {
    		case 2:
      			Row_Type = Get_16bit_row;
      			break;
    		case 3:
      			Row_Type = Get_24bit_row;
      			break;
    		case 4:
      			Row_Type = Get_32bit_row;
				AlphaPresent = TRUE;
      			break;
    		default:
      			throw new CFileException;
      			break;
    	}
    	break;
  	case 3:			// Grayscale image
    	components = 1;
    	SamplePerPixel = GRAYSample;
    	if (pixel_size == 1)
      		Row_Type = Get_8bit_gray_row;
   	 	else
      		throw new CFileException;
    	break;
  	default:
    	throw new CFileException;
    	break;
  	}	//end switch(subtype)
	
	
	//seek past the unwanted idfield area
	file.Seek(idlen, CFile::current);
	
	
	//if there is a colormap
  	if (maplen > 0) {
  		//only handle colormaps that are 256 bytes in length and 
  		// get indexed at 0
  	  	if (maplen > 256 || GET_2B(3) != 0)
      		throw new CFileException;
    	
		// Presently only handles 24-bit BGR format
		if (mapWidth != 24)
			throw new CFileException;
			
		// Allocate space to store the colormap
		colormap = new RGBTRIPLE[maplen];
		if (colormap == NULL)
			throw new CFileException;
    	// and read it from the file
    	
    	byteCount = ((mapWidth + 7) >> 3) * maplen;    	
    	//read the colormap with one huge read
    	file.Read((LPVOID) colormap, byteCount);	
	} 
	else {
    	if (cmaptype)		// but you promised a cmap!
      		throw new CFileException;
    	colormap = NULL;
  	}
  	
  	image_width = width;		//assign our class members
  	image_height = height;

	//allocate space for the image
	//NOTE: image_width, image_height, and SamplePerPixel must
	// have already been assigned
	if (!AllocateImage())
		throw new CFileException;

	bCount = pixel_size * image_width;			//number of bytes in one row
	source.buffer = new unsigned char[bCount];  //get the space for one row
	//was the allocation successful
	if (source.buffer == NULL)
		throw new CFileException;
	
	
	for (int row = 0; row < image_height; row++) {
		
		//if the image is upside down then count rowPtr from bottom
		if (is_bottom_up)
			rowPtr = pImage[(image_height-1) - row];
		else
			rowPtr = pImage[row];

		//read the row by ReadRleRow or with just a file read
		if (Rle_Data)
			ReadRleRow(&source, bCount, pixel_size);
		else
			file.Read((LPVOID) source.buffer, bCount);
	
		//switch based on the type of the image we have
		switch (Row_Type) {
			case (Get_8bit_gray_row):
			
				//if 8 bit then just memcopy the row to our image
				memcpy(rowPtr, source.buffer, image_width);
				break;
			
			case (Get_8bit_row):
				//if palettized image then get the index from the buffer and index
				// the colormap with it
				for (col = 0, s_index = 0; col < image_width*SamplePerPixel; col+=3) {
					t = UCH(source.buffer[s_index++]);
					//RGB values in the colormap are really BGR as read from file 
					rowPtr[col+0] = (U_CHAR) colormap[t].rgbtRed;
					rowPtr[col+1] = (U_CHAR) colormap[t].rgbtGreen;
					rowPtr[col+2] = (U_CHAR) colormap[t].rgbtBlue;
				}
				break;
				
			case (Get_16bit_row):
				//if 16 bit row then read from the buffer but do some formatting first
				for (col = 0, s_index = 0; col < image_width*SamplePerPixel; col+=3) {
					t = UCH(source.buffer[s_index++]);
					t += UCH(source.buffer[s_index++]) << 8;
					// We expand 5 bit data to 8 bit sample width.
					// The format of the 16-bit (LSB first) input word is
					//     xRRRRRGGGGGBBBBB
					
					rowPtr[col+2] = (U_CHAR) c5to8bits[t & 0x1F];
					t >>= 5; 
					rowPtr[col+1] = (U_CHAR) c5to8bits[t & 0x1F];
					t >>= 5;
					rowPtr[col+0] = (U_CHAR) c5to8bits[t & 0x1F];
				}
				break;
			
			case (Get_24bit_row):
				//strip the alpha component if present
				if (AlphaPresent)
					StripAlpha(source.buffer, bCount);
				//for 24 bit and 32 bit rows the data is sent straight to our image
				for (col = 0; col < image_width*SamplePerPixel; col+=3) {
					rowPtr[col+0] = (U_CHAR) UCH(source.buffer[col+2]); // change BGR to RGB order
					rowPtr[col+1] = (U_CHAR) UCH(source.buffer[col+1]);
					rowPtr[col+2] = (U_CHAR) UCH(source.buffer[col+0]);
				}						
				break;
				
			default:
				throw new CFileException;
				break;
			
		}//end switch

	} //end for
	
	} //end try
	catch (CFileException* e) {
	CString pcsMsg = "Error reading file ";
	if (source.rleBuf) delete [] source.rleBuf;
	if (source.buffer) delete [] source.buffer;
	if (colormap) 	   delete [] colormap;
	pcsMsg += filename;
	AfxMessageBox(pcsMsg);
	e->Delete();
	return FALSE;
	}
	
	//delete the buffers
	delete [] source.rleBuf;
	delete [] source.buffer;
	delete [] colormap;
	return TRUE;
}


//Read and decode one RLE encoded row in our image buffer
// n is buffer size in bytes
// bpp is number of bytes per pixel
static void ReadRleRow(TGA_source_ptr tga_ptr, unsigned int n, int bpp)
{
	unsigned int	value;
	unsigned int	i;
	unsigned char	*q;
	unsigned char	*p = tga_ptr->buffer;
	unsigned char	*rleBuf = tga_ptr->rleBuf;
	int				&inRawPacket = tga_ptr->inRawPacket;
	int				&inRLEPacket = tga_ptr->inRLEPacket;
	unsigned int	&packetSize	= tga_ptr->packetSize;
	CFile 			*pFile = tga_ptr->pFile;

	while (n > 0) {

		if (inRLEPacket) {

			if (packetSize * bpp > n) {
				value = n / bpp;		// calculate pixel count
				packetSize -= value;
				n = 0;
			}
			else {
				n -= packetSize * bpp;
				value = packetSize;
				packetSize = 0;
				inRLEPacket = 0;
			}
			while (value > 0) {
				*p++ = rleBuf[0];
				if (bpp > 1) *p++ = rleBuf[1];
				if (bpp > 2) *p++ = rleBuf[2];
				if (bpp > 3) *p++ = rleBuf[3];
				value--;
			}
		}
		else if (inRawPacket) {
		
			if (packetSize * bpp > n) {
				value = n;
				packetSize -= n / bpp;
				n = 0;
			}
			else {
				value = packetSize * bpp;	// calculate byte count
				n -= value;
				inRawPacket = 0;
			}
			for (i = 0; i < value; ++i) 
				*p++ = *q++;
		}
		else {
			
			// No accumulated data in buffers, so read from file
			pFile->Read((LPVOID) &packetSize, 1);
			
			if (packetSize & 0x80) {
				
				packetSize &= 0x7f;
				packetSize++;
				
				if (packetSize * bpp > n) {
					value = n / bpp;		// calculate pixel count
					packetSize -= value;
					inRLEPacket = 1;
					n = 0;
				}
				else {
					n -= packetSize * bpp;
					value = packetSize;
				}
				
				pFile->Read((LPVOID) rleBuf, bpp);
				
				while (value > 0) {
					*p++ = rleBuf[0];
					if (bpp > 1) *p++ = rleBuf[1];
					if (bpp > 2) *p++ = rleBuf[2];
					if (bpp > 3) *p++ = rleBuf[3];
					value--;
				}
			}
			else {
				packetSize++;
				
				// Maximum for packetSize is 128 so as long as RLEBUFSIZ
				// is at least 512, and bpp is not greater than 4
				// we can read in the entire raw packet with one operation.
				pFile->Read((LPVOID) rleBuf, bpp*packetSize);	
					
				
				// But is there enough room to copy them to our line buffer?
				if (packetSize * bpp > n) {
					value = n;				// number of bytes remaining
					packetSize -= n / bpp;
					inRawPacket = 1;
					n = 0;
				}
				else {
					value = packetSize * bpp;	// calculate byte count
					n -= value;
				}
				for (i = 0, q = rleBuf; i < value; ++i) 
					*p++ = *q++;
			}
		}
	}
}


//strips away the alpha component from a 32bit row
static void StripAlpha(unsigned char *s, int n )
{
	int				i;
	unsigned char	*p;

	// Copy the RGB components omitting the alpha.  Perform the
	// copy in its own buffer.  This algorithm is probably
	// specific to the 80x86 since byte orderring will be different
	// on a 680x0 processor.
	p = s;
	for (i = 0; i < n; ++i) {
		if ((i % 4) == 3) 
			++p;
		else 
			*s++ = *p++;
	}
}


/////////
//SaveTga()//////////////////////////////////////////////////////////
// Stores our image in a tga file
BOOL ImageFile::SaveTga(int mode24)
{
	char targaheader[18];   //the header fields for the file
	unsigned char *buffer;
	BOOL buf_alloc = FALSE;

	//try..catch block
	try {

	//open the file
	CFile file(filename, CFile::modeReadWrite | CFile::modeCreate);
  	
  	//Write the header///

  	// Set unused fields of header to 0
  	memset((void *) targaheader, 0, sizeof(targaheader));
  
 	//we omit the color table and its related fields
  
  	//set the width and height fields of the header
  	targaheader[12] = (char) (image_width & 0xFF);
  	targaheader[13] = (char) (image_width >> 8);
  	targaheader[14] = (char) (image_height & 0xFF);
  	targaheader[15] = (char) (image_height >> 8);
  	targaheader[17] = 0x20;		// Top-down, non-interlaced

	if (!mode24) {
    	targaheader[2] = 3;		// image type = uncompressed gray-scale
    	targaheader[16] = 8;	// bits per pixel
		if (SamplePerPixel == RGBSample) { //force grayscale
			buffer = new unsigned char[image_width];
			buf_alloc = TRUE;
		}
	} 
  	else {			// must be RGB
    	targaheader[2] = 2;		// image type = uncompressed RGB
    	targaheader[16] = 24;
		buffer = new unsigned char[image_width * RGBSample];
		buf_alloc = TRUE;
	}

	//write the 18 byte header
	file.Write((LPVOID) targaheader, 18);
			
	//Write the image data///	

	if (!mode24) {	//force gray
  	//if our image is grayscale
  	if (SamplePerPixel == GRAYSample) {
		for (int row = 0; row < image_height; row++)
			file.Write((LPVOID) pImage[row], image_width);
	}			
    //otherwise if it is RGB convert to grayscale    
    else {
    	for (int row = 0; row < image_height; row++) {
			int aux_col = 0;
			for (int col = 0; col < image_width; col++) {
				buffer[col] = (unsigned char) 
							  (pImage[row][aux_col+0] * .299 +
							   pImage[row][aux_col+1] * .587 +
							   pImage[row][aux_col+2] * .114);
			aux_col += 3;
			}
			file.Write((LPVOID) buffer, image_width);
		}
   	}
	} //if not mode24
	else {		//if is mode24
		if (SamplePerPixel == GRAYSample) {
			for (int row = 0; row < image_height; row++) {
				int aux_col = 0;
				for (int col = 0; col < image_width * RGBSample; col += 3) {
					buffer[col+0] = pImage[row][aux_col];
					buffer[col+1] = pImage[row][aux_col];
					buffer[col+2] = pImage[row][aux_col];
				
				aux_col++;
				}
				file.Write((LPVOID) buffer, image_width * RGBSample);
			}
		}
		else {
			for (int row = 0; row < image_height; row++) {
				for (int col = 0; col < image_width * RGBSample; col += 3) {
					buffer[col+0] = pImage[row][col+2];
					buffer[col+1] = pImage[row][col+1];
					buffer[col+2] = pImage[row][col+0];
				}
				file.Write((LPVOID) buffer, image_width * RGBSample);
			}
		}
	}
	if (buf_alloc)
		delete [] buffer;
	} //end try
	catch(CFileException* pe) {
		if (buf_alloc)
			delete [] buffer;
		CString pcsMsg = "Error: Can't save to file ";
		pcsMsg += filename;
		pe->Delete();
		AfxMessageBox(pcsMsg);
		return FALSE;
	}
	return TRUE;
}	
